/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "Adc64Core.h"
#include "Adc64SignalRouter.h"

#include "DominoDeviceManager.h"
#include "adc64-common/Adc64SystemFsm.h"
#include "DeserializersControl/DeserializersControlBuildParams.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "lib-common/ReadoutAppConfigDispatcher.h"
#include "remote-control-server/RemoteControlServerStatus.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_appcore)
Q_LOGGING_CATEGORY(lc_appcore, "afi.appcore")
}

Adc64Core::Adc64Core(QObject *parent)
    : ReadoutAppCore{parent}
    , configDb(new ReadoutAppConfigDb(this))
    , deviceMgr(new DominoDeviceManager(nullptr))
    , fsm(new Adc64SystemFsm(this))

{
    auto &signalRouter = Adc64SignalRouter::instance();
    signalRouter.setConfigDb(configDb);
    signalRouter.setReadoutAppConfigDispatcher(
                new ReadoutAppConfigDispatcher(this));

    signalRouter.setDominoDeviceManager(deviceMgr);
    signalRouter.setCore(this);
    signalRouter.setDeviceMgrFsm(fsm);
    signalRouter.connectAll();

    deviceMgr->setObjectName("Adc64DevMng");

    refreshProgramDescription();
}

void Adc64Core::setConfig(const ConfigKey &key, const BaseDeviceAppConfig &newConfig)
{
    if (newConfig == config)
        return;

    config = newConfig;
    config_key = key;
    updateRemoteStatus();
}

void Adc64Core::setDeserializersControlBuildParams(DeviceIndex index, const DeserializersControlBuildParams &bp)
{
    auto it = config.knownConfigs.find(index);
    if(it == config.knownConfigs.end())
        return;
    if(it->tqdcCore.getNCh() != bp.getNCh()) {
        it->tqdcCore.setNCh(bp.getNCh());
        emit configChanged(config);
    }
}

void Adc64Core::runError(const QString &text)
{
    if ((fsm->getState()==FsmState::Starting
         || fsm->getState()==FsmState::Run
         || fsm->getState()==FsmState::Stopping)
            && config.allowRemoteControl)
        emit sendRunError(rcRunId, text);
}

void Adc64Core::performLocalStart()
{
    fsm->setRequestState(FsmState::Starting);
}

void Adc64Core::performLocalStop()
{
    fsm->setRequestState(FsmState::Stopping);
}

void Adc64Core::performCalibrate()
{
    fsm->setRequestState(FsmState::Calibrate);
}

void Adc64Core::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    ReadoutAppCore::onFsmStateChange(state, oldState);

    switch (state) {
    case FsmState::Error :
        if(config.isRun()) {
            config.setRunState(false);
            emit configChanged(config);
        }
        break;
    case FsmState::Ready :
    {
        switch (oldState) {
        case FsmState::Calibrate :
        case FsmState::FindAdcDelay :
        case FsmState::Stopping:
            qInfo() << QString("%1 completed successfully")
                       .arg(fsmStateToString(oldState));
            break;
        default:;
        }
        break;
    }
    case FsmState::Starting :
    {
        config.setRunState(true);
        emit configChanged(config);
        fsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Calibrate :
    case FsmState::FindAdcDelay :
    {
        ++calKey;
        fsm->setMainCalKey(calKey);
        if(state == FsmState::Calibrate)
            emit startCalibration(calKey);
        else
            emit startAdcSrdDesCalibration(calKey, adcSerDesCalHardMode, adcSerDesCalResDir);
        fsm->setRequestState(FsmState::Ready);
        break;
    }
    case FsmState::Run :
    {
        qInfo() << QString("Start completed successfully");
        break;
    }
    case FsmState::Stopping :
    {
        config.setRunState(false);
        emit configChanged(config);
        fsm->setRequestState(FsmState::Ready);
        break;
    }
    default:;
    }
}

void Adc64Core::onRemoteStartRequest(const RcRunId &runId, const QVariant &start_event_number)
{
    qInfo(lc_appcore).noquote()
        << QString("Got remote request to start %1").arg(runId.toString());

    static bool warn_start_event_number = true;
    if(start_event_number.isValid() && warn_start_event_number) {
        qWarning(lc_appcore).noquote()
            << "Remote start with start_event_number not implemented yet.";
        warn_start_event_number = false;
    }

    if(rcRunId.runIndex == runId.runIndex &&
        rcRunId.runNumber > runId.runNumber) {
        const auto &msg = QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                              .arg(rcRunId.runNumber)
                              .arg(runId.runNumber);
        emit sendRunWarning(rcRunId, msg);
        return;
    }

    if(fsm->getState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Ready))
                              .arg(fsmStateToString(fsm->getState()));
        emit sendRunWarning(rcRunId, msg);
        return;
    }

    rcRunId = runId;
    fsm->setRequestState(FsmState::Starting);
}

void Adc64Core::onRemoteStopRequest(const RcRunId &runId, const QVariant &stop_event_number)
{
    static bool warnStopEvNum = true;
    if(stop_event_number.isValid() && warnStopEvNum) {
        warnStopEvNum = false;
        qWarning(lc_appcore).noquote()
            << "Stopping with 'stop_event_number' should be fixed";
    }
    bool success=true;
    if(rcRunId.runIndex != runId.runIndex) {
        qWarning(lc_appcore).noquote()
            << QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                   .arg(rcRunId.runIndex)
                   .arg(runId.runIndex);
        success = false;
    }
    if(success && rcRunId.runNumber != runId.runNumber){
        qWarning(lc_appcore).noquote()
            << QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                   .arg(rcRunId.runNumber)
                   .arg(runId.runNumber);
        success = false;
    }


    if(fsm->getState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(fsm->getState()));
        emit sendRunWarning(rcRunId, msg);
        if(fsm->getState() != FsmState::Ready && fsm->getState() != FsmState::Error)
            fsm->setRequestState(FsmState::Error);
        return;
    }
    fsm->setRequestState(FsmState::Stopping);
}

void Adc64Core::remoteStatusUpdated(const RemoteControlServerStatus &status)
{
    ReadoutAppCore::remoteStatusUpdated(status);

    if(underRemoteCtrl == status.portInUse)
        return;
    underRemoteCtrl = status.portInUse;

    const auto &curState = fsm->getState();
    if(curState != FsmState::Ready && curState != FsmState::Error) {
        qWarning(lc_appcore).noquote()
            << "RemoteControlServer disconnected: fsm -> Error";
        fsm->setRequestState(FsmState::Error);
    }
    if(status.portInUse)
        updateRemoteStatus();
}

void Adc64Core::requestAdcSerDesCalibration(bool hardMode, const QString &dir)
{
    if(fsm->getState() != FsmState::Ready){
        qWarning() << "AdcSerDesCalibration can be performed just in Ready state";
        return;
    }
    adcSerDesCalHardMode = hardMode;
    adcSerDesCalResDir = dir;
    fsm->setRequestState(FsmState::FindAdcDelay);
}

void Adc64Core::refreshProgramDescription()
{
    progDescr.setClientIndex(configDb->getClientIndex());

    emit ReadoutAppCore::updateProgramDescription(progDescr);
}

const BaseFsm *Adc64Core::getFsm() const
{
    return fsm;
}

void Adc64Core::updateAdcStatus(const AdcStatus &status)
{
    const auto &index = status.getDeviceIndex();
    emit setStatus(status);
    bool statusAfterReset = status.setup_key.has_value() && (status.setup_key.value() == config_key);

    if (!config.knownConfigs.contains(index)) return;
    //    if(devices[index] != sender()) {
    //        // Got status from previous instance of hw
    //        // DominoDevice was recreated
    //        return;
    //    }

    auto st = fsm->getState();
    if(st==FsmState::Run || st==FsmState::Starting || st==FsmState::Stopping) {
        if(!status.liveMagicCorrect) {
            runError("liveMagic incorrect");
        }
        if(statusAfterReset && !status.adcSelfTestOk) {
            runError("AdcSelfTest failed");
        }
        if(statusAfterReset && config.errOnTrigOnXOff && status.runStatistic) {
            auto blocked = status.runStatistic->getCounter(RunStatistic::Blocked);
            if(blocked) {
                runError("Trig on XOff detected");
                qWarning(lc_appcore).noquote()
                        << QString("[%1] trigOnXOff=%2").arg(index.getIdent()).arg(blocked);
            }
        }
    }

    //    if(hw->isConnected() && !status.liveMagicCorrect){
    //        AdcSetup &setup = systemSetup.adc[cfgIndex];
    //        qWarning(lc_appcore).noquote()
    //            << QString("[0x%2] Warning: Device Adc-%1 failed magic word check. Resetting latest configuration. (%3)")
    //                      .arg(cfgIndex).arg(setup.serialId, 8, 16, QChar('0'))
    //                      .arg(setup.host.toString());
    //        addProcessLog(row, "==== RESET ====");
    //        dumpers[row]->restart();
    //    }

    lastAdcStatus[index] = status;
}

void Adc64Core::updateRemoteStatus()
{
    auto devCnt = config.getDeviceIndexEnabledMap().getEnabled().size();
    QString stateStr;
    if(devCnt)
        stateStr = QString("%1 device(s)").arg(devCnt);
    else
        stateStr = "No devices";

    emit sendRemoteState(stateStr);
}
