#include "Adc64SignalRouter.h"

#include <QObject>

#include "Adc64Core.h"
#include "Adc64MainWindow.h"
#include "AdcCalibrationDialog.h"
#include "AdcConfigWidget.h"
#include "AdcDeviceProxy.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"
#include "DominoDeviceManager.h"
#include "MonitorEventCombiner.h"
#include "RunLogic/RunLogicWidget.h"
#include "adc64-common/DominoDevice.h"
#include "SparseDialog/SparseDialog.h"
#include "base/DeviceId.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "DeviceChTree/DeviceChModel.h"
#include "DeviceChTree/DeviceChTreeView.h"
#include "FirFilter/FirFilterWidget.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "lib-common/ReadoutAppConfigDispatcher.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mstream-mreg-dev-global/MstreamMregDevGlobal.h"
#include "mvc/DeviceStatusTableView.h"
#include "PreamplifierDialog.h"
#include "TrigCtrlStat/TrigCtrlStatWidget.h"
#include "WaveBlc/WaveBlcWidget.h"
#include "LemoMuxDialog.h"

#include "fsm/AdcDeviceMgrFsm.h"
#include "remote-control-server/RemoteControlServer.h"

static constexpr const auto device_filter_list = {
    DEVICE_ID_ADC64,
    DEVICE_ID_ADC64S2,
    DEVICE_ID_ADC64VE,
    DEVICE_ID_ADC64VE_XGE,
    DEVICE_ID_ADC64VE2,
    DEVICE_ID_ADC64ECAL,
    DEVICE_ID_ADC64WR,
    DEVICE_ID_ADC8BEDRS,
    DEVICE_ID_ADC8E,
    DEVICE_ID_EVADC,
    DEVICE_ID_TQDC16VS_DIG,
    DEVICE_ID_ADCM163,
    DEVICE_ID_ADC8E
};

Adc64SignalRouter::Adc64SignalRouter()
{
    setDeviceIdFilter(device_filter_list);
}

void Adc64SignalRouter::setCore(Adc64Core *p)
{
    core = p;
    setReadoutAppCore(p);
}

void Adc64SignalRouter::setMainWindow(Adc64MainWindow *p)
{
    mainWindow = p;
    setReadoutAppMainWindow(p);
}

void Adc64SignalRouter::setReadoutAppConfigDispatcher(ReadoutAppConfigDispatcher *p)
{
    readoutConfDisp = p;
    setConfigDispatcher(p);
}

void Adc64SignalRouter::connectAll() const
{
    ReadoutAppSignalRouter::connectAll();

    connect_DominoDeviceManager();
    connect_lldp();
    connect_CalibrationDialog();

    if (adcSerDesCalDialog && core)
        connect(adcSerDesCalDialog, &AdcSerDesCalDialog::requestCalibration,
                core, &Adc64Core::requestAdcSerDesCalibration);

    if (mainWindow) {
        connect(deviceMgr, &DominoDeviceManager::statusUpdated,
                mainWindow, &Adc64MainWindow::setAdcStatus);

        connect(mainWindow, &Adc64MainWindow::performLocalStart,
                core, &Adc64Core::performLocalStart);
        connect(mainWindow, &Adc64MainWindow::performLocalStop,
                core, &Adc64Core::performLocalStop);

        if(auto adcCfgWidget = mainWindow->getAdcConfigWidget()) {
            if (auto runLogicWidget = adcCfgWidget->getRunLogicWidget()) {
                if(core)
                    runLogicWidget->setController(core->runLogicController);
            }
        }
    }
}

void Adc64SignalRouter::connect_DeviceGlobal(const AdcDeviceProxy *devGlob, DominoDevice *dev) const
{
    ReadoutAppSignalRouter::connect_device(devGlob, dev);

    if(!dev)
        return;

    auto adcDeviceMgrFsm = qobject_cast<AdcDeviceMgrFsm *>(deviceMgrFsm);
    if (adcDeviceMgrFsm) {
        connect(dev, &DominoDevice::calibrateFinished,
                adcDeviceMgrFsm, &AdcDeviceMgrFsm::setDevCalKey);
    }

    if(calibDialog) {
        connect(dev, &DominoDevice::calResultChanged,
                calibDialog, &AdcCalibrationDialog::setCalResult);
    }
    if(core) {
        connect(core, &Adc64Core::startCalibration,
                dev, &DominoDevice::startCalibration);
        connect(core, &Adc64Core::startAdcSrdDesCalibration,
                dev, &ModularMregDevice::startAdcSrdDesCalibration);
    }
    if(mainWindow) {
        if(auto w = mainWindow->getLemoMuxDialog()) {
            connect(dev, &ModularMregDevice::adcLemoMuxBpUpdated,
                    w, &LemoMuxDialog::setAdcLemoMuxBp);
        }
        if(auto w = mainWindow->getAdcConfigWidget()) {
            connect(dev, &ModularMregDevice::adcBuildParamsUpdated,
                    w, &AdcConfigWidget::adcBuildParamsUpdated);

            if(auto tr = w->getTrigCtrlStatWidget()) {
                connect(dev, &ModularMregDevice::trigCtrlBpUpdated,
                        tr, &TrigCtrlStatWidget::setTrigCtrlBp);
            }
            if (auto ff = w->getFirFilterWidget()) {
                connect(dev, &ModularMregDevice::firFilterBpUpdated,
                        ff, &FirFilterWidget::setFirFilterBp);
            }
        }
    }
    if(monEvCombiner) {
        connect(devGlob, &MstreamMregDevGlobal::deviceDataUpdated,
                monEvCombiner, &MonitorEventCombiner::onMStreamDeviceDataUpdated);
    }
}

void Adc64SignalRouter::connect_FSM() const
{
    ReadoutAppSignalRouter::connect_FSM();

    if(!deviceMgrFsm)
        return;

    connect(readoutConfDisp, &ReadoutAppConfigDispatcher::configKeyChanged,
            deviceMgrFsm, &DeviceMgrFsm::setMainSetupKey);

    if(mainWindow) {
        connect(mainWindow, &Adc64MainWindow::printFsmInput,
                deviceMgrFsm, &BaseFsm::printFsmInput);

        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                mainWindow, &Adc64MainWindow::onFsmStateChange);
    }
}

void Adc64SignalRouter::connect_RemoteControlServer() const
{
    ReadoutAppSignalRouter::connect_RemoteControlServer();
    if(core) {
        //todo: move to base
        connect(core, &Adc64Core::sendRemoteState,
                remoteControlServer, &RemoteControlServer::sendState);

        connect(remoteControlServer, &RemoteControlServer::startRequest,
                core, &Adc64Core::onRemoteStartRequest);
        connect(core, &Adc64Core::sendRemoteState,
                remoteControlServer, &RemoteControlServer::sendState);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                core, &Adc64Core::onRemoteStopRequest);
        connect(core, &Adc64Core::sendRunError,
                remoteControlServer, &RemoteControlServer::onRunError);
        connect(core, &Adc64Core::sendRunWarning,
                remoteControlServer, &RemoteControlServer::onRunWarning);
    }
}

void Adc64SignalRouter::connect_MainWindow() const
{
    ReadoutAppSignalRouter::connect_MainWindow();
    if(!mainWindow)
        return;

    if(monEvCombiner) {
        connect(monEvCombiner, &MonitorEventCombiner::monitorEventChanged,
                mainWindow, &Adc64MainWindow::onMonitorEvent);
    }
    if(auto w = mainWindow->getAdcConfigWidget()) {
        connect(mainWindow, &ReadoutAppMainWindow::showRunLogicChanged,
                w, &AdcConfigWidget::setShowRunLogic);
    }
}

void Adc64SignalRouter::connect_ConfigurationDispatcher() const
{
    ReadoutAppSignalRouter::connect_ConfigurationDispatcher();

    auto cd = readoutConfDisp;
    if(!cd)
        return;

    if(core) {
        cd->subscribeConfig(core);
        cd->subscribeCS(core);
    }

    if(configurationManager)
        cd->subscribeCS(configurationManager);
    if(mainWindow) {
        auto adcConfigWidget = mainWindow->getAdcConfigWidget();
        cd->subscribeConfig(adcConfigWidget);
        cd->subscribeTrigCtrlStatConfig(adcConfigWidget->getTrigCtrlStatWidget());
        cd->subscribeFirFilterConfig(adcConfigWidget->getFirFilterWidget());
        cd->subscribeWaveBlcConfig(adcConfigWidget->getWaveBlcWidget());
        cd->subscribePreamplifierConfig(mainWindow->getPreamplifierDialog());
        cd->subscribeAdcLemoMuxConfig(mainWindow->getLemoMuxDialog());
        cd->subscribeConfig(mainWindow->getDeviceTreeModel());
    }
    if(configDb) {
        cd->subscribeConfig(configDb);
        cd->subscribeCS(configDb);
    }
    if(deviceDispatcher)
        deviceDispatcher->subscribeDevices(cd);
    if(deviceMgr) {
        connect(cd, &ReadoutAppConfigDispatcher::deviceListChanged,
                deviceMgr, &DominoDeviceManager::setDevices);
    }
    if(monEvCombiner) {
        connect(cd, &ReadoutAppConfigDispatcher::deviceListChanged,
                monEvCombiner, &MonitorEventCombiner::setDevices);
    }
    if(calibDialog)
        cd->subscribeConfig(calibDialog);
}

void Adc64SignalRouter::connect_DominoDeviceManager() const
{
    if(!deviceMgr)
        return;

    if(core) {
        connect(deviceMgr, &DominoDeviceManager::deserializersControlBpUpdated,
                core, &Adc64Core::setDeserializersControlBuildParams);
    }

    if (mainWindow) {
        connect(deviceMgr, &DominoDeviceManager::deserializersControlBpUpdated,
                mainWindow, &Adc64MainWindow::setDeserializersControlBuildParams);
        connect(deviceMgr, &DominoDeviceManager::statusUpdated,
                mainWindow, &ReadoutAppMainWindow::setStatus);
        if(auto chModel = mainWindow->getDeviceTreeModel()) {
            connect(deviceMgr, &DominoDeviceManager::statusUpdated,
                    chModel, &DeviceChModel::deviceStatusUpdated);
        }
        if(auto chView = mainWindow->getDeviceChTreeView()) {
            connect(deviceMgr, &DominoDeviceManager::statusUpdated,
                    chView, &DeviceChTreeView::deviceStatusUpdated);
        }

        //    connect(mainWindow, &MainWindow::setDeviceIdFilter,
        //            discoverDialog, &DiscoverDialog::setDeviceIdFilter);
    }
}

void Adc64SignalRouter::connect_deviceDispatcher() const
{
    ReadoutAppSignalRouter::connect_deviceDispatcher();

    if(!deviceDispatcher)
        return;

    if(deviceMgr) {
        connect(deviceDispatcher, &DeviceDispatcher::setHostAddress,
                deviceMgr, &DominoDeviceManager::setHostAddress);
    }

    if(mainWindow) {
        if(auto w = mainWindow->getLemoMuxDialog()) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    w, &LemoMuxDialog::setDevices);
        }
        if(auto w = mainWindow->getAdcConfigWidget()) {
            if(auto tr = w->getTrigCtrlStatWidget()) {
                connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                        tr, &TrigCtrlStatWidget::setDeviceList);
            }
        }
        if(auto chView = mainWindow->getDeviceChTreeView()) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    chView, &DeviceChTreeView::setDeviceList);
        }
    }
}

void Adc64SignalRouter::connect_lldp() const
{
    auto &lldpDicovery = LldpDiscovery::instance();
    if(deviceDispatcher)
        lldpDicovery.subscribe(deviceDispatcher);
    if(discoverDialog)
        lldpDicovery.subscribe(discoverDialog);
    //    if(mainWindow)
    //        lldpDicovery.subscribe(mainWindow);

}

void Adc64SignalRouter::connect_CalibrationDialog() const
{
    if(!calibDialog)
        return;

    if (deviceMgrFsm) {
        connect(deviceMgrFsm, &DeviceMgrFsm::stateChanged,
                calibDialog, &AdcCalibrationDialog::onFsmStateChange);
    }
    if(core) {
        connect(calibDialog, &AdcCalibrationDialog::performCalibrate,
                core, &Adc64Core::performCalibrate);
    }
}
