#include "MonitorEventCombiner.h"

#include <QDebug>
#include <QFile>
#include <QList>
#include <QMap>
#include <QMetaType>
#include <QPointer>
#include <QTimer>

#include "base/DeviceIndexEnabled.h"
#include "histogramm/DeviceData.h"

namespace {
const int monitorEventTimeMs = 300;
const int eventTimeoutMs = 500;
}

MonitorEventCombiner::MonitorEventCombiner(QObject *parent)
    : QObject(parent)
    , eventData(new CompleteEvent)
    , monitorEvent(new CompleteEvent)
    , eventTimeoutTimer(new QTimer(this))
    , monitorTimeoutTimer(new QElapsedTimer)
    , timestamp(new QElapsedTimer)
{
    qRegisterMetaType<DecodedData>();
    qRegisterMetaType<CompleteEvent>();
    qRegisterMetaType<TqdcEvent>();

    eventTimeoutTimer->setSingleShot(true);
    eventTimeoutTimer->setInterval(eventTimeoutMs);
    connect(eventTimeoutTimer, &QTimer::timeout, this, &MonitorEventCombiner::eventTimeout);
}

MonitorEventCombiner::~MonitorEventCombiner()
{
    delete eventData;
    delete monitorEvent;
    delete eventTimeoutTimer;
    delete monitorTimeoutTimer;
    delete timestamp;
}

void MonitorEventCombiner::setDevices(const DeviceIndexEnabledMap &deviceList)
{
    const DeviceIndexSet &newDeviceSet = deviceList.getEnabled();
    if(devSet == newDeviceSet)
        return;
    devSet = newDeviceSet;

    const auto &toDel = devSet-newDeviceSet;
    for(const auto &index : toDel) {
        eventData->tqdcEv.remove(index);
    }
}

void MonitorEventCombiner::onDeviceDataAcquired(const DeviceIndex &index, const TqdcEvent &tqdcEvent)
{
    Q_ASSERT(index == tqdcEvent.index);

    if(!devSet.contains(index)) {
        qWarning() << "Got device data from unknown device:" << index;
        return;
    }

    if (eventData->tqdcEv.size() > devSet.size()) {
        qWarning() << "Event demultiplexer: Too much different id numbers, event buffer reset";
        eventData->clear();
    }
    if(eventData->tqdcEv.isEmpty()) {
        eventTimeoutTimer->start();
        timestamp->start();
    }
    eventData->tqdcEv[index] = tqdcEvent;

    //    bool match = eventData->timestampTrigMatch();
    //    // emit data only when all event numbers match
    //    if (!match) {
    //        qWarning() << "Timestamp mismatch";
    //        return;
    //    }

    if (eventData->tqdcEv.size() == devSet.size()) {
        eventTimeoutTimer->stop();
        maybeSendMonitorEvent();
        eventData->clear();
        timestamp->invalidate();
        //        qDebug() << "Event number Ok";
    }

}

void MonitorEventCombiner::onMStreamDeviceDataUpdated(const DecodedData &data)
{
    for(auto ev : data.events) {
        for(auto it=ev.tqdcEv.cbegin(); it!=ev.tqdcEv.cend();++it) {
            onDeviceDataAcquired(it.key(), *it);
        }
    }
}

void MonitorEventCombiner::eventTimeout()
{
    if(timestamp->elapsed() > eventTimeoutMs)
        qWarning() << QString("EventData timed out, number of event blocks mismatch: %3 blocks, %4 devices")
                          .arg(eventData->tqdcEv.size()).arg(devSet.size());
}

void MonitorEventCombiner::maybeSendMonitorEvent()
{
    if(!timestamp->isValid())
        return;
    if(monitorTimeoutTimer->isValid() && monitorTimeoutTimer->msecsTo(*timestamp) < monitorEventTimeMs)
        return;

    *monitorTimeoutTimer = *timestamp;
    *monitorEvent = *eventData;
    emit monitorEventChanged(*monitorEvent);
}
