﻿//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "EvbMainWindow.h"
#include "ui_EvbMainWindow.h"

#include "AdjustBufferDialog.h"
#include "AdjustBufferModel.h"
#include "Client.h"
#include "Combiner.h"
#include "EvBConfigDispatcher.h"
#include "EvbFileNameDialog.h"
#include "EvBStartupOptions.h"
#include "EvbSignalRouter.h"
#include "FileSplitDialog.h"
#include "FileWriterStats.h"
#include "StatisticDialog.h"
#include "SizeHelper.h"
#include "ClientManagerFsm.h"

#include "add-pnp-dialog/AddPnpDialog.h"
#include "app/AppVersion.h"
#include "base/AbstractConfigConverter.h"
#include "base/DeviceId.h"
#include "base/DeviceIndex.h"
#include "base/Globals.h"
#include "base/ThreadController.h"
#include "base/threadQuit.h"
#include "base/RunDocument.h"
#include "base/RunDocumentConverter.h"
#include "configuration-manager/ConfigurationManager.h"
#include "pnp-server/PnpDiscovery.h"
#include "widgets/LogViewWidget.h"
#include "metric-sender/MetricSender.h"
#include "metric-sender/MetricTypes.h"
#include "remote-control-server/RemoteControlServer.h"
#include "TelnetRemoteControl/TelnetRemoteControl.h"

#include <QActionGroup>
#include <QCheckBox>
#include <QDateTime>
#include <QDebug>
#include <QElapsedTimer>
#include <QFileDialog>
#include <QJsonDocument>
#include <QJsonObject>
#include <QHostAddress>
#include <QInputDialog>
#include <QMessageBox>
#include <QMetaType>
#include <QThread>
#include <QTimer>
#include <QSemaphore>
#include <QSettings>
#include <QScreen>
#include <time.h>
#include <QTime>
#include <iostream>

namespace {
enum COLUMNS{ COL_READOUT_CB, COL_PROG_TYPE, COL_PROG_INDEX, COL_DATA_TYPE,
              COL_CONNECTION_STATE, COL_PROG_HOST, COL_STATUS,
              COL_EV_NUM, COL_TRASHED_EVENTS, COL_ALL_EVENTS,
              COL_FLUSHED_SIZE, COL_TIME_DISTR, COL_TCP_BUFF,
              COL_EV_BUFF, COL_TOTAL};

QString getTcpName(int tcpId) {
    return tcpId ? QString("Tcp #%1").arg(tcpId-1) : QString("Mon");
}
const DaqModuleSet EVB_SEARCH_TARGETS {
    DaqModule(DaqModule::Type::EventBuilder),
    DaqModule(DaqModule::Type::Generator),
    DaqModule(DaqModule::Type::GenericClient),
    DaqModule(DaqModule::Type::MStream),
    DaqModule(DaqModule::Type::VmeDaq),
    DaqModule(DaqModule::Type::Land)
};

static constexpr const char* prop_outType = "outType";
static constexpr const char* prop_id = "id";

void updateFsmLabel(QLabel *label, const QString &prefix, FsmState st)
{
    label->setText(prefix + ": " + fsmStateToString(st));
    auto p = label->palette();
    auto b = p.brush(label->foregroundRole());
    switch (st) {
    case FsmState::Error:
        b.setColor(Qt::red);
        break;
    case FsmState::Ready:
        b.setColor(QColor());
        break;
    case FsmState::Run:
        b.setColor(Qt::darkGreen);
        break;
    default:
        b.setColor(Qt::darkYellow);
    }
    p.setBrush(label->foregroundRole(), b);
    label->setPalette(p);
}

QColor getTcpClientStateColor(TcpClientState st, bool selected)
{
    switch (st) {
    case TcpClientState::Connected:
        return QColor();
    case TcpClientState::Disabled:
        if(selected)
            return {90, 90, 90};
        else
            return Qt::gray; // a0a0a4
    default:
        return Qt::red;
    }
}

constexpr const char *EVB_DATA_TYPE_KEY = "data_type";
constexpr const char *EVB_DATA_TYPE_NON_EVENT = "non-event";
bool isNonEventBaseProgram(const ProgramDescription &pd)
{
    switch(DaqModule::Type(pd.getDaqModule())) {
    case DaqModule::Type::MStream:
    {
        auto idStr = pd.options.value("dev_id");
        bool ok;
        auto id = idStr.toInt(&ok, 16);
        static const QSet<int> nonEvDevId { DEVICE_ID_MSC16VE_E,
                    DEVICE_ID_SOFTWARE, DEVICE_ID_UT24VE};
        return ok && (nonEvDevId.contains(id));
    }
    case DaqModule::Type::EventBuilder:
    {
        auto dtStr = pd.options.value(EVB_DATA_TYPE_KEY);
        return dtStr == EVB_DATA_TYPE_NON_EVENT;
    }
    default:;
    }
    return false;
}
}

using namespace std;

Q_DECLARE_METATYPE(Qt::GlobalColor)

EvbMainWindow::EvbMainWindow(const EvBStartupOptions &so, QWidget *parent)
    : BaseMainWindow(parent)
    , ui(new Ui::EvbMainWindow)
    , configuration_manager (new ConfigurationManager(this))
    , configDb(new EvBConfigDb(this))
    , configDispatcher(new EvBConfigDispatcher(this))
    , evb_fsm(new EventBuilderFsm(this))
    , client_manager_fsm(new ClientManagerFsm(this))
    , combiner(new Combiner())
    , combinerThread(new QThread(this))
    , remoteControl(new RemoteControlServer(this, 0))
    , labelFileErrSize(new QLabel(this))
    , statisticDialog(new StatisticDialog(this))
    , labelEvbFsmState(new QLabel(this))
    , labelCombinerFsmState(new QLabel(this))
    , labelClientManagerFsmState(new QLabel(this))
    , labelCombinerStatus(new QLabel(this))
    , fileSplitDialog(new FileSplitDialog(this))
    , fileNameDialog(new EvbFileNameDialog(this))
    , adjustBufferDialog(new AdjustBufferDialog(this))
    , adjustBufferModel(new AdjustBufferModel(this))
    , applyCacheTimer(new QTimer(this))
    , telnetRc(so.telnetPort ? new TelnetRemoteControl(*so.telnetPort) : nullptr)
{
    registerTypes();

    ui->setupUi(this);
    ui->mainToolBar->hide();
    labelEvbFsmState->setMinimumWidth(120);
    labelCombinerFsmState->setMinimumWidth(150);
    labelClientManagerFsmState->setMinimumWidth(180);
    ui->statusBar->addWidget(labelEvbFsmState);
    ui->statusBar->addWidget(labelCombinerFsmState);
    ui->statusBar->addWidget(labelClientManagerFsmState);
    ui->statusBar->addWidget(labelCombinerStatus);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);

    ui->splitter->addWidget(new LogViewWidget(ui->splitter));
    adjustBufferDialog->setModel(adjustBufferModel);

    auto evbSignalRouter = &EvbSignalRouter::instance();
    evbSignalRouter->setConfigurationManager(configuration_manager);
    evbSignalRouter->setEvBConfigDb(configDb);
    evbSignalRouter->setEvBConfigDispatcher(configDispatcher);
    evbSignalRouter->setMainWindow(this);
    evbSignalRouter->setRemoteControlServer(remoteControl);
    evbSignalRouter->setWidgetRunControl(ui->widgetRunControl);
    evbSignalRouter->setEvbFsm(evb_fsm);
    evbSignalRouter->setClientManagerFsm(client_manager_fsm);
    evbSignalRouter->setAdjustBufferModel(adjustBufferModel);
    evbSignalRouter->setStatisticDialog(statisticDialog);
    if(telnetRc) {
        evbSignalRouter->setTelnetRemoteControl(telnetRc);
        new ThreadController(telnetRc, this);
    }

    evbSignalRouter->connectAll();

    QThread::currentThread()->setObjectName("GUI thread");
    combinerThread->setObjectName("combinerThread");

    QActionGroup *dataOutActionGroup = new QActionGroup(this);
    dataOutActionGroup->addAction(ui->actionOutToFile);
    dataOutActionGroup->addAction(ui->actionOutToTcp);
    dataOutActionGroup->addAction(ui->actionCombinerIdle);
    connect(dataOutActionGroup, &QActionGroup::triggered,
            this, &EvbMainWindow::data_out_changed);
    connect(ui->radioButtonOutFile, &QRadioButton::toggled, this, &EvbMainWindow::dataOutChangedRB);
    connect(ui->radioButtonOutIdle, &QRadioButton::toggled, this, &EvbMainWindow::dataOutChangedRB);
    connect(ui->radioButtonOutTcp, &QRadioButton::toggled, this, &EvbMainWindow::dataOutChangedRB);

    ui->radioButtonOutFile->setProperty(prop_outType, COMBINER_FILE);
    ui->radioButtonOutTcp->setProperty(prop_outType, COMBINER_TCP);
    ui->radioButtonOutIdle->setProperty(prop_outType, COMBINER_IDLE);

    connect(ui->actionOutput_file_name, &QAction::triggered,
            fileNameDialog, &EvbFileNameDialog::show);

    new ThreadController(remoteControl, this);

    configDispatcher->subscribeCS(this);
    configDispatcher->subscribeCS(configDb);
    configDispatcher->subscribeCS(configuration_manager);
    configDispatcher->subscribeEvbConfig(this);
    configDispatcher->subscribeEvbConfig(configDb);
    configDispatcher->subscribeClients(this);
    configDispatcher->subscribeFileSplitterConfig(fileSplitDialog);
    configDispatcher->subscribeWriterConfig(fileNameDialog);
    configDispatcher->subscribeFileSplitterConfig(combiner);
    configDispatcher->subscribeClients(client_manager_fsm);

    configDispatcher->subscribeEvbConfig(adjustBufferDialog);
    configDispatcher->subscribeEvbConfig(adjustBufferModel);
    configDispatcher->subscribeWriterConfig(this);
    configDispatcher->subscribeEvNumberConfig(this);

    configuration_manager->setDefaultConfig(EvBConfig());

    applyCacheTimer->setSingleShot(true);
    applyCacheTimer->setInterval(500);
    connect(applyCacheTimer, &QTimer::timeout, this, &EvbMainWindow::applyCache);
    applyCacheTimer->start();
    init();
}

void EvbMainWindow::init()
{
    setupTable();

    ui->actionEnable_inf_output->setVisible(Globals::getDebugMode());
    ui->statusBar->addWidget(labelFileErrSize);
//    statusBar()->addPermanentWidget(labelFileErrSize);
    statusBar()->setToolTip("reg:\tNumber of regular events\n"
                            "skip:\tNumber of skipped events from clients\n"
                            "eob:\tNumber of 'end of burst' events\n"
                            "clients:\tClient's ids from last combined event\n"
                            "nextEv:\tEvent number of next expected event\n"
                            "monitor:\tNumber of sent events to monitor");
    labelFileErrSize->hide();

    initCombiner();
    updateOutTcpLabelList();
    setup_clients();
    updateGui();
}

void EvbMainWindow::setConfig(const EvBConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;

    applyConfig();
}

void EvbMainWindow::on_actionConfiguration_manager_triggered()
{
    configuration_manager->show();
}

void EvbMainWindow::applyCache()
{
    updatePendingClientStatus();
    updatePendingProgramDescription();
    applyCacheTimer->start();
}

void EvbMainWindow::doStateAction(const FsmState &reqSt)
{
    if(requestedState == reqSt)
        return;
    requestedState = reqSt;

    qDebug() << __PRETTY_FUNCTION__ << reqSt;

    QVariant requestArg;
    switch (reqSt) {
    case FsmState::Run:
        requestArg = QVariant::fromValue(loc_runId);
        break;
    case FsmState::PrepareStopping:
        requestArg = QVariant::fromValue(stopEvNum.value_or(0));
        break;
    default:;
    }

    switch (reqSt) {
    case FsmState::Configure:
    case FsmState::Run:
    case FsmState::PrepareStopping:
    case FsmState::Ready:
        client_manager_fsm->setRequestState(reqSt);
        emit requestState(reqSt, requestArg);
        if(reqSt == FsmState::Run){
            sendRunDocument();
        }
        break;
    default:;
    }
}

void EvbMainWindow::onEvbFsmStateChanged(const FsmState &cur, const FsmState &prev)
{
    updateFsmLabel(labelEvbFsmState, "EvbFsm", cur);
    switch (cur) {
    case FsmState::Error:
    {
        if(prev != FsmState::Init && prev != FsmState::Ready) {
            emit printFsmInput();
            qDebug() << __PRETTY_FUNCTION__ << "send stopping signal on Error state";
            doStateAction(FsmState::Ready);
        }
        break;
    }
    default:;
    }

    updateGuiEnableStates();
}

void EvbMainWindow::updateCombinerFsmLabel(const FsmState &cur_state)
{
    updateFsmLabel(labelCombinerFsmState, "CombinerFsm", cur_state);
}

void EvbMainWindow::updateClientFsmLabel(const FsmState &cur_state)
{
    updateFsmLabel(labelClientManagerFsmState, "ClientManagerFsm", cur_state);
}

void EvbMainWindow::startLocalRun()
{
    if(getFsmState() == FsmState::Ready)
        evb_fsm->setRequestState(FsmState::Starting);
}

void EvbMainWindow::stopLocalRun()
{
    if(getFsmState() == FsmState::Run) {
        stopEvNum.reset();
        evb_fsm->setLookForStopEvNum(stopEvNum.has_value());
        evb_fsm->setRequestState(FsmState::Stopping);
    }
}

void EvbMainWindow::configSelectorChanged(const ConfigIndexName& conf)
{
    BaseMainWindow::configSelectorChanged(conf);

    // Initialize MetricSender
    MetricSender::setBaseTag("index", conf.getIndex().toString());

    initProgramDescription(conf);
}

EvbMainWindow::~EvbMainWindow()
{
    QElapsedTimer timer;
    timer.start();

    auto allClients = clients.keys();
    ClientIndexSet toDel{allClients.cbegin(), allClients.cend()};
    deleteClients(toDel);

    disconnect(remoteControl, &RemoteControlServer::startRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::stopRequest, nullptr, nullptr);
    disconnect(remoteControl, &RemoteControlServer::connectionStatusUpdated, nullptr, nullptr);

    emit closeProgram(progDesc);

//    Q_ASSERT(ret);

    emit destroyAll();
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"destroyClient(0);";
    //todo: wait all clients are destroyed

    timer.start();

    //Give time to proceed destroyed signal from Combiner
//    qInfo()<<"[MainWindow] "<<timer.elapsed()<<"threadQuiting(combinerThread";
    if(!threadQuiting(combinerThread, 5000, &timer)){
        qWarning() << "Combiner's thread isn't stopped";
//        qInfo()<<"[MainWindow] "<<timer.elapsed()<<"Combiner's thread isn't stopped";
    }

    delete ui;
}

void EvbMainWindow::on_pushButtonAdd_clicked()
{
    QVector<ProgramDescription> freeProgs = getFreeProgs();
    if(!freeProgs.size())
        return;

    AddPnpDialog dialog(progDescrsCache, true, this);
    PnpDiscovery::subscribeCacheUpdaed(&dialog);

    dialog.set_function_filter(std::bind(&EvbMainWindow::isFreeProg, this, std::placeholders::_1));

    auto clients_map = config.getClientIndexEnabledMap();
    if(dialog.exec() == QDialog::Accepted) {
        for(auto pd : dialog.getProgDescr()) {
            auto index = pd.getClientIndex();
            clients_map[index] = true;
        }
        emit clientListChanged(clients_map);
    }
}

void EvbMainWindow::on_pushButtonRemove_clicked()
{
    auto cis = getSelectedClients();
    if(cis.isEmpty())
        return;

    QStringList list;
    for(auto c : cis)
        list << c.toString();
    QString message = QString("Are you shure to remove client(s):\n%1?")
                          .arg(list.join('\n'));

    int ret = QMessageBox::warning(this, tr("Confirm removing"), message,
                                   QMessageBox::Yes | QMessageBox::No,
                                   QMessageBox::No);
    if(ret == QMessageBox::No)
        return;

    auto clients_map = config.getClientIndexEnabledMap();
    for(auto index : cis) {
        if(index.isEmpty() || !clients_map.contains(index)) {
            qWarning() << "Don't know what client was selected; removing canceled";
            return;
        }

        clients_map.remove(index);
        qInfo() << QString("Client '%1' deleted").arg(index.toString());
    }
    emit clientListChanged(clients_map);
}

void EvbMainWindow::setupTable()
{
    QTableWidget *table = ui->tableWidget;
    table->setColumnCount(COL_TOTAL);
#if QT_VERSION >= 0x050000
    table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
    table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif

    table->setHorizontalHeaderItem(COL_READOUT_CB, new QTableWidgetItem("readout"));
    table->setHorizontalHeaderItem(COL_PROG_TYPE, new QTableWidgetItem("type"));
    table->setHorizontalHeaderItem(COL_PROG_INDEX, new QTableWidgetItem("index"));
    table->setHorizontalHeaderItem(COL_DATA_TYPE, new QTableWidgetItem("Data type"));
    table->setHorizontalHeaderItem(COL_CONNECTION_STATE, new QTableWidgetItem("Online"));
    table->setHorizontalHeaderItem(COL_PROG_HOST, new QTableWidgetItem("host"));
    table->setHorizontalHeaderItem(COL_STATUS, new QTableWidgetItem("state"));
    table->setHorizontalHeaderItem(COL_EV_NUM, new QTableWidgetItem("evNum"));
    table->setHorizontalHeaderItem(COL_TRASHED_EVENTS, new QTableWidgetItem("trashed"));
    table->setHorizontalHeaderItem(COL_ALL_EVENTS, new QTableWidgetItem("all"));
    table->setHorizontalHeaderItem(COL_FLUSHED_SIZE, new QTableWidgetItem("flushedSize"));

    table->setHorizontalHeaderItem(COL_TIME_DISTR, new QTableWidgetItem("Time distr"));
    table->setHorizontalHeaderItem(COL_TCP_BUFF, new QTableWidgetItem("Raw data buff"));
    table->setHorizontalHeaderItem(COL_EV_BUFF, new QTableWidgetItem("Ev. meta-info"));
}

void EvbMainWindow::on_tableWidget_itemSelectionChanged()
{
    updateGuiEnableStates();
    updateTablePalette();
}

void EvbMainWindow::updateClientStatus(const TcpClientStatus &status)
{
    pendingClientState[status.index] = status;
}

void EvbMainWindow::setup_clients()
{
    ClientIndexSet toDel;
    for(ClientIndex index : clients.keys()) {
        if(!config.clients.contains(index))
            toDel << index;
    }
    deleteClients(toDel);

    for(ClientIndex index : config.clients.keys()) {
        if(!clients.contains(index))
            addClient(config.clients[index]);
    }
    updateTableSize();

    updateAnalizerList();
}

void EvbMainWindow::deleteClients(const ClientIndexSet &set)
{
    for(ClientIndex index : set) {
        delete clients.take(index);
        int row = getTableRowByClientIndex(index);
        QWidget *chButton = ui->tableWidget->cellWidget(row, COL_READOUT_CB);
        QWidget *stateLabels = ui->tableWidget->cellWidget(row, COL_CONNECTION_STATE);
        Q_ASSERT(chButton);
        Q_ASSERT(stateLabels);
        delete chButton;
        delete stateLabels;
        ui->tableWidget->removeRow(getTableRowByClientIndex(index));
        lastClientState.remove(index);
    }
}

bool EvbMainWindow::addClient(const EvBClientConfig &client_config)
{
    Q_ASSERT(getFsmState()==FsmState::Init ||
             getFsmState()==FsmState::Ready ||
             getFsmState()==FsmState::Error);

    ClientIndex index = client_config.index;
    if (index.isEmpty())
        return false;
    if (clients.contains(index))
        return false;

    Client *clientManager = new Client(client_config, config.evNumCfg.getEvNumIndex(), this);
    makeConnections(clientManager);
    clients[index] = clientManager;
    addTableRow(client_config);
    emit updateClientReadout(index, client_config.enabled);
    updateTableSize();
    qDebug() << QString("Added client %1").arg(index.toString());
    return true;
}

void EvbMainWindow::makeConnections(Client *cm) const
{
    connect(this, &EvbMainWindow::requestState, cm, &Client::setClientRequestState);
    connect(cm, &Client::clientStateUpdated, client_manager_fsm, &ClientManagerFsm::setClientState);
    connect(this, &EvbMainWindow::destroyClient, cm, &Client::destroyClient);
    connect(this, &EvbMainWindow::destroyAll, cm, &Client::deleteLater);
    connect(this, &EvbMainWindow::updateClientReadout, cm, &Client::updateClientReadout);

    connect(cm, &Client::statUpdated, statisticDialog, &StatisticDialog::analyzerStatUpdated);

    connect(cm, &Client::evSizePeriodUpdated, adjustBufferModel, &AdjustBufferModel::clientEvSizePeriodUpdated);
    connect(this, &EvbMainWindow::printFsmInput, cm, &Client::reqPrintFsmInput);
}

void EvbMainWindow::addTableRow(const EvBClientConfig &cc) const
{
    QTableWidget *table = ui->tableWidget;
    int row = table->rowCount();
    table->setRowCount(row+1);
    QCheckBox *cb = new QCheckBox(table);
    cb->setChecked(cc.enabled);
    connect(cb, &QCheckBox::toggled, this, &EvbMainWindow::setClientReadOut);
    QLabel *label = new QLabel(table);
    label->setAlignment(Qt::AlignCenter);

    QTableWidgetItem *item = nullptr;
    table->setCellWidget(row, COL_READOUT_CB, cb);
    table->setCellWidget(row, COL_CONNECTION_STATE, label);
    table->setItem(row, COL_PROG_TYPE, new QTableWidgetItem(cc.index.program_type.toString()));
    table->setItem(row, COL_PROG_INDEX, item = new QTableWidgetItem(cc.index.program_index.toString()));
    item->setData(Qt::UserRole, QVariant::fromValue(cc.index));
    table->setItem(row, COL_DATA_TYPE, new QTableWidgetItem());
    table->setItem(row, COL_PROG_HOST, new QTableWidgetItem(""));
    table->setItem(row, COL_STATUS, new QTableWidgetItem("creating..."));
    table->setItem(row, COL_EV_NUM, new QTableWidgetItem());
    table->setItem(row, COL_TRASHED_EVENTS, new QTableWidgetItem());
    table->setItem(row, COL_ALL_EVENTS, new QTableWidgetItem());
    table->setItem(row, COL_FLUSHED_SIZE, new QTableWidgetItem());
    table->setItem(row, COL_TIME_DISTR, new QTableWidgetItem());
    table->setItem(row, COL_TCP_BUFF, new QTableWidgetItem());
    table->setItem(row, COL_EV_BUFF, new QTableWidgetItem());

    table->item(row, COL_TIME_DISTR)->setToolTip("Aver. time diff for earliest/\nAver. time full, ms");
    table->item(row, COL_TCP_BUFF)->setToolTip("free/data/trash");
    table->item(row, COL_EV_BUFF)->setToolTip("free/prepared/flushed");
}

void EvbMainWindow::updateGui() const
{
    ui->actionAdvanced_counters->setChecked(config.advancedCounters);
    ui->actionAllow_Remote_Control->setChecked(config.isRemoteEnabled);
    ui->actionEnable_client_time_distribution->setChecked(config.combinerCfg.timeDistrEn);
    ui->actionEnable_inf_output->setChecked(config.combinerCfg.infOutputEn);

    updateGuiCombinerOut();

    ui->actionFix_TLV_format->setChecked(config.combinerCfg.fixTlvFormat);
    ui->actionProtect_data_file_while_writing->setChecked(config.writerCfg.protectFile);
    ui->labelEvNumIndex->setVisible(config.evNumCfg.step != 1);
    ui->labelEvNumIndex->setText(QString("EvNum index: %1").arg(config.evNumCfg.getEvNumIndex()));
    updateSplitGui();
}

void EvbMainWindow::updateGuiCombinerOut() const
{
    ui->actionOutToFile->setChecked(config.isOutFile());
    ui->actionOutToTcp->setChecked(config.isOutTcp());
    ui->actionCombinerIdle->setChecked(config.isOutIdle());

    ui->radioButtonOutFile->setChecked(config.isOutFile());
    ui->radioButtonOutTcp->setChecked(config.isOutTcp());
    ui->radioButtonOutIdle->setChecked(config.isOutIdle());
}

void EvbMainWindow::updateSplitGui() const
{
    const auto &cfg = config.fileSplit;
    ui->groupBoxFileSplit->setVisible(config.isOutFile() && cfg.hasAnyEn());
    auto label = ui->labelSplitEv;
    label->setVisible(cfg.evNumEn);
    label->setText(QString("EvNum:\n%L1").arg(cfg.evNumFileSplit));
    label = ui->labelSplitSize;
    label->setVisible(cfg.fileSizeEn);
    label->setText(QString("File size: %1").arg(humanReadableSize(cfg.fileSizeSplit)));
    label = ui->labelSplitTime;
    label->setVisible(cfg.timeEn);
    label->setText(QString("Time: %L1s").arg(cfg.timeFileSplit));
}

void EvbMainWindow::updateGuiEnableStates() const
{
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    bool isStopped = (getFsmState()==FsmState::Ready) || (getFsmState()== FsmState::Error);
    if (ranges.empty() || !isStopped) {
        ui->pushButtonRemove->setEnabled(false);
    } else {
        ui->pushButtonRemove->setEnabled(true);
    }

    updateButtonAddEnableStates();
    for(int r=0; r<ui->tableWidget->rowCount(); ++r) {
        auto w = qobject_cast<QWidget *>(ui->tableWidget->cellWidget(r, COL_READOUT_CB));
        if(w)
            w->setEnabled(isStopped);
    }
    ui->actionSet_EvB_index->setEnabled(isStopped);
    ui->actionOutToFile->setEnabled(isStopped);
    ui->actionOutToTcp->setEnabled(isStopped);
    ui->groupBoxOutput->setEnabled(isStopped);
    ui->actionCombinerIdle->setEnabled(isStopped);
    ui->actionChange_storage_dir->setEnabled(isStopped);
    ui->actionReset_to_defaults->setEnabled(isStopped);
    ui->actionFix_TLV_format->setEnabled(isStopped);
}

void EvbMainWindow::updateButtonAddEnableStates() const
{
    bool isStopped = (getFsmState()==FsmState::Ready) || (getFsmState()== FsmState::Error);
    ui->pushButtonAdd->setEnabled(isStopped && hasFreeProgs());
}

void EvbMainWindow::updateTableSize()
{
    int width = ui->tableWidget->verticalHeader()->width();
    for(int i=0;i<ui->tableWidget->columnCount();++i){
        if(!ui->tableWidget->isColumnHidden(i))
            width += ui->tableWidget->columnWidth(i);
    }
    width += 5;

    QScreen *screen = QGuiApplication::primaryScreen();
    QRect rec = screen->geometry();
    int s_width = rec.width();

    if(width < (s_width*0.66))
        ui->tableWidget->setMinimumWidth(width);
}

void EvbMainWindow::updateAnalizerList()
{
    if(evb_fsm->getState() != FsmState::Error && evb_fsm->getState() != FsmState::Ready) {
        qCritical("Trying to '%s' while FSM not in Ready or Error state", __func__);
        evb_fsm->setRequestState(FsmState::Error);
    }
    QVector<Analizer *> analizers;
    ClientIndexSet clients_;
    nonEvBasedClients.clear();
    for(auto cfg : config.clients) {
        if(!cfg.enabled)
            continue;
        Q_ASSERT(clients.contains(cfg.index));
        if(clients.contains(cfg.index))
            analizers.push_back(clients[cfg.index]->getAnalizer());
        clients_.insert(cfg.index);
        if(isNonEventBaseClient(cfg.index))
            nonEvBasedClients << cfg.index;
    }

    enabledClientNum = static_cast<uint>(analizers.size());
    emit analizerListUpdated(clients_, nonEvBasedClients, analizers);
    updateRemoteStatus();
    updateClientDataType();
    refreshProgramDescription();
}

void EvbMainWindow::updateRemoteStatus()
{
    QString stateStr;
    int clientsEnabled=0;
    for(auto cfg : config.clients){
        if(cfg.enabled)
            ++clientsEnabled;
    }
    if(clientsEnabled)
        stateStr = QString::number(clientsEnabled).append(" clients");
    else
        stateStr = "no clients";
    emit sendRemoteState(stateStr);
}

void EvbMainWindow::initProgramDescription(const ConfigIndexName& conf)
{
    progDesc.setClientIndex({DaqModule(DaqModule::Type::EventBuilder),
                             conf.getIndex()});

    progDesc.ver_date = getAppDateStr();
    progDesc.ver_hash = getAppVersion();
}

void EvbMainWindow::refreshProgramDescription()
{
    bool hasRcInterface = progDesc.getProgramInterfaceNum(ProgramInterfaceRemoteControl);
    ProgramInterface rcInterface;
    if(hasRcInterface)
        rcInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    progDesc.clear();
    if(hasRcInterface)
        progDesc.interfaces.push_back(rcInterface);

    for (int i=0; i<lastCombinerStats.tcpStatus.size(); ++i) {
        const OutputTcpStatus &tcpStatus = lastCombinerStats.tcpStatus[i];
        if(!tcpStatus.listening)
            continue;
        ProgramInterface tcpInterface;
        tcpInterface.setType(i==0 ? ProgramInterfaceMonitorData : ProgramInterfaceDataFlow);
        tcpInterface.enabled = true;
        tcpInterface.id = tcpStatus.index ? tcpStatus.index-1 : 0;
        tcpInterface.port = *tcpStatus.localPort;
        tcpInterface.isFree = !tcpStatus.connected;
        if(tcpStatus.connected) {
            tcpInterface.peers << ProgramInterfacePeer(tcpStatus.peer.host, tcpStatus.peer.port);
        }
        progDesc.interfaces.push_back(tcpInterface);
    }

    int clientCnt=0;
    bool nonEvBasedMode = true;
    for (auto cfg : config.clients) {
        if(cfg.enabled) {
            ++clientCnt;
            if(nonEvBasedMode)
                nonEvBasedMode = isNonEventBaseClient(cfg.index);
        }
    }
    if(clientCnt)
        progDesc.options["Clients"] = QString::number(clientCnt);
    if(nonEvBasedMode)
        progDesc.options[EVB_DATA_TYPE_KEY] = EVB_DATA_TYPE_NON_EVENT;

#ifdef GIT_DESCR
    progDesc.ver_hash = GIT_DESCR;
#endif
#ifdef GIT_LAST_COMMIT
    progDesc.ver_date = GIT_LAST_COMMIT;
#endif
    progDesc.options["fsm"] = fsmStateToString(getFsmState());
    progDesc.options["runIndex"] = loc_runId.runIndex;
    progDesc.options["runNumber"] = QString::number(loc_runId.runNumber);
    progDesc.options["output"] = config.getOutputStr();

    emit updateProgramDescription(progDesc);
}

void EvbMainWindow::updateTcpPortLabels()
{
    for (const auto &st : lastCombinerStats.tcpStatus) {
        QLabel *l = getTcpLabel(st.index);
        Q_ASSERT(l);

        QString text = QString("%1: ").arg(getTcpName(st.index));
        if(st.listening) {
            text += QString("%1").arg(*st.localPort);
            if(st.connected) {
                text += "->" + st.peer.toString();
            } else {
                text += " (free)";
            }
        } else {
            text += QString("disabled");
        }
        text += QString(" (%1%)").arg((int)(st.dataPart));
        l->setText(text);

        l->setStyleSheet(QString("QLabel { color : %1; }")
                             .arg(!st.connected ? (st.index ? "red" : "black") : "green"));
    }
}

void EvbMainWindow::setOutTcpServerNumber(int tcpNum)
{
    if(tcpNum < 1)
        tcpNum = 1;
    if(config.combinerCfg.combinerOutTcpNum != tcpNum){
        config.combinerCfg.combinerOutTcpNum = tcpNum;
        emit configChanged(config);
    }

    updateOutTcpLabelList();

    emit configChanged(config);
}

QVector<ProgramDescription> EvbMainWindow::getFreeProgs() const
{
    QVector<ProgramDescription> freeProgs;
    for (const ProgramDescription &descr : progDescrsCache) {
        if(isFreeProg(descr))
            freeProgs.push_back(descr);
    }
    return freeProgs;
}

bool EvbMainWindow::hasFreeProgs() const
{
    for (const ProgramDescription &descr : progDescrsCache) {
        if(isFreeProg(descr))
            return true;
    }
    return false;
}

bool EvbMainWindow::isFreeProg(const ProgramDescription &pd) const
{
    if(clients.contains(pd.getClientIndex()))
        return false;

    int interface_id = config.evNumCfg.getEvNumIndex();
    for (const ProgramInterface &i : pd.interfaces) {
        if(i.type == ProgramInterface::getTypeName(ProgramInterfaceDataFlow)
                && i.enabled && i.isFree && i.id == interface_id) {
            return true;
        }
    }

    return false;
}

int EvbMainWindow::getTableRowByClientIndex(const ClientIndex &index) const
{
    QTableWidget *table = ui->tableWidget;
    for(int row =0; row < table->rowCount(); ++row) {
        if(index == getClientIndexByRow(row))
            return row;
    }
    return -1;
}

QSet<int> EvbMainWindow::getSelectedRow() const
{
    QSet<int> set;
    QList<QTableWidgetSelectionRange> ranges = ui->tableWidget->selectedRanges();
    if(!ranges.isEmpty()){
        for(auto range : ranges) {
            for(int row=range.topRow(); row<=range.bottomRow(); ++row)
                set.insert(row);
        }
    }
    return set;
}

ClientIndex EvbMainWindow::getClientIndexByRow(int row) const
{
    const auto &item = ui->tableWidget->item(row, COL_PROG_INDEX);
    return item ? item->data(Qt::UserRole).value<ClientIndex>() : ClientIndex();
}

ClientIndexSet EvbMainWindow::getSelectedClients() const
{
    ClientIndexSet cis;
    auto rows = getSelectedRow();
    for(auto r : rows) {
        auto id = getClientIndexByRow(r);
        if(!id.isEmpty())
            cis.insert(id);
    }
    return cis;
}

void EvbMainWindow::updateTableTextColor(int row, ClientIndex index) const
{
    if(index.isEmpty())
        return;

    auto color = getTcpClientStateColor(lastClientState.value(index), false);
    QTableWidgetItem *i = ui->tableWidget->item(row, COL_STATUS);
    if(i->foreground().color() != color)
        i->setForeground(color);
}

void EvbMainWindow::combinerOutUpdated()
{
    if(config.isOutTcp() && config.combinerCfg.combinerOutTcpNum==0){
        setOutTcpServerNumber(1);
        emit configChanged(config);
    }

    ui->actionSet_number_of_outputTcpServers->setEnabled(config.isOutTcp());
    ui->labelDataFileName->setVisible(config.isOutFile());
    ui->labelCombinerIdle->setVisible(config.isOutIdle());
    updateOutTcpLabelList();
    updateTcpLabelVisibility();
    emit configChanged(config);
}

QLabel *EvbMainWindow::getTcpLabel(int index) const
{
    bool ok;
    for(int i=0; i<ui->verticalLayoutOutTcp->count(); ++i) {
        QLayoutItem *item = ui->verticalLayoutOutTcp->itemAt(i);
        if(item) {
            QWidget *w = item->widget();
            if(w) {
                QLabel *l = qobject_cast<QLabel *>(w);
                if(l) {
                    if(l->property(prop_id).toInt(&ok)==index && ok)
                        return l;
                }
            }
        }
    }
    return nullptr;
}

void EvbMainWindow::updateTcpLabelVisibility() const
{
    for(int id=1; id<=config.combinerCfg.combinerOutTcpNum; ++id) {
        auto l = getTcpLabel(id);
        Q_ASSERT(l);
        if(l)
            l->setVisible(config.isOutTcp());
    }
}

void EvbMainWindow::registerTypes()
{
    qRegisterMetaType<Qt::GlobalColor>();
    qRegisterMetaType<AnalyzerStat>();
    qRegisterMetaType<RcRunId>();
    qRegisterMetaType<ClientIndex>();
    qRegisterMetaType<ClientIndexEnabledMap>();
    qRegisterMetaType<ClientIndexSet>();
    qRegisterMetaType<EvbFileSplitterConfig>();
    qRegisterMetaType<EvbCombinerConfig>();
    qRegisterMetaType<EvbEvNumberConfig>();
    qRegisterMetaType<EventSizePeriod>();
    AbstractConfigConverter<ClientIndex>::registerTypes();
    AbstractConfigConverter<RunDocument>::registerTypes();
}

void EvbMainWindow::initCombiner()
{
    combiner->initOutputBuffers(config.combinerCfg);

    combiner->moveToThread(combinerThread);
    combinerThread->start();

    configDispatcher->subscribeCombinerConfig(combiner);
    configDispatcher->subscribeEvNumberConfig(combiner);

    connect(this, &EvbMainWindow::requestState, combiner, &Combiner::setRequestState);
    connect(this, &EvbMainWindow::destroyClient, combiner, &Combiner::destroyClient);
    connect(this, &EvbMainWindow::analizerListUpdated, combiner, &Combiner::updateAnalizerList);

//    connect(combiner, &Combiner::clientStateUpdated, fsm, &EventBuilderFsm::setClientState);
    connect(combiner, &Combiner::error, this, &EvbMainWindow::combinerError);

    connect(combiner, &Combiner::destroyed, combinerThread, &QThread::quit);
    connect(combiner, &Combiner::status, this, &EvbMainWindow::updateStatus);
    connect(combiner, &Combiner::statusRaw, this, &EvbMainWindow::updateCombinerStats);
    connect(combiner, &Combiner::fileWriterStatsUpdated, this, &EvbMainWindow::fileWriterStatsChanged);
    connect(this, &EvbMainWindow::setMonitorPortNum, combiner, &Combiner::setMonitorPortNum);
}

void EvbMainWindow::applyConfig()
{
    updateGui();

    //TODO: implement check that all parts are corresponding to Config
    emit allowRemoteControl(config.isRemoteEnabled);
    setup_clients();
    update_combiner_to_config();
}

void EvbMainWindow::update_combiner_to_config()
{
    updateAnalizerList();
    combinerOutUpdated();
    setOutTcpServerNumber(config.combinerCfg.combinerOutTcpNum);

    emit evNumberConfigChanged(config.evNumCfg);
    if(config.monitorPort)
        emit setMonitorPortNum(config.monitorPort);
}

void EvbMainWindow::updateOutTcpLabelList()
{
    QLayout *layout = ui->verticalLayoutOutTcp;

    const auto &tcpNum = config.combinerCfg.combinerOutTcpNum;
    for(int index=0; index < layout->count(); ) {
        bool ok;
        QWidget *w = layout->itemAt(index)->widget();
        if(w->property(prop_id).toInt(&ok) > tcpNum && ok) {
            layout->removeWidget(w);
            delete w;
        } else {
            ++index;
        }
    }

    for(int tcpId=0; tcpId <= tcpNum; ++tcpId) {
        if(getTcpLabel(tcpId)) continue;
        QLabel *label = new QLabel(QString("%1: creating...").arg(getTcpName(tcpId)), ui->widgetOutTcp);
//        if(tcpId == 0) label->setVisible(false);
        label->setProperty(prop_id, tcpId);
        layout->addWidget(label);
        label->setAlignment(Qt::AlignHCenter);
        if(tcpId)
            label->setVisible(config.isOutTcp());
    }
}

void EvbMainWindow::combinerError(const QString &errStr)
{
    qCritical() << "Got error from combiner" << errStr;
    if(isInRun())
        emit sendRunError(loc_runId, errStr);
}

void EvbMainWindow::deviceError(const ClientIndex &index, const QString &errStr)
{
    QString msg = QString("Got error from %1").arg(index.toString());
    qCritical() << msg << errStr;
    if(isInRun())
        emit sendRunError(loc_runId, errStr);
}

void EvbMainWindow::updateStatus(const QString &status)
{
    labelCombinerStatus->setText(status);
    emit sendRemoteState(status);
}

void EvbMainWindow::updateCombinerStats(const CombinerStats &ss)
{
    if(backPreasure) {
        if(ss.getMaxOutputDataPart() < 0.5) {
            backPreasure = false;
            qInfo() << "BackPressure disabled";
            emit sendBackPreasure(loc_runId, false);
        }
    } else {
        if(ss.getMaxOutputDataPart() > 0.9) {
            backPreasure = true;
            QString mess = "Enable backPressure";
            qWarning() << mess;
            emit sendRunWarning(loc_runId, mess);
            emit sendBackPreasure(loc_runId, true);
        }
    }
    bool outDataInterfaceChanged = false;
    bool outDataPartsChanged = false;
    for (int id : ss.tcpStatus.keys()) {
        if(!lastCombinerStats.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
        const OutputTcpStatus &prevS = lastCombinerStats.tcpStatus[id];
        const OutputTcpStatus &nextS = ss.tcpStatus[id];
        outDataInterfaceChanged |= prevS.listening!=nextS.listening
                || prevS.connected!=nextS.connected
                || prevS.localPort!=nextS.localPort
                || prevS.peer.hostName!=nextS.peer.hostName;
        outDataPartsChanged |= !qFuzzyCompare(prevS.dataPart, nextS.dataPart);
    }
    for (int id : lastCombinerStats.tcpStatus.keys()) {
        if(!ss.tcpStatus.contains(id)) {
            outDataInterfaceChanged = true;
            outDataPartsChanged = true;
            break;
        }
    }
    if(ss.partedEvents && lastCombinerStats.partedEvents != ss.partedEvents) {
        static QElapsedTimer partedEventsTimer;
        if(!partedEventsTimer.isValid() || partedEventsTimer.elapsed() > 1500) {
            partedEventsTimer.start();
            auto mess = QString("Incompleted events: %L1").arg(ss.partedEvents);
            emit sendRunWarning(loc_runId, mess);
        }
    }
    const auto prevClientsStat = lastCombinerStats.clientsStat; // todo: use lastCombinerStats instead
    lastCombinerStats = ss; // todo: move updating lastCombinerStats after table update
//    fsm->setExpectedEvNum(ss.expectedEvNum);
    QTableWidget * t = ui->tableWidget;
    const bool tableUpdatesEnabled = t->updatesEnabled();
    t->setUpdatesEnabled(false);
    if(outDataInterfaceChanged)
        refreshProgramDescription();
    if(outDataInterfaceChanged || outDataPartsChanged)
        updateTcpPortLabels();

    if(ss.clientsStat.isEmpty()) {
        t->setUpdatesEnabled(tableUpdatesEnabled);
        return;
    }
    for(int row=0; row<t->rowCount(); ++row){
        ClientIndex index = getClientIndexByRow(row);
        if(index.isEmpty())
            continue;
        auto clSt = ss.clientsStat[index];
        QString str;
        if(clSt.lastEventNum.has_value())
            str = QString("%L1").arg(clSt.lastEventNum.value());
        t->item(row, COL_EV_NUM)->setText(str);
        auto item = t->item(row, COL_TRASHED_EVENTS);
        item->setText(QString("%L1").arg(clSt.trashedEvents));
        QColor color;
        if(clSt.trashedEvents) {
            color = Qt::red;
            if(prevClientsStat[index].trashedEvents != clSt.trashedEvents) {
                static QElapsedTimer trashedEventsTimer;
                if(!trashedEventsTimer.isValid() || trashedEventsTimer.elapsed() > 1500) {
                    trashedEventsTimer.start();
                    auto mess = QString("Trashed events from %1: %L2")
                            .arg(index.toString()).arg(ss.partedEvents);
                    emit sendRunWarning(loc_runId, mess);
                }
            }
        }
        if(item->foreground().color() != color)
            item->setForeground(color);
        t->item(row, COL_ALL_EVENTS)->setText(QString("%L1").arg(clSt.eventBlocks));
        t->item(row, COL_FLUSHED_SIZE)->setText(humanReadableSize(clSt.flushedSize));
        t->item(row, COL_FLUSHED_SIZE)->setToolTip(QString::number(clSt.flushedSize));
        if(config.combinerCfg.timeDistrEn) {
            if(ss.timeDistrCnt && clSt.timeDistr.has_value())
                t->item(row, COL_TIME_DISTR)->
                        setText(QString::number(clSt.timeDistr.value()/ss.timeDistrCnt));
            else
                t->item(row, COL_TIME_DISTR)->setText("-");
        }

        t->item(row, COL_EV_BUFF)->setText(QString("%1/%2/%3")
                                               .arg(clSt.freeEvPart)
                                               .arg(clSt.preparedEvPart)
                                               .arg(clSt.flushedEvPart));

        // metrics
        const QMap<QString, QString> tags({{"client", index.toString()}});
        MetricFields fields{{"drop", clSt.trashedEvents},
                            {"pass_event", clSt.eventBlocks},
                            {"pass_json", clSt.jsonBlocks},
                            {"pass_stat", clSt.statBlocks}};
        MetricSender::send("evb.in.subevents", fields, tags);
    }

    {
        QString mess;
//        if(!enabled)
//            mess.append("stopped; ");
        QString clientsStr;
        int first=-1;
        int last=-1;
        for(int r=0; r<ui->tableWidget->rowCount(); ++r){
            const auto &i = getClientIndexByRow(r);
            if(!ss.lastEventClients.contains(i))
                continue;
            if(first==-1) {
                first = r;
            } else if(last+1!=r) {
                if(first == last)
                    clientsStr.append(QString("%1 ").arg(first));
                else
                    clientsStr.append(QString("%1-%2 ").arg(first).arg(last));
                first = r;
            }
            last = r;
        }

        if(first!=-1){
            if(first == last)
                clientsStr.append(QString("%1 ").arg(first));
            else
                clientsStr.append(QString("%1-%2 ").arg(first).arg(last));
        }

        mess.append(QString("reg:%1;skip:%2;eob:%3;clients:%4;nextEv:%5  monitor:%6")
                    .arg(ss.regularEvents)
                    .arg(ss.skipped)
                    .arg(ss.eob)
                    .arg(clientsStr)
                    .arg(ss.expectedEvNum)
                    .arg(ss.monitorEvCount)
                    );
//        for(int i=0;i<events.size(); ++i){
//            const EventMetaInfo* ev = events[i];
//            if(ev == nullptr) continue;
//            mess.append(QString("; #%1=%2").arg(i).arg(events[i]->evNum));
//        }
        updateStatus(mess);
    }

    labelFileErrSize->hide();
    labelFileErrSize->setText(QString("fileErr:%1B").arg(ss.missedFileSize));
    if(ss.missedFileSize){
        labelFileErrSize->show();
        labelFileErrSize->setStyleSheet("QLabel { color : red;}");
    }
    labelFileErrSize->setStyleSheet(QString("QLabel { color : %1;}")
                                    .arg(ss.missedFileSize?"red" : "black"));
    updateTableSize();

    t->setUpdatesEnabled(tableUpdatesEnabled);
}

void EvbMainWindow::fileWriterStatsChanged(const FileWriterStats &fwStat)
{
    QString dataFileNameText;
    if(!fwStat.fileName.isEmpty()){
        QString suf;
        if(fwStat.writtenFileSize != fwStat.writtenTotalSize)
            suf = humanReadableSize(fwStat.writtenTotalSize).prepend('/');
        dataFileNameText = QString("%1 (%2%3) (%4%)")
                .arg(fwStat.fileName)
                .arg(humanReadableSize(fwStat.writtenFileSize))
                .arg(suf).arg(fwStat.dataPart*100, 0, 'f', 1);

        if(fwStat.missedFileSize)
            dataFileNameText +=QString(" (missed: %1)")
                    .arg(humanReadableSize(fwStat.missedFileSize));
    } else {
        dataFileNameText = QString(fwStat.predefStorage).append("...");
    }
    ui->labelDataFileName->setText(dataFileNameText);

    auto p = ui->labelDataFileName->palette();
    p.setColor(ui->labelDataFileName->foregroundRole(), fwStat.dataPart>0.8 ? Qt::darkYellow : Qt::black);
    ui->labelDataFileName->setPalette(p);
}

void EvbMainWindow::remoteStartRequest(const RcRunId &rcRunId, const QVariant &startEvNumber)
{
    auto msg = QString("Got remote request to start %1").arg(rcRunId.toString());
    if(startEvNumber.isValid()) {
        startEvNum = startEvNumber.toULongLong();
        msg += QString(" (StartEvNum=%L1)").arg(*startEvNum);
    } else {
        startEvNum.reset();
    }

    qInfo() << msg;
    emit configChanged(config);

    static bool warn_start_event_number = true;
    if(startEvNum && warn_start_event_number) {
        qWarning() << "Remote start with start_event_number not implemented yet."
                   << "startEvNumber:" << *startEvNum;
        warn_start_event_number = false;
    }

    if(loc_runId.runIndex == rcRunId.runIndex &&
            loc_runId.runNumber > rcRunId.runNumber) {
        const auto &msg = QString("Previous runNumber(%1) is greater that remote control runNumber(%2)")
                              .arg(loc_runId.runNumber)
                              .arg(rcRunId.runNumber);
        qWarning() << msg;
        emit sendRunWarning(loc_runId, msg);
        return;
    }

    if(getFsmState() != FsmState::Ready) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Ready))
                              .arg(fsmStateToString(getFsmState()));
        qWarning() << msg;
        emit sendRunWarning(rcRunId, msg);
        return;
    }

    loc_runId = rcRunId;
    evb_fsm->setRequestState(FsmState::Starting);
}

void EvbMainWindow::remoteStopRequest(const RcRunId &rcRunId, const QVariant &runStopEvNum)
{
    auto msg = QString("Got remote request to stop %1").arg(rcRunId.toString());
    if(runStopEvNum.isValid()) {
        stopEvNum = runStopEvNum.toULongLong();
        msg += QString(" (StopEvNum=%L1)").arg(*stopEvNum);
    } else {
        stopEvNum.reset();
    }
    qInfo() << msg;

    if(loc_runId.runIndex != rcRunId.runIndex) {
        qWarning()<< QString("Got stop request with wrong run index: runIndex (%1); remote control runIndex (%2)")
                     .arg(loc_runId.runIndex)
                     .arg(rcRunId.runIndex);
        return;
    }
    if(loc_runId.runNumber != rcRunId.runNumber) {
        qWarning()<< QString("Got stop of unknown run: runNumber (%1); remote control runNumber (%2)")
                     .arg(loc_runId.runNumber)
                     .arg(rcRunId.runNumber);
        return;
    }

    if(getFsmState() != FsmState::Run) {
        const auto &msg = QString("Application isn't in '%1' state (curState='%2')")
                              .arg(fsmStateToString(FsmState::Run))
                              .arg(fsmStateToString(getFsmState()));
        qWarning() << msg;
        emit sendRunWarning(loc_runId, msg);
        if(getFsmState() != FsmState::Ready && getFsmState() != FsmState::Error)
            evb_fsm->setRequestState(FsmState::Error);
        return;
    }

    evb_fsm->setLookForStopEvNum(stopEvNum.has_value());
    evb_fsm->setRequestState(FsmState::Stopping);
}

void EvbMainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &rcSerStat)
{
    bool pdChanged = false;
    if(!progDesc.containsProgramInterface(ProgramInterfaceRemoteControl,0)){
        ProgramInterface remoteControlInterface;
        remoteControlInterface.setType(ProgramInterfaceRemoteControl);
        remoteControlInterface.id = 0;
        progDesc.interfaces.push_back(remoteControlInterface);
        pdChanged = true;
    }

    ProgramInterface &remoteInterface = progDesc.getProgramInterface(ProgramInterfaceRemoteControl);
    pdChanged |= remoteInterface.port != rcSerStat.tcpPort ||
            remoteInterface.enabled != rcSerStat.listening ||
            remoteInterface.isFree != !rcSerStat.portInUse;
    remoteInterface.port = rcSerStat.tcpPort;
    remoteInterface.enabled = rcSerStat.listening;
    remoteInterface.isFree = !rcSerStat.portInUse;

    if(rcSerStat.portInUse) {
        ProgramInterfacePeer p{rcSerStat.peerHost, rcSerStat.peerPort};
        if(!remoteInterface.peers.contains(p)) {
            remoteInterface.setPeer(p);
            pdChanged = true;
        }
        updateRemoteStatus();
        qInfo() << "Remote control connected.";
    } else {
        qWarning() << "Remote control disconnected.";
        loc_runId = RcRunId();

        if(getFsmState() != FsmState::Ready) {
            qCritical() << "Stopping current run";
            stopEvNum.reset();
            evb_fsm->setLookForStopEvNum(stopEvNum.has_value());
            evb_fsm->setRequestState(getFsmState() == FsmState::Run
                         ? FsmState::Stopping : FsmState::Error);
        }
    }

    if(ui->widgetRunControl->isEnabled() == rcSerStat.portInUse){
        ui->widgetRunControl->setEnabled(!rcSerStat.portInUse);
    }
    if(pdChanged)
        refreshProgramDescription();
}

/*
void MainWindow::addLogHistory(QString log, Qt::GlobalColor color)
{
//    qDebug()<<QString("[%1] %2").arg(QDateTime::currentDateTime().toString("yy-MM-dd hh:mm:ss")).arg(log);
    int repeat=0;
    int now = time(nullptr);
    bool found=false;
    for(int i=0;i<logHistory.size();++i) {
        LogHistorEntry &e = logHistory[i];
        if(log == e.mess){
            found=true;
            repeat = e.repeatCnt++;
            if(now - e.time > 5){
                e.time = now;
                logHistory.move(i,logHistory.size()-1);
            } else
                return;
        }
    }
    if(!found){
        if(logHistory.size() >= (3*clients.size()+3))
            logHistory.removeAt(0);
        LogHistorEntry e;
        e.mess = log;
        e.repeatCnt = 0;
        e.time = now;
        logHistory.append(e);
    }

    QTextCharFormat tf = ui->plainTextEdit->currentCharFormat();
    tf.setForeground(QBrush(color));

    SyslogMessageSender::log(color==Qt::red ? LOG_PRI_WARNING : LOG_PRI_NOTICE, log.toStdString().c_str());
    if(repeat)
        log.append(QString(" (repeat: %1 times)").arg(repeat));
    ui->plainTextEdit->setCurrentCharFormat(tf);
    ui->plainTextEdit->appendPlainText(QString("[%1] %2").arg(QTime::currentTime().toString(), log));
}
*/

void EvbMainWindow::on_actionChange_storage_dir_triggered()
{
//    QFileDialog::QFileDialog ( QWidget * parent = 0, const QString & caption = QString(), const QString & directory = QString(), const QString & filter = QString() )
//    QFileDialog dialog(this, "Set predefined storage for mpd data file", "", "*.data");
//    dialog.setOption(QFileDialog::ShowDirsOnly);
//    dialog.exec();
//    qDebug() << dialog.selectedFiles();

    auto &storageDir = config.writerCfg.storageDir;
    QString dir = QFileDialog::getExistingDirectory( this, "Choose storage directory...",
                                                     storageDir, QFileDialog::ShowDirsOnly );
    if(!dir.isEmpty() && storageDir != dir){
        storageDir = dir;
        emit writerConfigChanged(config.writerCfg);
    }
}

void EvbMainWindow::on_actionReset_to_defaults_triggered()
{
    emit configChanged({});
}

void EvbMainWindow::on_actionAdvanced_counters_toggled(bool showColumns)
{
    if(config.advancedCounters != showColumns) {
        config.advancedCounters = showColumns;
        emit configChanged(config);
    }

    QTableWidget *table = ui->tableWidget;
//    table->setColumnHidden(COL_EV_NUM, !showColumns);
//    table->setColumnHidden(COL_ID, !showColumns);
    table->setColumnHidden(COL_TCP_BUFF, !showColumns);
    table->setColumnHidden(COL_EV_BUFF, !showColumns);
    updateTableSize();
}

void EvbMainWindow::on_actionAllow_Remote_Control_toggled(bool checked)
{
    if(config.isRemoteEnabled != checked) {
        config.isRemoteEnabled = checked;
        emit configChanged(config);
    }
    emit allowRemoteControl(config.isRemoteEnabled);
}

void EvbMainWindow::data_out_changed(QAction *a)
{
    auto &out = config.combinerCfg.combinerOut;
    if(a == ui->actionCombinerIdle) out = COMBINER_IDLE;
    else if(a == ui->actionOutToTcp) out = COMBINER_TCP;
    else if(a == ui->actionOutToFile) out = COMBINER_FILE;
    else {
        qCritical() << "Data out was switch to unknown state.";
    }

    combinerOutUpdated();
    emit configChanged(config);
    refreshProgramDescription();
    updateGuiCombinerOut();
    updateSplitGui();
}

void EvbMainWindow::dataOutChangedRB()
{
    auto rb = qobject_cast<QRadioButton*>(sender());
    if(!rb->isChecked())
        return;

    auto out = rb->property(prop_outType).value<COMBINER_OUT>();
    if(out == config.combinerCfg.combinerOut)
        return;
    config.combinerCfg.combinerOut = static_cast<COMBINER_OUT>(out);

    combinerOutUpdated();
    emit configChanged(config);
    refreshProgramDescription();
    updateGuiCombinerOut();
    updateSplitGui();
}

void EvbMainWindow::on_actionSet_number_of_outputTcpServers_triggered()
{
    bool ok;
    const int tcpNum = QInputDialog::getInt(this, tr("TCP servers for output"),
                                 tr("Set number of TCP servers for output:"), config.combinerCfg.combinerOutTcpNum, 1, 20, 1, &ok);
    if (ok && config.combinerCfg.combinerOutTcpNum != tcpNum)
        setOutTcpServerNumber(tcpNum);
}

void EvbMainWindow::on_actionSet_EvB_index_triggered()
{
    bool ok;
    quint32 startNumber = QInputDialog::getInt(this, tr("Event builder index"),
                                 tr("Set 1st event number (EvB index):"), config.evNumCfg.startEvNum, 0, 19, 1, &ok);
    if(!ok)
        startNumber = config.evNumCfg.startEvNum;
    quint32 stepNumber = QInputDialog::getInt(this, tr("EvB event numbet step"),
                                 tr("Set event number step:"), config.evNumCfg.step, 1, 20, 1, &ok);
    if(!ok)
        startNumber = config.evNumCfg.step;

    if (config.evNumCfg.startEvNum != startNumber || config.evNumCfg.step != stepNumber) {
        int evbIndexOld = config.evNumCfg.getEvNumIndex();
        EvbEvNumberConfig cfg{startNumber, stepNumber};
        emit evNumberConfigChanged(cfg);
        if(evbIndexOld != cfg.getEvNumIndex()){
            QMessageBox msgBox;
            msgBox.setText("Event Builder index was changed, configuration was saved.\n"
                           "Needed manual restart\n"
                           "Closing...");
            msgBox.exec();
            close();
        }
    }
}

void EvbMainWindow::on_actionSet_ring_buffer_sizes_triggered()
{
    adjustBufferDialog->show();
}

void EvbMainWindow::on_actionSet_file_split_options_triggered()
{
    fileSplitDialog->show();
}

void EvbMainWindow::on_actionShow_statistics_triggered()
{
    statisticDialog->show();
}

void EvbMainWindow::on_actionEnable_client_time_distribution_toggled(bool en)
{
    if(config.combinerCfg.timeDistrEn != en) {
        config.combinerCfg.timeDistrEn = en;
        emit configChanged(config);
    }

    QTableWidget *table = ui->tableWidget;
    table->setColumnHidden(COL_TIME_DISTR, !config.combinerCfg.timeDistrEn);
    updateTableSize();
}

void EvbMainWindow::on_actionSet_monitor_port_number_triggered()
{
    bool ok;
    const int MIN_PORT_NUMBER = 0;
    const int MAX_PORT_NUMBER = 32999;
    const int portNum = QInputDialog::getInt(this, tr("Set monitor port number"),
                                             QString("Set monitor port number (in range %1-%2):")
                                             .arg(MIN_PORT_NUMBER).arg(MAX_PORT_NUMBER),
                                             config.monitorPort ? config.monitorPort : MAX_PORT_NUMBER, MIN_PORT_NUMBER, MAX_PORT_NUMBER, 1, &ok);
    if (ok && config.monitorPort != portNum) {
        config.monitorPort = portNum;
        emit configChanged(config);
        emit setMonitorPortNum(static_cast<quint16>(portNum));
//        QMessageBox msgBox;
//        msgBox.setText("Monitor port number was changed, configuration was saved.\n"
//                       "Needed manual restart\n"
//                       "Closing...");
//        msgBox.exec();
//        close();
    }
}

void EvbMainWindow::on_actionPrint_FSM_input_triggered()
{
    emit printFsmInput();
}

void EvbMainWindow::on_actionFix_TLV_format_toggled(bool en)
{
    if(config.combinerCfg.fixTlvFormat == en)
        return;
    config.combinerCfg.fixTlvFormat = en;
    emit configChanged(config);
}

void EvbMainWindow::on_actionProtect_data_file_while_writing_toggled(bool protect)
{
    if(config.writerCfg.protectFile == protect)
        return;
    config.writerCfg.protectFile = protect;
    emit writerConfigChanged(config.writerCfg);
    emit configChanged(config);
}

void EvbMainWindow::setClientReadOut(bool readout)
{
    QTableWidget *table = ui->tableWidget;
    for(int row=0; row<table->rowCount(); ++row){
        if(table->cellWidget(row, COL_READOUT_CB)==reinterpret_cast<QWidget *>(sender())){
//            int clientId = table->item(row, COL_ID)->text().toInt(&ok);
//            qInfo() << QString("client at row=%1 clientId=%2 was %3abled")
//                       .arg(row).arg(clientId).arg(readout?"en":"dis");
//            Q_ASSERT(config.clients.contains(clientId));
            ClientIndex client_index = getClientIndexByRow(row);
            Q_ASSERT(!client_index.isEmpty());
            Q_ASSERT(config.clients.contains(client_index));
            if(config.clients[client_index].enabled != readout) {
                config.clients[client_index].enabled = readout;
                emit configChanged(config);
                Q_ASSERT(clients.contains(client_index));
            }
            emit updateClientReadout(client_index, readout);
            updateAnalizerList();
            emit configChanged(config);
            return;
        }
    }
}

void EvbMainWindow::on_actionEnable_inf_output_toggled(bool en)
{
    if(config.combinerCfg.infOutputEn == en)
        return;
    config.combinerCfg.infOutputEn = en;
    emit configChanged(config);
}

void EvbMainWindow::pnpCacheUpdated(const PnpInfoCache &cache)
{
    progDescrsCache = cache;
    hasNewPnpCache = true;
}

void EvbMainWindow::updateTablePalette() const
{
    auto cis = getSelectedClients();
    if(cis.isEmpty())
        return;
    QPalette p = ui->tableWidget->palette();
    QColor color;
    for(auto id : cis) {
        auto c = getTcpClientStateColor(lastClientState.value(id), true);
        if(c == Qt::red) {
            color = c;
            break;
        }
        if(c.isValid())
            color = c;
    }
    if(p.color(QPalette::HighlightedText) != color) {
        p.setColor(QPalette::HighlightedText, color);
        ui->tableWidget->setPalette(p);
    }
}

void EvbMainWindow::updateClientDataType() const
{
    auto t = ui->tableWidget;
    bool hideCol = true;
    for(int r=0; r<t->rowCount(); ++r) {
        auto i = getClientIndexByRow(r);
        bool nonEvBased = isNonEventBaseClient(i);
        hideCol &= !nonEvBased;
        t->item(r, COL_DATA_TYPE)->setText(nonEvBased ? "nonEvBased" : "EvBased");
    }
    t->setColumnHidden(COL_DATA_TYPE, hideCol);
}

bool EvbMainWindow::isNonEventBaseClient(const ClientIndex &index) const
{
    for(auto pd : progDescrsCache) {
        if(pd.getClientIndex() == index)
            return isNonEventBaseProgram(pd);
    }
    return false;
}

FsmState EvbMainWindow::getFsmState() const
{
    return evb_fsm->getState();
}

bool EvbMainWindow::isInRun() const
{
    return getFsmState() != FsmState::Ready && getFsmState() != FsmState::Error && getFsmState() != FsmState::Error;
}

void EvbMainWindow::updatePendingClientStatus()
{
    QTableWidget *table = ui->tableWidget;
    bool restoreUpdate = false;
    for(auto st : pendingClientState) {
        const auto &index = st.index;
        if(!config.clients.contains(index))
            continue;

        bool connectionChanged = !lastClientState.contains(index)
                || lastClientState[index] != st.state;
        lastClientState[index] = st.state;
        if(connectionChanged)
            updateTablePalette();
        const int row = getTableRowByClientIndex(index);
        if(row == -1)
            continue;

        if(!restoreUpdate) {
            table->setUpdatesEnabled(false);
            restoreUpdate = true;
        }

        if(table->item(row, COL_STATUS)->text() != st.stateStr)
            table->item(row, COL_STATUS)->setText(st.stateStr);

        {
            QLabel *label = qobject_cast<QLabel*>(table->cellWidget(row, COL_CONNECTION_STATE));
            QPixmap p;
            if (st.state==TcpClientState::Connected) {
                label->setToolTip("Client sends discover packets");
                p = QApplication::style()->standardPixmap(QStyle::SP_DialogApplyButton);
            } else {
                if(isInRun() && config.clients.contains(index) && config.clients[index].enabled){
                    QString errStr = QString("%1 disconnected").arg(index.toString());
                    qWarning() << errStr;
                    // todo handle mstream restart period
                    if(lastCombinerStats.hasPendingEvent && lastCombinerStats.regularEvents)
                        deviceError(index, errStr);
                }
                label->setToolTip("Device doesn't send discover packets");
                p = QApplication::style()->standardPixmap(QStyle::SP_DialogCloseButton);
            }
            if(connectionChanged)
                updateTableTextColor(row, index);

            label->setPixmap(p);
        }

        {
            QString text = QString("%1/%2/%3")
                    .arg(st.statistic.freePart, 4, 'f', 1)
                    .arg(st.statistic.dataPart, 4, 'f', 1)
                    .arg(st.statistic.trashPart, 4, 'f', 1);
            if(table->item(row, COL_TCP_BUFF)->text() != text)
                table->item(row, COL_TCP_BUFF)->setText(text);
        }

        {
            const auto &pd = st.lastProgDescription;
            const QString h = pd.hostName.split('.').first();
            if(table->item(row, COL_PROG_HOST)->text() != h)
                table->item(row, COL_PROG_HOST)->setText(h);
        }
    }

    if(restoreUpdate) {
        updateTableSize();
        table->setUpdatesEnabled(true);
    }
    pendingClientState.clear();
}

void EvbMainWindow::updatePendingProgramDescription()
{
    if(!hasNewPnpCache)
        return;
    hasNewPnpCache = false;

    for(auto pd : progDescrsCache) {
        if(!EVB_SEARCH_TARGETS.contains(pd.getDaqModule()))
            continue;

        auto index = pd.getClientIndex();
        if(!clients.contains(index))
            continue;

        if(isNonEventBaseProgram(pd) != nonEvBasedClients.contains(index)) {
            //what if in run?!
            updateAnalizerList();
        }

        if(index.getProgramType() == DaqModule(DaqModule::Type::MStream)) {

            const int row = getTableRowByClientIndex(index);
            if(row != -1) {
                auto typeStr = pd.options.value("dev_id_str");

                if(typeStr.isEmpty()) {
                    auto idStr = pd.options.value("dev_id");
                    typeStr = QString(getDeviceTypeName(idStr.toInt(nullptr, 16)));
                }
                if(!typeStr.isEmpty()) {
                    auto item = ui->tableWidget->item(row, COL_PROG_TYPE);
                    if(item->text() != typeStr)
                        item->setText(typeStr);
                }
            }
        }
    }

    updateButtonAddEnableStates();
}

void EvbMainWindow::sendRunDocument() const
{
    auto cfgObj = QVariant::fromValue(config).value<QJsonObject>();
    RunDocument rd(loc_runId, configDb->getClientIndex(), cfgObj);
    emit sendRunDocumentToRc(rd);
}
