#ifndef EVBTLVTYPES_H
#define EVBTLVTYPES_H

#include <QList>

#include "lib-common/MpdRawTypes.h"
#include "base/RcRunId.h"

// ==== General types ====
struct TlvBaseRec {
    TlvBaseRec(quint32 _type, const QByteArray &p = {})
        : type(_type)
        , payload(p)
    {}
    virtual ~TlvBaseRec(){}
    quint32 totSize() const { return payload.size()+2*sizeof(quint32); }
    QByteArray getData() const {
        QByteArray a;
        a.append(reinterpret_cast<const char*>(&type), sizeof(type));
        const quint32 &len = payload.size();
        a.append(reinterpret_cast<const char*>(&len), sizeof(len));
        a.append(payload);
        return a;
    }
    quint32 type;
    QByteArray payload;
};

struct TlvNumberRecord : public TlvBaseRec {
    TlvNumberRecord(quint32 _type, quint32 val)
        : TlvBaseRec(_type, {reinterpret_cast<const char*>(&val), sizeof(val)})
    {}
};

struct ComplexTlvRecord : public TlvBaseRec {
    ComplexTlvRecord(quint32 _type) : TlvBaseRec(_type) {}
    virtual ~ComplexTlvRecord(){}
    void setPayload(const QList<TlvBaseRec> &list) {
        payload.clear();
        for(auto v : list)
            payload.append(v.getData());
    }
};

// ==== Specific types ====

struct RunNumberRecord : public TlvNumberRecord {
    RunNumberRecord(quint32 runNum)
        : TlvNumberRecord(MPD_SYNC_RUN_NUMBER, runNum)
    {}
};

struct RunIndexRecord : public TlvBaseRec {
    RunIndexRecord(QString index)
        : TlvBaseRec(MPD_SYNC_RUN_INDEX)
    {
        payload = QByteArray(index.toLatin1());
        // complete 32-bit format
        static const constexpr char PADDING_CHAR = 0;

        auto v = payload.size()%sizeof(quint32);
        if(v)
            payload.append(sizeof(quint32)-v, PADDING_CHAR);
        Q_ASSERT(payload.size()%sizeof(quint32) == 0);
    }

    QByteArray runIndexLatin1;
};

struct FileIdRecord : public TlvNumberRecord {
    FileIdRecord(quint32 fileId)
        : TlvNumberRecord(MPD_SYNC_FILE_ID, fileId)
    {}
};

struct EvbEvOrderRecord : public TlvNumberRecord {
    EvbEvOrderRecord(quint32 fileId)
        : TlvNumberRecord(MPD_SYNC_EV_ORDER, fileId)
    {}
};

struct SartStopRecord : public ComplexTlvRecord {
    SartStopRecord(bool start, quint32 runNumber, QString runIndex)
        : ComplexTlvRecord(start ? MPD_SYNC_RUN_START : MPD_SYNC_RUN_STOP)
    {
        setPayload({RunNumberRecord(runNumber), RunIndexRecord(runIndex)});
    }
};

struct FileSplitRecord : public ComplexTlvRecord
{
    FileSplitRecord(bool start, RcRunId runId, uint splitId, std::optional<quint32> evbOrder)
        : ComplexTlvRecord(start ? MPD_SYNC_FILE_BEGIN: MPD_SYNC_FILE_END)
    {
        QList<TlvBaseRec> list{RunNumberRecord(static_cast<quint32>(runId.runNumber)),
                    RunIndexRecord(runId.runIndex),
                    FileIdRecord(splitId)};
        if(evbOrder.has_value())
            list << EvbEvOrderRecord(evbOrder.value());
        setPayload(list);
    }
};

#endif // EVBTLVTYPES_H
