#include "MscThresholdModel.h"
#include "MscModelItem.h"

#include <QDebug>
#include <QColor>

#include "config/MscConfigDispatcher.h"
#include "base/DeviceIndexEnabled.h"

namespace {
    static const QVector<int> CFG_ROLES = {
        Qt::DisplayRole,
        Qt::EditRole
    };
    static const QVector<int> STATUS_ROLES = {
        Qt::DisplayRole
    };
}

MscThresholdModel::MscThresholdModel(QObject *parent)
    : QAbstractItemModel(parent)
    , rootItem(new MscModelItem())
{
    auto configDispatcher = &MscConfigDispatcher::instance();
    configDispatcher->subscribeConfig(this);
}

QVariant MscThresholdModel::data(const QModelIndex &index, int role) const
{
    if(!index.isValid())
        return QVariant();
    switch (role) {
    case Qt::DisplayRole:
        return displayData(index);
    case Qt::EditRole:
        return editData(index);
    case Qt::ForegroundRole:
        return foregroundRoleData(index);
    }
    return QVariant();
}

bool MscThresholdModel::setData(const QModelIndex &model_index,
                                const QVariant &value,
                                int role)
{
    if(role != Qt::EditRole)
        return false;

    if(!model_index.isValid())
        return false;

    const auto item = getItemByIndex(model_index);
    const auto col = getColumnByIndex(model_index);
    if(!item)
        return false;
    int ch = item->item_row;

    if (!config.contains(item->index))
        return false;
    auto &chThreshold = config.findDevice(item->index).mscCore.chThreshold;

    if(col != MscThresholdColumns::COL_THRESHOLD)
        return false;

    for(int i : chThreshold.keys()) {
        if(ch < 0 || ch == i)
            chThreshold[i] = value.toInt();
    }

    columnDataChanged(MscThresholdColumns::COL_THRESHOLD);
    emit configChanged(config);
    return false;
}

QVariant MscThresholdModel::headerData(int section,
                                       Qt::Orientation orientation,
                                       int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<MscThresholdColumns>(section)) {
        case MscThresholdColumns::COL_CH:
            return "Ch #";
        case MscThresholdColumns::COL_THRESHOLD:
            return "Threshold";
        case MscThresholdColumns::COL_STATUS:
            return "Status";
        default:
            return {};
        }
    }
    return {};
}

QModelIndex MscThresholdModel::index(int row, int column, const QModelIndex &parent) const
{
    if(!hasIndex(row, column, parent))
        return QModelIndex();
    const MscModelItem *parentItem = parent.isValid() ?
        static_cast<MscModelItem*>(parent.internalPointer()) :
        rootItem;
    MscModelItem *childItem = parentItem->child(row);
    if(childItem) {
        return createIndex(row, column, childItem);
    }
    return QModelIndex();
}

QModelIndex MscThresholdModel::parent(const QModelIndex &index) const
{
    if(!index.isValid())
        return QModelIndex();

    const MscModelItem *childItem =
        static_cast<MscModelItem*>(index.internalPointer());
    if(!childItem)
        return QModelIndex();
    MscModelItem *parentItem = childItem->parent;
    if(parentItem == rootItem)
        return QModelIndex();

    return createIndex(parentItem->row(), 0, parentItem);
}

int MscThresholdModel::rowCount(const QModelIndex &parent) const
{
    const auto *parentItem = parent.isValid() ?
        static_cast<MscModelItem*>(parent.internalPointer()) :
        rootItem;
    return parentItem->childCount();
}

void MscThresholdModel::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(config == newConfig)
        return;


    const auto &oldSet = config.getDeviceIndexEnabledMap().getEnabled();
    const auto &newSet = newConfig.getDeviceIndexEnabledMap().getEnabled();
    config = newConfig;

    QMap<DeviceIndex, int> chNum_new;
    for (const auto &index : config.deviceKeys()) {
        chNum_new[index] = config.findDevice(index).mscCore.hist.size();
    }

    if((newSet != oldSet)||(chNum_new != chNum)){
        chNum = chNum_new;
        reset();
    }
    else{
        auto top = index(0,0);
        auto lastDev = index(rowCount(top)-1, columnCount(top)-1, top);
        auto bottom = index(rowCount(lastDev)-1, columnCount(lastDev)-1, lastDev);
        emit dataChanged(top, bottom, {Qt::DisplayRole});
    }
}

void MscThresholdModel::statusUpdated(const MscStatus &status)
{
    devStatus = status;

    int col = static_cast<int>(MscThresholdColumns::COL_STATUS);
    auto firstDev = index(0,col);
    auto lastDev = index(rowCount()-1,col);
    auto lastCh = index(rowCount(lastDev)-1, col, lastDev);
    emit dataChanged(firstDev, lastCh, {Qt::DisplayRole, Qt::ForegroundRole});
}

QVariant MscThresholdModel::displayData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    if(!item) return QVariant();
    const auto col = getColumnByIndex(index);

    auto ch = item->item_row;
    switch (col) {
    case MscThresholdColumns::COL_CH:
        return ch < 0 ? item->index.getSerialStr()
                      : QString::number(item->item_row);
    case MscThresholdColumns::COL_STATUS:
        if(ch<0)
            return {};
        if (!devStatus.inputStatus.has_value())
            return {};
        return QChar(devStatus.inputStatus.value().hitState &
                     (1<<item->item_row) ? 0x2713 : 0x2715);
    default:;
    }

    return QVariant();
}

QVariant MscThresholdModel::editData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto &col = getColumnByIndex(index);
    if(!item)
        return QVariant();

    if(col!=MscThresholdColumns::COL_THRESHOLD)
        return QVariant();

    auto ch = std::max(item->item_row, 0);
    const auto &chThreshold = config.findDevice(item->index).mscCore.chThreshold;
    return chThreshold[ch];
}

QVariant MscThresholdModel::foregroundRoleData(const QModelIndex &index) const
{
    const auto item = getItemByIndex(index);
    const auto col = getColumnByIndex(index);
    if(!item || item->item_row<0)
        return QVariant();

    if (!devStatus.inputStatus.has_value())
        return {};
    auto hitSt = devStatus.inputStatus.value().hitState;
    switch (col) {
    case MscThresholdColumns::COL_STATUS:
        return hitSt & (1<<index.row()) ? QColor(Qt::darkGreen) : QColor();
    default: return QVariant();
    }
}

void MscThresholdModel::columnDataChanged(MscThresholdColumns col)
{
    QVector<int> roles;
    switch (col) {
        case MscThresholdColumns::COL_THRESHOLD:
            roles = CFG_ROLES;
        break;
        case MscThresholdColumns::COL_STATUS:
            roles = STATUS_ROLES;
        break;
        default:;
    }

    auto top = index(0,0);
    auto devs = rowCount(top);
    for(int d=0; d<devs; ++d) {
        auto dev = index(d, static_cast<int>(col), top);
        emit dataChanged(dev, dev, roles);
        auto chNum = rowCount(dev);
        for(int ch=0; ch<chNum; ++ch) {
            auto chItem = index(ch, static_cast<int>(col), dev);
            emit dataChanged(chItem, chItem, roles);
        }
    }
}

void MscThresholdModel::reset()
{
    beginResetModel();
    rootItem->children.clear();
    for (const auto &key : config.deviceKeys()) {
        auto devItem = new MscModelItem(rootItem);
        devItem->index = key;
        for(int hId=0; hId<chNum[devItem->index]; ++hId) {
            auto chItem = new MscModelItem(devItem);
            chItem->index = key;
            chItem->item_row = hId;
            devItem->addChild(chItem);
        }
        rootItem->addChild(devItem);
    }
    endResetModel();
    return;
}

MscModelItem *MscThresholdModel::getItemByIndex(const QModelIndex &index) const
{
    return static_cast<MscModelItem*>(index.internalPointer());
}

MscThresholdColumns MscThresholdModel::getColumnByIndex(const QModelIndex &index) const
{
    return static_cast<MscThresholdColumns>(index.column());
}
