/*
**    Copyright 2014-2015 Ivan Filippov
**    Copyright 2014-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "VmeTrigDevice.h"

#include <QDebug>

#include "mregdev/MregDeviceAdcm.h"
#include "regio/regop.h"

namespace {
enum {
    REG_FVME2TMWR_STATUS = 0x201, // 16-bit RO, bit[0] - tm_valid
    REG_FVME2TMWR_RUN_EVENT_NUMBER = 0x204, // 64-bit RO
    REG_FVME2TMWR_WR_LINK_ERROR_COUNTER = 0x100, // Reg 0x100, 32-bit RO counter of PHY code error
    REG_FVME2TMWR_WR_SYNC_LOST_COUNTER  = 0x102, // Reg 0x102, 32-bit RO counter of white rabbit sync lost
    REG64_FVME2TMWR_BUSY_TIMER = 0x300,
    REG64_FVME2TMWR_IDLE_TIMER = 0x310
};

constexpr fw_version_t REV_HOLD_COUNTERS(1, 0, 22708);
}

VmeTrigDevice::VmeTrigDevice(const DeviceIndex &deviceIndex, QObject *parent) :
    ModularMregDevice(deviceIndex, status, parent),
    status(deviceIndex)
{}

bool VmeTrigDevice::readStatus()
{
    return
        ModularMregDevice::readStatus() &&
        readVmeTrigStatus();
}

bool VmeTrigDevice::readVmeTrigStatus()
{
    status.ident = getIdent();

    WrStatus wrStatus;
    VmeTrigCoreStatus coreStatus;

    RegOpVector r;
    if (getDeviceFwVersion() >= REV_HOLD_COUNTERS)
        r.RegWrite16(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 2);
    r.RegRead64(REG_FVME2TMWR_RUN_EVENT_NUMBER, &coreStatus.evNum);
    r.RegRead32(REG_FVME2TMWR_WR_SYNC_LOST_COUNTER, &wrStatus.syncFail);
    r.RegRead32(REG_FVME2TMWR_WR_LINK_ERROR_COUNTER, &wrStatus.linkError);
    r.RegRead16(REG_FVME2TMWR_STATUS, &coreStatus.deviceStatus);
    r.RegRead64(REG64_FVME2TMWR_BUSY_TIMER, &coreStatus.busyTimer);
    r.RegRead64(REG64_FVME2TMWR_IDLE_TIMER, &coreStatus.idleTimer);
    if (getDeviceFwVersion() >= REV_HOLD_COUNTERS)
        r.RegWrite16(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0);

    if (!regOpExec(r))
        return false;

    status.wrStatus = wrStatus;
    status.coreStatus = coreStatus;

    emit updateStatus(getDeviceIndex(), status);
    return true;
}

void VmeTrigDevice::vmeReadoutUpdated(const DeviceIndex &serial, bool _evNumRead)
{
    Q_ASSERT(getDeviceIndex() == serial);
    if (getDeviceIndex() != serial)
        return;
    if (doEvNumRead== _evNumRead)
        return;
    doEvNumRead = _evNumRead;
    qDebug().noquote() << getIdent() << "set readout" << doEvNumRead;
}

std::optional<quint64> VmeTrigDevice::readGlobEvNum()
{
    quint64 globEvNum;
    RegOpVector r;
    if(getDeviceFwVersion() >= REV_HOLD_COUNTERS)
        r.RegWrite16(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0x2); // lock
    r.RegRead64(REG_FVME2TMWR_RUN_EVENT_NUMBER, &globEvNum);
    if(getDeviceFwVersion() >= REV_HOLD_COUNTERS)
        r.RegWrite16(mlink::MregDeviceAdcm::ADCM16_CTRL_REG, 0); // unlock

    if (!regOpExec(r))
        return {};
    return globEvNum;
}

void VmeTrigDevice::evNumRequest(int checkIndex)
{
    if(!doEvNumRead) return;

    auto opt = readGlobEvNum();
    if (opt.has_value()) {
        const auto &evNum = opt.value();
        emit evNumResponse(checkIndex, getDeviceIndex(), evNum);
    }
}
