/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "FitterGausProfile.h"
#include <cmath>

namespace dsplib {

static const int MAX_FITGAUS_POINTS = 21;

bool FitterGausProfile::isValid() const
{
    if ((fititer < 0) || (fititer > 10000)) {
        return false;
    }
    if ((fitside < 0) || (fitside > 63)) {
        return false;
    }
    return true;
}

int FitterGausProfile::get_dec_fititer() const
{
    int fiter;
    switch (fprof) {
    case FIT_PROFILE_SUPERFAST:
        fiter = 1;
        break;
    case FIT_PROFILE_FAST:
        fiter = 1;
        break;
    case FIT_PROFILE_OPTIMAL:
        fiter = 2;
        break;
    case FIT_PROFILE_PRECISION:
        fiter = 26;
        break;
    case FIT_PROFILE_CUSTOM:
        fiter = fititer;
        break;
    default:
        fiter = default_fititer;
        break;
    }

    if (fiter > 1000)
        fiter = default_fititer;
    return fiter;
}

/*fit_algo_t*/
int FitterGausProfile::get_dec_fitalgo() const
{
    int algo = default_fitalgo;
    if (fprof == FIT_PROFILE_CUSTOM) algo = fitalgo;

    return algo;
}

int FitterGausProfile::get_dec_fitfixpar() const
{
    int fixpar = default_fitfixpar;
    if (fprof == FIT_PROFILE_CUSTOM) fixpar = fitfixpar;

    return fixpar;
}

int FitterGausProfile::get_fit_half_np(double pulse_width_steps) const
{
    int fside = 0;
    double k = 1.0, b = 0;
    if (fprof == FIT_PROFILE_SUPERFAST) {k = 0.7; }
    if (fprof == FIT_PROFILE_FAST) {k = 1.0; }
    if (fprof == FIT_PROFILE_OPTIMAL) { k=1.2; b = 1; }
    if (fprof == FIT_PROFILE_PRECISION) {k = 1.6; b = 1; }
    if (fprof == FIT_PROFILE_CUSTOM) {fside = fitside;}

    if (fside <= 0) // auto
        fside = static_cast<int>(trunc(pulse_width_steps * k + b));
    if (fside<=0 || fside > 63)
        fside = default_fitside;
    if (fside < 2)
        fside = 2;
    // safe check
    int maxfside = (MAX_FITGAUS_POINTS-1)/2;
    if (fside > maxfside)
        fside = maxfside;

    //  printf("ch %2d: pwid=%f, k=%.1f, b=%.0f  fside=%d\n", ch, dec.pwid[ch], k, b, fside);
    return fside;
}

} // namespace dsplib
