/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ClientEnableTableModel.h"

#include <optional>

int ClientEnableTableModel::rowCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return clients.size();
}

int ClientEnableTableModel::columnCount(const QModelIndex &parent) const
{
    Q_UNUSED(parent)
    return COL_TOTAL;
}

ClientIndex ClientEnableTableModel::clientIndexOf(int row) const
{
    if (row < 0)
        return {};
    int r = 0;
    for (auto it = clients.cbegin(); it != clients.cend(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

int ClientEnableTableModel::rowOf(const ClientIndex &clientIndex) const
{
    return std::distance(clients.begin(), clients.find(clientIndex));
}

static const double ResetModelRatioThreshold = 0; // always use resetModel

void ClientEnableTableModel::setClientList(const ClientIndexEnabledMap &map)
{
    if(map == clients)
        return;

    const ClientIndexSet newSet(map);
    const ClientIndexSet oldSet(clients);
    const auto toRemove = oldSet-newSet;
    const auto toAdd = newSet-oldSet;
    bool reset = false;
    reset |= (toRemove.size() >= clients.size() * ResetModelRatioThreshold);
    reset |= (toAdd.size() >= clients.size() * ResetModelRatioThreshold);
    if (reset) {
        beginResetModel();
        for(auto i : toRemove)
            clients.remove(i);
        for(auto i : toAdd)
            clients.insert(i, map[i]);
        endResetModel();
    } else {
        for(auto i : toRemove) {
            removeRow(rowOf(i));
            clients.remove(i);
        }
        for(auto i : toAdd) {
            clients.insert(i, map[i]);
            insertRow(rowOf(i));
        }
    }
    for(auto i : (newSet & oldSet)) {
        if(clients[i] == map[i])
            continue;
        clients[i] = map[i];
        auto modIndex = index(rowOf(i), COL_ENABLE);
        emit dataChanged(modIndex, modIndex);
    }
}

void ClientEnableTableModel::setEnabled(bool en)
{
    enabled = en;

    auto indexFrom = createIndex(0, 0);
    auto indexTo = createIndex(rowCount()-1, columnCount()-1);
    emit dataChanged(indexFrom, indexTo);
}

Qt::ItemFlags ClientEnableTableModel::flags(const QModelIndex & index) const
{
    auto f = QAbstractTableModel::flags(index);
    switch (index.column()) {
    case COL_ENABLE:
        if(enabled)
            f |= Qt::ItemIsUserCheckable;
        else
            f &= ~Qt::ItemIsEnabled;
        break;
    default:;
    }
    return f;
}

bool ClientEnableTableModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    const auto &clientIndex = clientIndexOf(index.row());
    if (clientIndex.isEmpty())
        return {};

    const auto it = clients.find(clientIndex);
    if (it == clients.end())
        return false;

    std::optional<bool> st;
    if (role == Qt::CheckStateRole && index.column() == COL_ENABLE)
            st = (value == Qt::Checked);
    if (role == Qt::DisplayRole && index.column() == COL_ENABLE)
        st = value.value<bool>();

    if(!st.has_value())
        return false;
    if (*it == st.value())
        return true;

    *it = st.value();
    emit dataChanged(index, index);
    emit clientListChanged(clients);
    return true;
}

QVariant ClientEnableTableModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    const auto &clientIndex = clientIndexOf(index.row());
    if (clientIndex.isEmpty())
        return {};

    const auto &en = clients[clientIndex];

    switch(column) {
    case COL_ENABLE:
        switch (role) {
        case Qt::EditRole: return en;
        case Qt::CheckStateRole: return en ? Qt::Checked : Qt::Unchecked;
        }
        break;
    case COL_TYPE:
        switch (role) {
        case Qt::DisplayRole: return clientIndex.getProgramType().toString();
        }
        break;
    case COL_INDEX:
        switch (role) {
        case Qt::DisplayRole: return clientIndex.getProgramIndex().toString();
        }
        break;
    }
    return {};
}

QVariant ClientEnableTableModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (static_cast<Columns>(section)) {
        case COL_ENABLE: return tr("En");
        case COL_TYPE: return tr("Type");
        case COL_INDEX: return tr("Index");
        case COL_TOTAL: return {};
        }
    }
    return ClientIndexTableModel::headerData(section, orientation, role);
}

QVector<ClientEnableTableModel::Columns> ClientEnableTableModel::hiddenColumns()
{
    return  {
        COL_ENABLE,
        COL_TYPE,
        COL_INDEX
    };
}
