/*
**    Copyright 2010-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef FIRDESIGNDIALOG_H
#define FIRDESIGNDIALOG_H

#include "dsplib/WindowFunction.h"
#include <complex>
#include <vector>
#include <QDialog>
#include <QVector>

class DecoderConfig;
class FirDesignConfig;
class FirDesignDialogPrivate;
struct gauss_param_t;

class QAbstractButton;
class WaveformData;

using dsplib::WindowFunction;

namespace Ui {
class FirDesignDialog;
} // namespace Ui

class FirDesignDialog : public QDialog
{
    Q_OBJECT
public:
    explicit FirDesignDialog(QWidget *parent = nullptr);
    DecoderConfig getDecoderConfig() const;
    ~FirDesignDialog() override;

signals:
    void decoderConfigChanged(const DecoderConfig &);

public Q_SLOTS:
    void setAdcConfig(const FirDesignConfig &r);
    void setDecoderConfig(const DecoderConfig &r);

private slots:
    void on_comboBoxWindowingFunction_activated(int);
    void on_sbfirnp_valueChanged(int);
    void on_sbsnf_valueChanged(double); // Noise floor of source
    void on_sbrnf_valueChanged(double); // Noise floor of target
    void on_sbpos_valueChanged(double);
    void on_sbwid_valueChanged(double);
    void on_sbexp_valueChanged(double);
    void on_pushButtonGenerateFir_clicked();
    void on_buttonBox_clicked(QAbstractButton *button);
    void on_buttonBoxShape_clicked(QAbstractButton *button);
    void on_buttonBoxFir_clicked(QAbstractButton *button);

protected:
    void setButtons();
    bool modified() const;
    void apply();
    void restoreDefaults();
    void replotINP();
    void replotINPFFT();
    void replotFIR();
    void replotFIRFFT();
    void replotTEST();
    void Restore();
    void Redraw();
    double reqf(double i) const;
    double reqfunscaled(double i) const;
    //    double ca( unsigned int ch, int i ) const;
    double data_latency() const;
    std::vector<double> getCalAvgY() const;
    std::vector<double> getCalAvgXInNs() const;
    size_t getRefSize() const;
    std::vector<double> getRefY_n(double amplitudeScaler, int n) const;
    std::vector<double> getRefY() const;
    std::vector<double> getRefXInNs() const;
    double getCalAvgAmplitude() const;
    WindowFunction::fwinf_t win_type() const;
    gauss_param_t find_hit(const std::vector<double> &testResult) const;
    std::vector<std::complex<double>> makeRealDFT(const std::vector<double> &v) const;
    static std::vector<double> magnitudesDb(const std::vector<std::complex<double> > &v);
    static std::vector<double> fftFrequencies(int timePoints);
    template<typename Container>
    bool writeNumbersFile(const QString &fileName, const Container &data);
    QVector<double> readNumbersFile(const QString fileName);

    void clearFir();
    void makeTestShape();
    void readShapeFile();
    void writeShapeFile();
    void writeFirFile();
    void readFirFile();

private:
    Ui::FirDesignDialog *const ui;
    Q_DECLARE_PRIVATE(FirDesignDialog)
    FirDesignDialogPrivate *const d_ptr;
    int restoring = 0;
};

#endif // FIRDESIGNDIALOG_H
