/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ProxyChain.h"

#include <QSortFilterProxyModel>

#include "DeviceDynamicStateFilterProxyModel.h"
#include "DeviceIdFilter.h"
#include "DeviceIndexFilterProxyModel.h"
#include "DeviceIdentityProxyModel.h"
#include "NetworkFilterProxyModel.h"

ProxyChain::ProxyChain(QObject *parent)
    : QAbstractProxyModel(parent)
{
}

QAbstractProxyModel *ProxyChain::createProxyChain(
        QAbstractItemModel *model,
        DeviceIdFilterMode deviceIdFilterMode,
        DeviceIndexFilterMode deviceIndexFilterMode,
        DeviceStateFilterMode deviceStateFilterMode,
        NetworkFilterMode networkFilterMode)
{
    deviceIdFilter = new DeviceIdFilter(this);
    deviceIdFilter->setSourceModel(model);
    if (deviceIdFilterMode == EnableDeviceIdFilter)
        connect(this, &ProxyChain::setDeviceIdFilter,
                deviceIdFilter, &DeviceIdFilter::setFilter);

    deviceIndexFilter = new DeviceIndexFilterProxyModel(this);
    deviceIndexFilter->setSourceModel(deviceIdFilter);
    deviceIndexFilter->setInverseMode(deviceIndexFilterMode != NormalDeviceFilter);
    if (deviceIndexFilterMode != DisableDeviceFilter)
        connect(this, &ProxyChain::setDeviceIndexFilter,
                deviceIndexFilter, &DeviceIndexFilterProxyModel::setFilter);

    deviceDynamicStateFilter = new DeviceDynamicStateFilterProxyModel(this);
    deviceDynamicStateFilter->setSourceModel(deviceIndexFilter);
    deviceDynamicStateFilter->setFilter(
                deviceStateFilterMode!= DisableDeviceStateFilter,
                deviceStateFilterMode== DynamicDeviceStateFilter);

    networkFilter = new NetworkFilterProxyModel(this);
    networkFilter->setSourceModel(deviceDynamicStateFilter);
    if (networkFilterMode == EnableNetworkFilter)
        connect(this, &ProxyChain::setNetworkFilter,
                networkFilter, &NetworkFilterProxyModel::setFilter);

    sortProxy = new QSortFilterProxyModel(this);
    sortProxy->setSourceModel(networkFilter);

    identityProxy = new DeviceIdentityProxyModel(this);
    identityProxy->setSourceModel(sortProxy);
    return identityProxy;
}

QModelIndex ProxyChain::mapToDataSource(const QAbstractItemModel *model, const QModelIndex &index)
{
    Q_ASSERT(model);
    if (!model)
        return QModelIndex();
    Q_ASSERT(index.isValid());
    if (!index.isValid())
        return QModelIndex();
    QModelIndex sourceIndex(index);

    int i=0;
    while (auto proxyModel = qobject_cast<const QAbstractProxyModel *>(model)) {
        sourceIndex = proxyModel->mapToSource(sourceIndex);
        Q_ASSERT(sourceIndex.isValid());
        auto source = proxyModel->sourceModel();
        if (source == model)
            break; // loop
        model = source;
        i++;
    }
    return sourceIndex;
}

QModelIndex ProxyChain::mapFromSource(const QModelIndex &sourceIndex) const
{
    Q_ASSERT(deviceIdFilter);
    Q_ASSERT(deviceIndexFilter);
    Q_ASSERT(deviceDynamicStateFilter);
    Q_ASSERT(networkFilter);
    Q_ASSERT(sortProxy);
    Q_ASSERT(identityProxy);
    QModelIndex index = sourceIndex;
    index = deviceIdFilter->mapFromSource(index);
    index = deviceIndexFilter->mapFromSource(index);
    index = deviceDynamicStateFilter->mapFromSource(index);
    index = networkFilter->mapFromSource(index);
    index = sortProxy->mapFromSource(index);
    index = identityProxy->mapFromSource(index);
    return index;
}

QModelIndex ProxyChain::mapToSource(const QModelIndex &proxyIndex) const
{
    return mapToDataSource(identityProxy, proxyIndex);
}

QModelIndex ProxyChain::index(int row, int column, const QModelIndex& parent) const
{
    Q_ASSERT(identityProxy);
    return identityProxy->index(row, column, parent);
}

QModelIndex ProxyChain::parent(const QModelIndex& child) const
{
    Q_ASSERT(identityProxy);
    return identityProxy->parent(child);
}

int ProxyChain::columnCount(const QModelIndex& parent) const
{
    Q_ASSERT(identityProxy);
    return identityProxy->columnCount(parent);
}

int ProxyChain::rowCount(const QModelIndex& parent) const
{
    Q_ASSERT(identityProxy);
    return identityProxy->rowCount(parent);
}
