#include "HitCoreModule.h"

#include "HitCore/HitCoreConfig.h"
#include "base/DeviceId.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "HitCoreStatus.h"

namespace {
enum {
    REG32_ID = 0x00,
    REG_LOCK = 0x03,
    REG64_HIT_CORE_TS = 0x04,
    REG_CHANNEL_COUNT = 0x08,
    REG64_INPUT_SIGNAL_POLARITY = 0x18, // 1-negative, 0-positive
    REG64_HIT_CMP_STATE = 0x10,         // RO, Comparators State
    REG32_HIT_CNT_CH_BASE = 0x80,       // RO, 32-bit
};

inline constexpr const quint16 REG_LOCK_BIT_LOCK = 0x0001;
} // namespace

HitCoreModule::HitCoreModule(const MregBusInterface &bus) : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        if (getDeviceId() == DEVICE_ID_PHOS_FEC)
            setBaseAddr(0x600, 0xFF);
    }
}

QList<AFI_SDB_ID> HitCoreModule::getSupportedSdbIds()
{
    return {AFI_SDB_Input_Hit_Core};
}

bool HitCoreModule::writeConfig(const HitCoreConfig &config)
{
    RegOpVector r;
    r.RegWrite16(REG64_INPUT_SIGNAL_POLARITY, config.polarity);
    return regOpExecRebased(r);
}

bool HitCoreModule::readStatus(HitCoreStatus *result)
{
    HitCoreStatus status;
    {
        RegOpVector r;
        r.RegRead32(REG32_ID, &status.id);
        r.RegRead16(REG_CHANNEL_COUNT, &status.nch);
        if (!regOpExecRebased(r))
            return {};
    }
    if (!status.isValid())
        return {};
    status.cnt.resize(status.nch);

    RegOpVector r;
    r.RegWrite16(REG_LOCK, 0);                 // unlock (update)
    r.RegWrite16(REG_LOCK, REG_LOCK_BIT_LOCK); // lock
    r.RegRead64(REG64_HIT_CORE_TS, &status.ts);
    r.RegRead64(REG64_HIT_CMP_STATE, &status.compState);
    for (int i=0; i<status.cnt.size(); i++)
        r.RegRead32(REG32_HIT_CNT_CH_BASE + 2*i, &status.cnt[i]);
    if (!regOpExecRebased(r))
        return {};

    if (result)
        *result = status;
    return true;
}

std::optional<HitCoreStatus> HitCoreModule::readStatus()
{
    HitCoreStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

QVector<quint64> HitCoreModule::readCmpState(int num_samples)
{
    QVector<quint64> cmpState(num_samples);
    RegOpVector r;
    for (int i = 0; i < num_samples; ++i)
        r.RegRead64(REG64_HIT_CMP_STATE, &cmpState[i]); // read multiple times
    if (!regOpExecRebased(r))
        return {};
    return cmpState;
}
