#include "MlinkCsrModule.h"

#include "base/DeviceId.h"
#include "base/util.h"
#include "device-modules/OneWire/OneWire.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

QList<AFI_SDB_ID> MlinkCsrModule::getSupportedSdbIds()
{
    return { AFI_SDB_MLINK_DIG_CSR, AFI_SDB_MLINK_CSR };
}

MlinkCsrModule::CsrType MlinkCsrModule::getCsrType() const
{
    switch(getSdbId()) {
    case AFI_SDB_MLINK_DIG_CSR: return CsrType::Dig;
    case AFI_SDB_MLINK_CSR: return CsrType::Default;
    default: return CsrType::Default;
    }
}

MlinkCsrModule::CsrType MlinkCsrModule::getCsrType(DeviceId deviceId)
{
    switch (deviceId) {
    case DEVICE_ID_8CADC:
    case DEVICE_ID_ADC8E:
    case DEVICE_ID_ADC16LTCM:
    case DEVICE_ID_ADC32P:
    case DEVICE_ID_ADC64:
    case DEVICE_ID_ADC64S2:
    case DEVICE_ID_ADC64WR:
    case DEVICE_ID_ADC64VE:
    case DEVICE_ID_ADC64VE2:
    case DEVICE_ID_ADC64VE_XGE:
    case DEVICE_ID_ADC64ECAL:
    case DEVICE_ID_ADCM16:
    case DEVICE_ID_ADCM163:
    case DEVICE_ID_ADCM164:
    case DEVICE_ID_BRIC:
    case DEVICE_ID_CRU16WR:
    case DEVICE_ID_TTVXS:
    case DEVICE_ID_TDC64VHLE:
    case DEVICE_ID_TDC64VLE:
    case DEVICE_ID_TDC72VXS:
    case DEVICE_ID_VME_TDC64V:
    case DEVICE_ID_VME_TDC32VL:
    case DEVICE_ID_VME_TDC72VHL:
    case DEVICE_ID_VME_TTB9V:
    case DEVICE_ID_TQDC16VS_DIG:
    case DEVICE_ID_VMM128RO:
        return CsrType::Dig;

    case DEVICE_ID_ADC8BEDRS:
    case DEVICE_ID_BPM4100:
    case DEVICE_ID_CTF6:
    case DEVICE_ID_EVADC:
    case DEVICE_ID_FVME:
    case DEVICE_ID_HRB6ASD:
    case DEVICE_ID_HRB128E:
    case DEVICE_ID_MSC8E:
    case DEVICE_ID_MSC16VE_TM:
    case DEVICE_ID_NUCLORF:
    case DEVICE_ID_TEST:
    case DEVICE_ID_TTB9V:
    case DEVICE_ID_TRC:
    case DEVICE_ID_U40VE:
    case DEVICE_ID_U40VE_8SPI:
    case DEVICE_ID_U40VE_IDEAS:
    case DEVICE_ID_U40VE_RC:
    case DEVICE_ID_U40VE_SI:
    case DEVICE_ID_UT24VE:
    case DEVICE_ID_VME_TRIG:
        return CsrType::Default;

    case DEVICE_ID_PHOS_FEC:
    case DEVICE_ID_VME_MSC16V:
    case DEVICE_ID_MSC16VE_E:
    case DEVICE_ID_TQDC16VS_E:
        return CsrType::Default;

    case DEVICE_ID_UNKNOWN:
    case DEVICE_ID_AnyDevice:
        return CsrType::Default;
    }
    return CsrType::Default;
}

void MlinkCsrModule::post_init()
{
    if (!isAvailable()) {
        // no-SDB compatibility
        bus.setAddrRange({0x0040, 0x001F});
    }
}

// MlinkCsr registers
enum {
    REG_MLINK_CSR_DEVICE_ID   = 0x02,
    REG_MLINK_CSR_LIVEMAGIC   = 0x03,
    REG64_MLINK_CSR_OW_ID     = 0x10,
    REG_MLINK_CSR_TEMP        = 0x14,
    REG_MLINK_CSR_PCB_VER     = 0x15,
    REG_MLINK_CSR_FW_VER      = 0x16,
    REG_MLINK_CSR_FW_REV      = 0x17,
};

// Legacy (dig_csr registers)
enum {
    REG_MLINK_ADCM_LIVEMAGIC    = 0x08,
};

bool MlinkCsrModule::detect()
{
    return readDeviceId()
        && readOneWireId()
        && readFirmwareId();
}

std::optional<MlinkCsrStatus> MlinkCsrModule::readStatus()
{
    if (!readDeviceId())
        return {};
    if (!readOneWireId())
        return {};
    if (!readFirmwareId())
        return {};
    if (!readTemp())
        return {};

    return status;
}

bool MlinkCsrModule::readDeviceId()
{
    RegOpVector r;
    quint16 tmp = 0;
    r.RegRead16 (REG_MLINK_CSR_DEVICE_ID, &tmp);
    if (!regOpExecRebased(r))
        return {};
    status.device_id = static_cast<DeviceId>((tmp >> 8) & 0xFF);
    return true;
}

bool MlinkCsrModule::readTemp()
{
    quint16 r16;
    RegOpVector r;
    r.RegRead16 (REG_MLINK_CSR_TEMP, &r16);
    if (!regOpExecRebased(r))
        return {};
    // TODO: check for temp ready
    if (r16 == 0xFFFF)
        return {};
    status.temps = {{"1-Wire", convertOneWireTemp(r16)}};
    return true;
}

bool MlinkCsrModule::readLiveMagic(quint16 *value)
{
    quint16 reg;
    switch (getCsrType()) {
    case CsrType::Dig: reg = REG_MLINK_ADCM_LIVEMAGIC; break;
    default: reg = REG_MLINK_CSR_LIVEMAGIC;
    }
    quint16 r16;
    RegOpVector r;
    r.RegRead16 (reg, &r16);
    if (!regOpExecRebased(r))
        return {};
    if (value)
        *value = r16;
    return true;
}

bool MlinkCsrModule::writeLiveMagic(quint16 value, quint16 *readback)
{
    quint16 reg;
    switch (getCsrType()) {
    case CsrType::Dig: reg = REG_MLINK_ADCM_LIVEMAGIC; break;
    default: reg = REG_MLINK_CSR_LIVEMAGIC;
    }
    quint16 r16;
    RegOpVector r;
    r.RegWrite16(reg, value);
    r.RegRead16(reg, &r16);
    if (!regOpExecRebased(r))
        return {};
    if (readback)
        *readback = r16;
    return true;
}

bool MlinkCsrModule::readOneWireId()
{
    quint64 ow_id;
    RegOpVector r;
    r.RegRead64(REG64_MLINK_CSR_OW_ID, &ow_id);
    if (!regOpExecRebased(r))
        return {};

    if ((ow_id == 0) || (ow_id == ~0ull)) {
        qWarning() << "No identification chip present";
        return {};
    }
    if (! OneWire::oneWireIdValid(ow_id)) {
        qWarning() << QString("Unknown identification chip");
        return {};
    }
    if (!OneWire::oneWireCrcValid(ow_id)) {
        qWarning() << QString("Identification chip CRC mismatch (ow_id: 0x%1").arg(ow_id, 0, 16);
        return {};
    }
    status.id_chip_present = true;
    status.device_serial = (ow_id << 8) >> 16;
    return true;
}

bool MlinkCsrModule::readFirmwareId()
{
    std::array<quint16,3> val;
    RegOpVector r;
    r.RegRead16(REG_MLINK_CSR_FW_VER, &val[0]);
    r.RegRead16(REG_MLINK_CSR_FW_REV, &val[1]);
    r.RegRead16(REG_MLINK_CSR_PCB_VER, &val[2]);
    if (!regOpExecRebased(r))
        return {};
    status.fw_version = fw_version_t(val.at(0), val.at(1));
    quint16 hw_ver = val.at(2);
    status.hw_version = fw_version_t(hw_ver, 0);
    if (hw_ver != 0) {
        switch (static_cast<DeviceId>(getDeviceId())) {
        case DEVICE_ID_TTVXS:
        case DEVICE_ID_TDC72VXS:
        case DEVICE_ID_TDC64VHLE:
            if (!status.fw_version.isNull() && status.fw_version <= fw_version_t(1, 3, 0)) {
                quint8 pcb_ver_maj = 1 + ((hw_ver>>2) & 0x7);
                quint8 pcb_ver_min = hw_ver & 0x3;
                status.hw_version = fw_version_t(pcb_ver_maj, pcb_ver_min, 0);
            }
        default:;
        }
    }
    return true;
}
