#include "PhysicalTriggerModule.h"

#include "PhysicalTriggerConfig.h"
#include "PhysicalTriggerStatus.h"

#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

#include "device-modules/RcCounters/RcCounters.h"
#include "device-modules/RcCounters/RcCountersModule.h"
#include "device-modules/RcTrig/RcTrigModule.h"
#include "device-modules/SpillControl/SpillControlModule.h"
#include "device-modules/SpillControl/SpillStatus.h"

enum : quint16 {
    REG_RC_CORE_BASE        = 0x000,
    REG_RC_CORE_MAKS        = 0x01F,
    REG_CH_BASE             = 0x200,
    REG_CH_MASK             = 0x07F,
    REG_SPILL_BASE          = 0x300,
    REG_SPILL_MAKS          = 0x03F,
    REG_EVENT_MEM_BASE      = 0x500,
    REG_EVENT_MEM_MAKS      = 0x00F,
    REG_INPUT_COUNTERS_BASE = 0x600,
    REG_BA_COUNTERS_BASE    = 0x800, // BEFORE_AFTER
    REG_REDUC_COUNTERS_BASE = 0xA00, // after reduction
    REG_COUNTERS_MAKS       = 0x1FF,
};

// PhysicalTrigger Core
enum : quint16 {
    REG_CORE_CTRL           = 0x00,
    REG_CH_NUM              = 0x01,
    REG_MATCH_WIN           = 0x02,
    REG_TRIG_DELAY          = 0x03,
    REG32_CH_EN             = 0x06,
    REG32_START_CH_MASK     = 0x08,
    REG32_CAL_TRIG_OFFSET   = 0x10,
    REG32_CAL_TRIG_PERIOD   = 0x12,
    REG_CAL_TRIG_COUNT      = 0x14,
    REG32_READ_CH_MASK      = 0x16,
};

// PhysicalTrigger Channel
enum : quint16 {
    REG64_CH_SHIFT    = 0x4,
};

// Event memory
enum : quint16 {
    REG_CH_WF_LATENCY       = 0x0,
    REG_CH_WF_READ_POINTS   = 0x1,
};

enum : quint16 {
    REG_CSR_BIT_CNT_RESET   = 1 << 0,
    REG_CSR_BIT_CNT_LOCK    = 1 << 1,
    REG_CSR_BIT_OCS_DELAYED = 1 << 2,
    REG_CSR_BIT_USE_LUT     = 1 << 3,
};

template<typename  T>
class TrcSubModule : public T
{
public:
    TrcSubModule(quint16 localShift, quint16 localMask, const MregBusInterface &bus)
        : T(bus)
        , localShift(localShift)
        , localMask(localMask)
    {}

    void onParentBaseAddrUpdate(quint16 phyTrigBaseAddr) {
        T::setBaseAddr(phyTrigBaseAddr+localShift, localMask);
    }
private:
    quint16 localShift;
    quint16 localMask;
};

PhysicalTriggerModule::PhysicalTriggerModule(const MregBusInterface &bus) :
    BaseDeviceModule(bus),
    rcTrigMod(new TrcSubModule<RcTrigModule>(REG_RC_CORE_BASE, REG_RC_CORE_MAKS, bus)),
    spillMod(new TrcSubModule<SpillControlModule>(REG_SPILL_BASE, REG_SPILL_MAKS, bus)),
    rcCountersMods({
        {DevCountersTrcIn,
         new TrcSubModule<RcCountersModule>(REG_INPUT_COUNTERS_BASE, REG_COUNTERS_MAKS, bus)},
        {DevCountersTrcBa,
         new TrcSubModule<RcCountersModule>(REG_BA_COUNTERS_BASE, REG_COUNTERS_MAKS, bus)},
        {DevCountersTrcReduc,
         new TrcSubModule<RcCountersModule>(REG_REDUC_COUNTERS_BASE, REG_COUNTERS_MAKS, bus)},
    })
{}

PhysicalTriggerModule::~PhysicalTriggerModule()
{
    delete rcTrigMod;
    delete spillMod;
    for(auto mod : rcCountersMods)
        delete mod;
}

QList<AFI_SDB_ID> PhysicalTriggerModule::getSupportedSdbIds()
{
    return {AFI_SDB_TRC_SYSTEM};
}

void PhysicalTriggerModule::post_init()
{
    BaseDeviceModule::post_init();

    rcTrigMod->onParentBaseAddrUpdate(baseAddr());
    spillMod->onParentBaseAddrUpdate(baseAddr());
    for(auto mod : rcCountersMods)
        mod->onParentBaseAddrUpdate(baseAddr());
}

void PhysicalTriggerModule::softClear()
{
    RegOpVector r;
    r.RegWrite16(REG_CORE_CTRL, getControlReg() | REG_CSR_BIT_CNT_RESET);
    r.RegWrite16(REG_CORE_CTRL, getControlReg());
    regOpExecRebased(r);
}

bool PhysicalTriggerModule::writeConfig(const PhysicalTriggerConfig &cfg)
{
    RegOpVector localCommands;

    oscilogramDelayed = cfg.oscDelayed;
    localCommands.RegWrite16(REG_CORE_CTRL, getControlReg());
    localCommands.RegWrite16(REG_MATCH_WIN, cfg.matchWin);
    localCommands.RegWrite16(REG_TRIG_DELAY, cfg.trigDelay);

    auto &ct = cfg.calibTrigger;

    localCommands.RegWrite32(REG32_CAL_TRIG_OFFSET, ct.offset);
    localCommands.RegWrite32(REG32_CAL_TRIG_PERIOD, ct.period);
    localCommands.RegWrite16(REG_CAL_TRIG_COUNT, ct.count);


    localCommands.RegWrite32(REG32_CH_EN, 0xFFFF | cfg.startChMask
                         | cfg.readChMask);
    localCommands.RegWrite32(REG32_START_CH_MASK, cfg.startChMask);
    localCommands.RegWrite32(REG32_READ_CH_MASK, cfg.readChMask);

    localCommands.RegWrite16(REG_EVENT_MEM_BASE + REG_CH_WF_LATENCY,
                      cfg.chWfLatency);
    localCommands.RegWrite16(REG_EVENT_MEM_BASE + REG_CH_WF_READ_POINTS,
                      cfg.chWfReadPoints);

    for(quint16 ch=0; ch<cfg.chCfg.size(); ++ch) {
        quint16 addr = REG_CH_BASE+ch*REG64_CH_SHIFT;
        auto reg = cfg.chCfg[ch].reg;
        reg.bit.after_protect_time = std::min(cfg.trigDelay, reg.bit.after_protect_time);
        localCommands.RegWrite64(addr, reg.all);

    }

    bool ok = true;
    ok &= rcTrigMod->writeConfig(cfg.trig);
    ok &= spillMod->writeConfig(cfg.spill);
    ok &= regOpExecRebased(localCommands);
    ok &= rcTrigMod->writeConfig(cfg.trig);
    ok &= spillMod->writeConfig(cfg.spill);

    return ok;
}

PhysicalTriggerBuildParams PhysicalTriggerModule::getBuildParam()
{
    quint16 v;
    regRead16(REG_CH_NUM, &v);

    PhysicalTriggerBuildParams st;
    st.phyChNum = v & 0xFF;
    st.auxChNum = (v>>8) & 0xFF;

    return st;
}

bool PhysicalTriggerModule::readStatus(PhysicalTriggerStatus *status)
{
    if (!setCountersLock(true))
        return {};
    // rcTrigMod has no status yet
    if (!spillMod->readStatus(&status->spill))
        return {};
    for (auto it = rcCountersMods.begin(); it != rcCountersMods.end(); ++it) {
        const auto &key = it.key();
        auto &map = status->rcCounters[key];
        if (!rcCountersMods[key]->readStatus(&map))
            return {};
    }
    if (!setCountersLock(false))
        return {};
    return true;
}

bool PhysicalTriggerModule::setCountersLock(bool lock)
{
    RegOpVector r;
    r.RegWrite16(REG_CORE_CTRL, getControlReg() | (lock ? REG_CSR_BIT_CNT_LOCK : 0));
    return regOpExecRebased(r);
}

quint16 PhysicalTriggerModule::getControlReg() const
{
    return oscilogramDelayed ? REG_CSR_BIT_OCS_DELAYED : 0;

}
