#include "WaveRecModule.h"

#include "CalResult.h"
#include "base/fw_version.h"
#include "WaveRec/WaveRecConfig.h"
#include "WaveRec/WaveRecStatus.h"
#include "base/DeviceId.h"
#include "device-modules/MregBusInterface.h"
#include "regio/regop.h"
#include "sdb/AfiSdbId.h"

namespace {
// ADC Core Registers
enum {
    // Ver 2.0
    REG_ID                      = 0x00,
    REG_COTROL                  = 0x02,
    REG_LOCK                    = 0x03,
    REG_LATENCY                 = 0x04,
    REG_MATCH_WIN               = 0x05,
    REG_PREWIN                  = 0x06,
    REG_WIN                     = 0x07,
    REG_MODE                    = 0x08,
    REG_CH_EN                   = 0x10,
    REG_SIGNAL_POLARITY         = 0x18,
    REG_DIG_ZS_EN               = 0x20,
    REG_DIG_ZS_POLARITY         = 0x28,
    REG_BUILD_PARAMS            = 0x30,
    REG_DIG_ZS_THRESHOLD        = 0x80,

    // Ver 1.0
    REG_ADC_LAT                 = 0x00,
    REG_ADC_MATCH_WIN           = 0x01,
    REG_ADC_PREWIN              = 0x02,
    REG_ADC_WIN                 = 0x03, // Samples to readout
    REG_ADC_MEM_DEPTH           = 0x04, // in kilosamples
    REG_ADC_MODE                = 0x05, // 1:0 - mode (1-norm,2-baseline)
    REG_ADC_GAIN_CRTL           = 0x06, // 7:0 - for ch pairs
    REG64_ADC_CH_EN             = 0x08,

    REG_ADC_BUILD_PARAM_1       = 0x10, // [15:8] - ADC Bits
    REG_ADC_BUILD_PARAM_2       = 0x11, // [7:0] - ADC chips number; [15:8] - ADC chip channel number
    REG_ADC_BUILD_PARAM_3       = 0x12, // [7:0] - log2(max ADC latency); [15:8] - log2(max ADC match window)
//    REG_ADC_BUILD_PARAM_4   = 0x13, // reserved

    REG_SPARSE_CTRL             = 0x38, // since v2.1
    REG_SPARSE_READ_CELL_NUM    = 0x39, // since v2.1
    REG_SPARSE_PERIOD           = 0x3A, // since v2.1

    REG64_ADC_CH_HFIFO          = 0x40, // bit# = ch#, RO
    REG64_ADC_CH_MFIFO          = 0x48, // bit# = ch#, RO
    REG64_ADC_CH_EFIFO          = 0x50, // bit# = ch#, RO

    REG_ADC_CH_DATA_BASE        = 0x80, // RO, ADC Channel #0 Data
};

enum {
    REG_ADC_MODE_NORM = 1,
    REG_ADC_MODE_BASELINE = 2,
};
enum {
    REG_ADC_MODE_V2_DISABLE_ZS = 1,
    REG_ADC_MODE_V2_ANLG_ZS = 2,
    REG_ADC_MODE_V2_DIG_ZS = 4,
};

} // namespace

WaveRecModule::WaveRecModule(const MregBusInterface &bus)
    : BaseDeviceModule(bus)
{
    if (!isAvailable()) {
        if (getDeviceId() == DEVICE_ID_PHOS_FEC)
            setBaseAddr(0x0300, 0x00FF);
    }
}

QList<AFI_SDB_ID> WaveRecModule::getSupportedSdbIds()
{
    return {AFI_SDB_ADC_Core, AFI_SDB_WAVE_REC};
}

bool WaveRecModule::writeConfig(const WaveRecConfig &config)
{
    RegOpVector r;

    r.RegWrite16(isV2() ? REG_LATENCY : REG_ADC_LAT, config.latency); // Set number of points saved after incoming trigger
    r.RegWrite16(isV2() ? REG_MATCH_WIN : REG_ADC_MATCH_WIN, config.matchWin);
    r.RegWrite16(isV2() ? REG_PREWIN : REG_ADC_PREWIN, config.adcPreWin);
    r.RegWrite16(isV2() ? REG_WIN : REG_ADC_WIN, config.readCellNumber);
    quint16 zsMode;
    switch (config.zsMode) {
    case WaveRecConfig::ZsAnlg:
        zsMode = isV2() ? REG_ADC_MODE_V2_ANLG_ZS : REG_ADC_MODE_NORM;
        break;
    case WaveRecConfig::ZsDig:
        zsMode = isV2() ? REG_ADC_MODE_V2_DIG_ZS : REG_ADC_MODE_BASELINE;
        break;
    default:
        zsMode = isV2() ? REG_ADC_MODE_V2_DISABLE_ZS : REG_ADC_MODE_BASELINE;
        break;
    }
    r.RegWrite16(isV2() ? REG_MODE : REG_ADC_MODE, zsMode);

    {
        quint64 chEnMaskAdc = 0;
        for (auto it = config.chAdcEn.begin(); it != config.chAdcEn.end(); ++it) {
            auto ch = it.key();
            bool en = it.value();
            if (en)
                chEnMaskAdc |= quint64(1) << ch;
        }
        r.RegWrite64(isV2() ? REG_CH_EN : REG64_ADC_CH_EN, chEnMaskAdc);
    }
    if(isV2() && buildParams) {
        quint64 signalInv = 0;
        quint64 chZsEn = 0;
        quint64 chZsRising = 0;
        for(auto ch=0; ch<buildParams->adcNCh; ++ch) {
            if(config.chSignalInvert.value(ch, false))
                signalInv |= 1<<ch;
            if(config.chZsEn.value(ch, false))
                chZsEn |= 1<<ch;
            if(!config.chZsRising.value(ch, false))
                chZsRising |= 1<<ch;
            r.RegWrite16(REG_DIG_ZS_THRESHOLD+ch, config.chZsThr.value(ch));
        }
        r.RegWrite64(REG_SIGNAL_POLARITY, signalInv);
        r.RegWrite64(REG_DIG_ZS_EN, chZsEn);
        r.RegWrite64(REG_DIG_ZS_POLARITY, chZsRising);
    }
    static const fw_version_t sparse_ver(2,1);
    if(getVersion() >= sparse_ver) {
        if(!config.sparse.en)
            r.RegWrite16(REG_SPARSE_CTRL, 0x0);
        r.RegWrite16(REG_SPARSE_PERIOD, config.sparse.period);
        r.RegWrite16(REG_SPARSE_READ_CELL_NUM, config.sparse.readCellNum);
        if(config.sparse.en)
            r.RegWrite16(REG_SPARSE_CTRL, 0x1);
    }
    return regOpExecRebased(r);
}

bool WaveRecModule::readStatus(WaveRecStatus *result)
{
    WaveRecStatus status;
    RegOpVector r;
    r.RegRead64(REG64_ADC_CH_HFIFO, &status.hfifo);
    r.RegRead64(REG64_ADC_CH_MFIFO, &status.mfifo);
    r.RegRead64(REG64_ADC_CH_EFIFO, &status.efifo);
    if (!regOpExecRebased(r))
        return {};

    if (result)
        *result = status;
    return true;
}

std::optional<WaveRecStatus> WaveRecModule::readStatus()
{
    WaveRecStatus status;
    if (!readStatus(&status))
        return {};
    return status;
}

QMap<int, QVector<int> > WaveRecModule::getAllChannelData(int repeat)
{
    if(!buildParams || isV2())
        return {};
    RegOpVector r;
    QMap<int, QVector<quint16>> rawdata;
    for (int ch = 0; ch < buildParams->adcNCh; ch++) {
        auto &samples = rawdata[ch];
        samples.resize(repeat);
        for (int k = 0; k < repeat; k++)
            r.RegRead16(REG_ADC_CH_DATA_BASE + ch, &samples[k]);
    }
    if (!regOpExecRebased(r))
        return {};
    QMap<int, QVector<int> > result;
    for (auto it = rawdata.cbegin(); it != rawdata.cend(); ++it) {
        auto ch = it.key();
        const auto &samples = it.value();
        for (auto sample : samples)
            result[ch].push_back(static_cast<qint16>(sample));
    }
    return result;
}

bool WaveRecModule::readBaseline(MultichannelBaselineStats *result, int repeat)
{
    const auto &allChannelData = getAllChannelData(repeat);
    if (allChannelData.empty())
        return {};
    auto stats = MultichannelBaselineStats(allChannelData);
    if (result)
        *result = stats;
    return true;
}

RegOpVector WaveRecModule::setAdcChEnRegOp(quint64 mask)
{
    RegOpVector r;
    r.RegWrite64(baseAddr()+isV2() ? REG_CH_EN : REG64_ADC_CH_EN, mask);
    return r;
}

RegOpVector WaveRecModule::setModeRegOp(bool zsEn)
{
    RegOpVector r;
    const auto &mode = zsEn ? REG_ADC_MODE_NORM : REG_ADC_MODE_BASELINE;
    r.RegWrite16(baseAddr()+(isV2() ? REG_MODE : REG_ADC_MODE), mode);
    return r;
}

RegOpVector WaveRecModule::setLat(quint16 lat)
{
    RegOpVector r;
    r.RegWrite16(baseAddr()+(isV2() ? REG_LATENCY : REG_ADC_LAT), lat);
    return r;
}

RegOpVector WaveRecModule::setMatchWin(quint16 matchWin)
{
    RegOpVector r;

    r.RegWrite16(baseAddr()+(isV2() ? REG_MATCH_WIN : REG_ADC_MATCH_WIN), matchWin);
    return r;
}

RegOpVector WaveRecModule::setPreWin(quint16 preWin)
{
    RegOpVector r;
    r.RegWrite16(baseAddr()+(isV2() ? REG_PREWIN : REG_ADC_PREWIN), preWin);
    return r;
}

RegOpVector WaveRecModule::setAdcWin(quint16 adcWin)
{
    RegOpVector r;
    r.RegWrite16(baseAddr()+(isV2() ? REG_WIN : REG_ADC_WIN), adcWin);
    return r;
}

bool WaveRecModule::readBuildParams()
{
    buildParams.reset();
    std::array<quint16, 8> val;
    RegOpVector r;
    if(isV2()) {
        r.RegRead16(REG_BUILD_PARAMS, &val[0]);
        r.RegRead16(REG_BUILD_PARAMS+1, &val[1]);
        r.RegRead16(REG_BUILD_PARAMS+2, &val[2]);
    } else {
        r.RegRead16(REG_ADC_BUILD_PARAM_1, &val[0]);
        r.RegRead16(REG_ADC_BUILD_PARAM_2, &val[1]);
        r.RegRead16(REG_ADC_BUILD_PARAM_3, &val[2]);
    }

    if (!regOpExecRebased(r))
        return {};

    AdcBuildParams bp;
    bp.adcBits = val[0] >> 8;

    bp.adcChips = val[1] & 0xFF;
    bp.adcChPerChip = val[1] >> 8;
    bp.adcNCh = bp.adcChips * bp.adcChPerChip;

    bp.adcMaxLat = 1 << (val[2] & 0xFF);
    bp.adcMaxMatchWin = 1 << (val[2] >> 8);

    if (getDeviceId() == DEVICE_ID_TQDC16VS_E) {
        // remove it after introducing normal TADC module to TQDC16VS
        bp.adcBits = 14;
        bp.adcChips = 2;
        bp.adcChPerChip = 8;
        bp.adcNCh = bp.adcChips * bp.adcChPerChip;
    }
    buildParams = bp;
    return true;
}

bool WaveRecModule::isV2() const
{
    return majorVersion() >= 2;
}
