#include "ClockDividerReset.h"

#include "ModularMregDevice.h"
#include "ClockDividerControl/ClockDividerControlModule.h"
#include "DeserializersControl/DeserializersControlModule.h"

#include <QDeadlineTimer>
#include <QDebug>

ClockDividerReset::ClockDividerReset(ModularMregDevice *parent)
    : dev (parent)
{}

void ClockDividerReset::init()
{
    clkCtrl = dev->getFirstModule<ClockDividerControlModule>();
    desCtrl = dev->getFirstModule<DeserializersControlModule>();

    if (!clkCtrl || !desCtrl) {
        deinit();
        return;
    }
}

void ClockDividerReset::deinit()
{
    clkCtrl = nullptr;
    desCtrl = nullptr;
}

void ClockDividerReset::syncClock()
{
    const constexpr static auto funcName = __func__;
    auto log = [&](const QString &msg, bool ok=true)
    {
        auto dbg = ok ? qInfo() : qWarning();
        dbg << QString("[%1] %2: %3")
                       .arg(dev->getDeviceIndex().getIdent())
                       .arg(funcName)
                   .arg(msg);
    };
    if (!clkCtrl || !desCtrl) {
        log("skip, missing modules");
        return;
    }

    auto st = clkCtrl->readStatus();
    if(!st) {
        log("skip, can't read status", false);
        return;
    }
    if(!st->needSync()) {
        log("skip, all Ok");
        return;
    }

    QDeadlineTimer deadline(1500);
    if(clkCtrl->setResetEn(true)) {
        forever {
            QThread::msleep(50);
            auto st = clkCtrl->readStatus();
            if(!st) {
                log("failed, can't read status", false);
                break;
            }
            if(!st->needSync()) {
                log("done");
                break;
            }
            if(deadline.hasExpired()) {
                log("time expired");
                break;
            }
        }
    }
    clkCtrl->setResetEn(false);

    desCtrl->ResetDeser();
    desCtrl->ResetAdcStatus();
    QThread::msleep(1);
}
