﻿/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ClockControlWidget.h"
#include "base/DeviceIndexEnabled.h"
#include "ui_ClockControlWidget.h"

#include <QComboBox>
#include <QDebug>
#include <QLabel>
#include <QSizePolicy>

#include "base/gui_utils.h"
#include "device-modules/ClockControl/ClockControlType.h"

ClockControlWidget::ClockControlWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::ClockControlWidget)
{
    ui->setupUi(this);
    updateControlGui();
}

ClockControlWidget::~ClockControlWidget()
{
    delete ui;
}

void ClockControlWidget::setConfig(const ClockControlConfig &newConfig)
{
    if (newConfig == cfg)
        return;
    cfg = newConfig;

    updateControlGui();
    updateStatus();
}

void ClockControlWidget::setClockControlBp(const DeviceIndex &index, const ClockBuildParams &bp)
{
    if(devBp.contains(index) && devBp[index] == bp)
        return;
    devBp[index] = bp;
    updateComboBoxList();
    updateControlGui();
}

void ClockControlWidget::setClockControlStatus(const DeviceIndex &index, const std::optional<ClockStatus> &st)
{
    if(!devSet.contains(index) || !st)
        return;
    auto it = devStatus.find(index);
    if(it != devStatus.end() && *it == *st)
        return;
    devStatus[index] = *st;
    updateStatus();
}

void ClockControlWidget::setDeviceList(const DeviceIndexEnabledMap &map)
{
    auto set = map.getEnabled();
    if(devSet == set)
        return;
    devSet = set;
    for(auto i : devStatus.keys())
        if(!devSet.contains(i))
            devStatus.remove(i);
    updateComboBoxList();
    updateStatus();
}

void ClockControlWidget::updateControlGui()
{
    RecursiveWidgetSignalBlocker blocker(this);

    auto i = ui->comboBoxClockSource->findData(cfg.source);
    ui->comboBoxClockSource->setCurrentIndex(i);
    ui->checkBoxPllBypass->setChecked(cfg.pll_bypass);
}

void ClockControlWidget::updateComboBoxList()
{
    auto bp = getBuildParams().value_or(0);
    auto cb = ui->comboBoxClockSource;
    QSignalBlocker blocker(cb);
    cb->clear();
    cb->setEnabled(bp);
    if(bp) {
        for(auto i=0; i<16; ++i) {
            auto type = static_cast<ClockControlType>(bp & (1<<i));
            if(type == 0)
                continue;
            auto name = clockName(type);
            cb->addItem(name, type);
        }
    } else {
        cb->addItem("No build params");
    }
    updateControlGui();
}

void ClockControlWidget::updateStatus() const
{
    QStringList missedDev;
    DeviceIndexSet okDev;
    for(auto i : devSet) {
        if(!devStatus.contains(i)) {
            missedDev << QString("%1: missed").arg(i.getIdent());
            continue;
        }
        QStringList err;
        auto st = devStatus[i];
        if(st.pll_bypass != cfg.pll_bypass) {
            err << "pll mismatch";
        }
        if(cfg.source != ClockControlAuto && st.selected_clock != cfg.source) {
            err << "source mismatch";
        }

        if(err.isEmpty())
            okDev << i;
        missedDev << QString("%1: %2").arg(i.getIdent()).arg(err.isEmpty() ? "ok": err.join(", "));
    }

    ui->labelStatus->setText(okDev == devSet ? "ok" : missedDev.join('\n'));
}

std::optional<quint16> ClockControlWidget::getBuildParams() const
{
    std::optional<quint16> bp;
    for(auto i : devSet) {
        auto it = devBp.constFind(i);
        if(it == devBp.cend())
            return {};
        bp = bp.value_or(0xFFFF) & it->clockCapability;
    }
    return bp;
}

void ClockControlWidget::on_comboBoxClockSource_currentIndexChanged(int)
{
    cfg.source = static_cast<ClockControlType>(ui->comboBoxClockSource->currentData().toInt());
    emit configChanged(cfg);
    updateStatus();
}


void ClockControlWidget::on_checkBoxPllBypass_toggled(bool bypass)
{
    cfg.pll_bypass = bypass;
    emit configChanged(cfg);
    updateStatus();
}

