﻿#include "DeviceChTreeView.h"

#include <QHeaderView>
#include <QSettings>

#include "DeviceChModel.h"
#include "TreeColumns.h"
#include "SpinBoxDelegate.h"

#include "base/DeviceIndexEnabled.h"
#include "sdb/AfiSdbId.h"
#include "ModularDeviceStatus.h"

const QString EXPAND_DEVICE_MAP = "expand_device_map";

DeviceChTreeView::DeviceChTreeView(QWidget *parent)
    : QTreeView(parent)
    , anlgDelegate(new AnlgThrSpinBoxDelegate(this))
    , digDelegate(new DigThrSpinBoxDelegate(this))
{
    connect(this, &QTreeView::expanded, this, &DeviceChTreeView::expanded);
    connect(this, &QTreeView::collapsed, this, &DeviceChTreeView::collapsed);
    header()->setStretchLastSection(false);
    header()->setSectionResizeMode(QHeaderView::ResizeToContents);
}

void DeviceChTreeView::setTreeModel(DeviceChModel *treeModel)
{
    QTreeView::setModel(treeModel);
    setItemDelegateForColumn(static_cast<int>(TreeColumns::COL_ANLG_THR), anlgDelegate);
    setItemDelegateForColumn(static_cast<int>(TreeColumns::COL_DIG_TRIG_LVL), digDelegate);
    setItemDelegateForColumn(static_cast<int>(TreeColumns::COL_ZS_DIG_THR), digDelegate);

    connect(treeModel, &QAbstractItemModel::modelReset, this, &DeviceChTreeView::onModelReset);
    updateColVis();
}

void DeviceChTreeView::saveAppGuiSettings(QSettings &s) const
{
    s.beginGroup(EXPAND_DEVICE_MAP);
    QStringList devList;
    for(DeviceIndex index : expand_device_map.keys()){
        s.setValue(index.toString(), expand_device_map[index]);
        devList << index.toString();
    }
    s.setValue("devList", devList);
    s.endGroup();
}

void DeviceChTreeView::loadAppGuiSettings(const QSettings &s)
{
    expand_device_map.clear();
    const auto &devList = s.value(EXPAND_DEVICE_MAP+"/devList").toStringList();
    for(const QString &key : devList){
        DeviceIndex index = DeviceIndex::fromString(key);
        if(!index.isNull())
            expand_device_map[index] = s.value(EXPAND_DEVICE_MAP+"/"+key).toBool();
    }
}

void DeviceChTreeView::setseparateChEn(bool showSepCols)
{
    if(separateChEn == showSepCols)
        return;
    separateChEn = showSepCols;
    updateColVis();
}

void DeviceChTreeView::expanded(const QModelIndex &index)
{
   if(!getDeviceIndex(index).isNull()){
        expand_device_map[getDeviceIndex(index)] = true;
   }
}

void DeviceChTreeView::collapsed(const QModelIndex &index)
{
    if(!getDeviceIndex(index).isNull()){
        expand_device_map[getDeviceIndex(index)] = false;
    }
}

void DeviceChTreeView::onModelReset()
{
    expandTopItem();
    addDelegate();
    expandAll();
    updateColVis();
}

void DeviceChTreeView::expandTopItem()
{
    auto model = qobject_cast<const DeviceChModel*>(this->model());
    auto top = model->index(0,0);
    expand(top);

    for(DeviceIndex index : expand_device_map.keys()){
        auto row = model->getDevRow(index);
        if(expand_device_map[index])
            expand(model->index(row, 0, top));
        else
            collapse(model->index(row, 0, top));
    }

    for(int row=0; row<this->model()->rowCount(); ++row){
        DeviceIndex index = static_cast<DeviceChModelItem*>(model->index(row, 0, top).internalPointer())->dev;
        if(!expand_device_map.contains(index))
            expand_device_map.remove(index);
    }
}

void DeviceChTreeView::deviceStatusUpdated(const ModularDeviceStatus &status)
{
    const auto &index = status.getDeviceIndex();
    if(!devSet.contains(index))
        return;

    auto it = devStMap.find(index);

    bool sdbChange;
    if(it == devStMap.end()) {
        sdbChange = true;
        devStMap.insert(index, status);
    } else {
        sdbChange = it->sdb != status.sdb;
        *it = status;
    }

    if(!sdbChange)
        return;

    updateColVis();
}

void DeviceChTreeView::setDeviceList(const DeviceIndexEnabledMap &map)
{
    const auto &newSet = map.getEnabled();
    if(devSet == newSet)
        return;
    devSet = newSet;
    updateColVis();
}

void DeviceChTreeView::resizeEvent(QResizeEvent *event)
{
    if(auto m = model()) {
        m->layoutChanged();
    }
    QTreeView::resizeEvent(event);
}

DeviceIndex DeviceChTreeView::getDeviceIndex(const QModelIndex &index) const
{
    return static_cast<DeviceChModelItem*>(index.internalPointer())->dev;
}

void DeviceChTreeView::addDelegate(const QModelIndex &parent)
{
    const auto cols = {static_cast<int>(TreeColumns::COL_ANLG_THR),
                       static_cast<int>(TreeColumns::COL_DIG_TRIG_LVL),
                       static_cast<int>(TreeColumns::COL_ZS_DIG_THR)};
    if(parent.isValid()) {
        for(auto col : cols) {
            const auto &i = parent.siblingAtColumn(col);
            if(!isPersistentEditorOpen(i))
                openPersistentEditor(i);
        }
    }

    const auto &m = model();
    int r1 = m->rowCount(parent);
    for(int i=0; i<r1; ++i) {
        auto chIndex = m->index(i, 0, parent);
        addDelegate(chIndex);
    }
}

void setTreeColumnHidden(QTreeView *v, TreeColumns col, std::optional<bool> hasCol)
{
    v->setColumnHidden(static_cast<int>(col), !hasCol.value_or(false));
}

void DeviceChTreeView::updateColVis()
{
    static const QMap<AFI_SDB_ID, QList<TreeColumns> > sdbToColMap
            ({
                 {AFI_SDB_HPTDC_RegIO_JTAG, {TreeColumns::COL_TDC_EN}},
                 {AFI_SDB_AD5328_x2_SPI, {TreeColumns::COL_TRIG_EN,
                                          TreeColumns::COL_ANLG_THR}},
                 {AFI_SDB_WAVE_REC, {TreeColumns::COL_ADC_EN,
                                     TreeColumns::COL_DIG_TRIG_EN,
                                     TreeColumns::COL_DIG_TRIG_EDGE,
                                     TreeColumns::COL_DIG_TRIG_LVL,
                                     TreeColumns::COL_ZS_DIG_EN,
                                     TreeColumns::COL_ZS_DIG_EDGE,
                                     TreeColumns::COL_ZS_DIG_THR}},
                 {AFI_SDB_Input_Hit_Core, {TreeColumns::COL_INVERT,
                                           TreeColumns::COL_RATE,
                                           TreeColumns::COL_LEVEL}},
             });
    QMap<AFI_SDB_ID, std::optional<bool> > hasSdbMod;

    for(auto sdbId : sdbToColMap.keys()) {
        auto &hasMod = hasSdbMod[sdbId];
        for(const auto &status : devStMap) {
            if(!hasMod || !hasMod.value())
                hasMod = !status.sdb.filterDevices({sdbId}).isEmpty();
        }
    }

    for(auto it=sdbToColMap.cbegin(); it!=sdbToColMap.cend(); ++it) {
        for(auto col : *it) {
            auto hasCol = hasSdbMod[it.key()];
            if(!separateChEn && (col == TreeColumns::COL_TDC_EN ||
                                 col == TreeColumns::COL_ADC_EN))
                hasCol = false;
            setTreeColumnHidden(this, col, hasCol);
        }
    }
}
