#ifndef FLASHDEV_H
#define FLASHDEV_H

#include <cstdint>
#include <optional>
#include <string>
#include <vector>

#include <QObject>
#include <QByteArray>

#include "FlashDevTask.h"
#include "FlashIdData.h"

enum FlashCommand : uint16_t;

class FlashDev : public QObject
{
    Q_OBJECT
public:
    enum class State {Init, Ready, Write, Verify};

    explicit FlashDev(QObject *parent = nullptr);

    void configureFlash(FlashIdData *result, bool supportAddr4bMode);
    bool isConfigured() const { return state != State::Init; }

    // async interface
    bool abortTask();
    bool isReady() const;
    std::optional<FlashDevTask> getTasks() const;
    std::optional<FlashDevTask> getFinishedTasks() const;
    bool runTasks();
    bool submitTask(const FlashDevTask &newTask);

    uint32_t getImageOffset(FlashImageType imageType) const;
    uint32_t getImageSizeLimit(FlashImageType imageType) const;
    FlashIdData getFlashIdData() const { return flashIdData; }
    bool isAddr4bMode() const { return useAddr4bMode; }
    void setSingleCheckMode(bool v) { singleCheckMode = v; }

signals:
    void taskFinished(const FlashDevTask &task);

protected:
    enum class CmdStatus { Error, Ok, Timeout };
    QString cmdStatusStr(CmdStatus value);
    QString commandStr(FlashCommand value);
    bool eraseFlashStep();
    bool writeFlashStep();
    bool verifyFlashStep();

    void dumpVerify(const std::vector<uint16_t> &original, const std::vector<uint16_t> &readback) const;
    void reverseByteOrder(std::vector<uint16_t> &dataVec);
    CmdStatus waitCmdExecution(int timeoutMs = 0, bool extraWaitCheck = false);
    CmdStatus execCmd(FlashCommand cmd, int timeoutMs = 0, bool extraWaitCheck = false);
    bool singleSectorErase(uint32_t pos);
    bool singleBlockWrite(uint32_t pos, std::vector<uint16_t> blockData);
    bool singleSectorProgram(const uint32_t pos, const std::vector<uint16_t> &sectorData);
    std::vector<uint16_t> singleBlockRead(uint32_t pos);
    std::vector<uint16_t> singleBlockRead(uint32_t pos, int retries, int rightCnt);
    bool bulkErase();

    void readStat();
    bool read_rdsr(uint16_t *result);
    virtual bool devReadReg(int regAddr, uint16_t *result) = 0;
    virtual bool devWriteReg(int regAddr, uint16_t data) = 0;
    virtual bool devWriteReg32(int regAddr, uint32_t data) = 0;
    virtual bool devBlkWriteReg(uint32_t count, const uint32_t *addr, uint16_t *data) = 0;
    virtual bool devBlkReadReg(uint32_t count, const uint32_t *addr, uint16_t *data) = 0;

    virtual QString getPrefix() const = 0;
    bool updateFlashIdData();
    uint32_t getFlashSize() const;
    bool setState(State newState);

    State state = State::Init;
    std::optional<FlashDevTask> task;
    std::optional<FlashDevTask> finishedTask;
    int cmdPerformed = 0;
    FlashIdData flashIdData;
    int attemptFactor = 0;
    bool singleCheckMode = false;
    bool useAddr4bMode = false;
};

#endif // FLASHDEV_H
