//
//    Copyright 2015 AFI Electronics
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "AbstractRawReader.h"

#include <QDebug>
#include <QTimer>
#include <QSemaphore>

namespace {
const int DATA_READ_PERIOD_MS = 1000;
const int STATS_SEND_PERIOD_MS = 1000;
}

AbstractRawReader::AbstractRawReader(QSemaphore * freeSpace, QObject *parent)
    : QObject(parent),
      streamPos(0),
      syncLost(false),
      freeSpaceSem(freeSpace),
      listenEnabled(true)
{
    QTimer *timer1 = new QTimer(this);
    connect(timer1, SIGNAL(timeout()), SLOT(sendStats()));
    timer1->start(STATS_SEND_PERIOD_MS);
}

AbstractRawReader::~AbstractRawReader()
{

}

QString AbstractRawReader::rawDataErrorString(qint64 pos, const QString &text)
{
    return QString("[%1]: %2").arg(streamPos + pos, 8, 16, QChar('0')).toUpper().arg(text);
}

void AbstractRawReader::sendStats()
{
    if (sentStats == stats) return;
    emit textTableStatsUpdated(stats);
    sentStats = stats;
}

void AbstractRawReader::dataAcquired(const QByteArray &d)
{
    data.append(d);
//    qInfo() <<"AbstractRawReader::dataAcquired cap" << data.capacity() << "size:"<< data.size();
    analyzeData();
}

void AbstractRawReader::releaseReadData(int offset)
{
//    Q_ASSERT(offset <= data.size());
    if (offset > data.size()) { // fail-safe
        qDebug() << "releaseReadData: offset " << offset << ", data.size" << data.size();
        offset = data.size();
    }
    data = data.right(data.size()-offset);
//    qInfo() <<"AbstractRawReader::releaseReadD cap" << data.capacity() << "size:"<< data.size();
    streamPos += offset;
    freeSpaceSem->release(offset);
}

void AbstractRawReader::reset()
{
    freeSpaceSem->release(data.size());
    data.clear();
    stats.clear();
    syncLost = true;
    streamPos = 0;
}

void AbstractRawReader::clearStat()
{
    stats.clear();
}

void AbstractRawReader::analyzeData()
{
    if(!listenEnabled){
        QTimer::singleShot(DATA_READ_PERIOD_MS, this, SLOT(analyzeData()));
        return;
    }

    restartEventPrint();

    int readerPos = 0;
    while (readerPos < data.size()) {
        int syncPos = findSync(readerPos);
        if (syncPos > 0) {
            stats.inc("Sync Dropped Bytes", syncPos-readerPos);
            readerPos = syncPos;
        }
        // decode one raw event
//        qInfo() << "Ev pos:" << streamPos + readerPos;
        int newReaderPos = readEvent(readerPos);
        if (newReaderPos < 0 || newReaderPos > data.size()) {
            rawDataErrorString(readerPos, QString("Invalid result of readEvent: %1, data size %2").arg(newReaderPos).arg(data.size()));
            newReaderPos = data.size();
        }
        if (newReaderPos == readerPos) break; // not enough data
        readerPos = newReaderPos;
    }
//    qInfo() << "readerPos:"<< readerPos;
    releaseReadData(readerPos);
}

void AbstractRawReader::listenStateUpdated(bool listen)
{
    listenEnabled = listen;
}
