/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ReadoutAppSignalRouter.h"

#include "AdcMpdTrigDialog/AdcMpdTrigDialog.h"
#include "AdcMpdTrigDialog/AdcMpdTrigModel.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"
#include "AsyncDeviceCommander.h"
#include "BaseConfigDispatcher.h"
#include "ClockControl/ClockControlDialog.h"
#include "CrosspointSwitchDilog/CrosspointSwitchDilog.h"
#include "DeviceStatusZmqPublisher/DeviceStatusZmqPublisher.h"
#include "FeLinkDialog/FeLinkStatusDialog.h"
#include "HptdcStatusDialog/HptdcStatusDialog.h"
#include "ModularMregDevice.h"
#include "NetworkPortWidget/NetworkPortConfigDialog.h"
#include "ReadoutAppCore.h"
#include "ReadoutAppMainWindow.h"
#include "RunLogic/RunLogicController.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "fsm/DeviceMgrFsm.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "mldiscover/DeviceDispatcher.h"
#include "mldiscover/LldpDiscovery.h"
#include "mstream-mreg-dev-global/MstreamMregDevGlobal.h"
#include "mvc/DeviceStatusTableViewFancy.h"
#include "remote-control-server/EvNumChecker.h"
#include "remote-control-server/RemoteControlServer.h"
#include "SparseDialog/SparseDialog.h"
#include "widgets/DeviceStatusDialog.h"
#include "zmq-lib/ZmqDialog.h"

void ReadoutAppSignalRouter::setDeviceIdFilter(const DeviceIdSet &filter)
{
    deviceIdFilter = filter;
}

void ReadoutAppSignalRouter::setDeviceMgrFsm(DeviceMgrFsm *p)
{
    deviceMgrFsm = p;
    setMainFsm(p);
}

void ReadoutAppSignalRouter::setReadoutAppMainWindow(ReadoutAppMainWindow *p)
{
    BaseSignalRouter::setPnpBaseMainWindow(p);

    readoutAppMainWindow = p;
    setAdcMpdTrigDialog(readoutAppMainWindow->adcMpdTrigDialog);
    setAdcSerDesCalDialog(readoutAppMainWindow->adcSerDesCalDialog);
    setClockControlDialog(readoutAppMainWindow->clockControlDialog);
    setCrosspointSwitchConfigDialog(readoutAppMainWindow->crosspointSwitchDilog);
    setDeviceStatusDialog(readoutAppMainWindow->deviceStatusDialog);
    setDiscoverDialog(readoutAppMainWindow->discoverDialog);
    //setLogDialog(readoutAppMainWindow->logDialog);
    setMStreamStatisticDialog(readoutAppMainWindow->mStreamStatisticDialog);
    setNetworkPortConfigDialog(readoutAppMainWindow->networkPortConfigDialog);
    setSparseDialog(readoutAppMainWindow->sparseDialog);
    setFeLinkStatusDialog(readoutAppMainWindow->feLinkStatusDialog);
    setHptdcStatusDialog(readoutAppMainWindow->hptdcStatusDialog);
    setDeviceStatusTableView(readoutAppMainWindow->getDeviceStatusTableView());
}

void ReadoutAppSignalRouter::setReadoutAppCore(ReadoutAppCore *p)
{
    BaseSignalRouter::setBaseAppCore(p);

    readoutAppCore = p;
    setDeviceCommander(readoutAppCore->deviceCommander);
    setDeviceDispatcher(readoutAppCore->deviceDispatcher);
    setEvNumChecker(readoutAppCore->evNumChecker);
//    setDeviceMgrFsm(readoutAppCore->getFsm()); // todo??: readoutAppCore stores BaseFsm not DeviceMgrFsm
}

void ReadoutAppSignalRouter::connectAll() const
{
    BaseSignalRouter::connectAll();

    connect_Core();
    connect_deviceCommander();
    connect_deviceDispatcher();
    connect_ConfigurationDispatcher();
    connect_EvNumChecker();

    if (discoverDialog) {
        LldpDiscovery::subscribe(discoverDialog, deviceIdFilter);
    }
    if (deviceStatusTableView && deviceStatusDialog) {
        connect(deviceStatusTableView, &DeviceStatusTableView::currentDeviceIndexChanged,
                deviceStatusDialog, &DeviceStatusDialog::selectDevice);
        connect(deviceStatusTableView, &DeviceStatusTableViewFancy::showDeviceStatus,
                deviceStatusDialog, &DeviceStatusDialog::showDevice);
    }
    // File menu
    if (readoutAppMainWindow && discoverDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionDiscoverDialog,
                discoverDialog, &QDialog::show);
    }

    // Tools menu
    if (readoutAppMainWindow && adcMpdTrigDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionAdcMpdTrigDialog,
                adcMpdTrigDialog, &QDialog::show);
    }
    if (readoutAppMainWindow && adcSerDesCalDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionAdcSerDesCalibration,
                adcSerDesCalDialog, &QDialog::show);
    }
    if (readoutAppMainWindow && networkPortConfigDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionNetworkPortConfig,
                networkPortConfigDialog, &QDialog::show);
    }
    if (readoutAppMainWindow && sparseDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionSparseDialog,
                sparseDialog, &QDialog::show);
    }
    if (clockControlDialog) {
        if(readoutAppMainWindow) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionClockControlConfig,
                clockControlDialog, &QDialog::show);
        }

        if (deviceDispatcher) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    clockControlDialog, &ClockControlDialog::setDeviceList);
        }
    }
    if (readoutAppMainWindow && crosspointSwitchDilog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionCpsConfig,
                crosspointSwitchDilog, &QDialog::show);
    }
    if (readoutAppMainWindow && deviceStatusDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionDeviceStatus,
                deviceStatusDialog, &QDialog::show);
    }
    if (readoutAppMainWindow) {
        if(auto d = readoutAppMainWindow->zmqDialog) {
            connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionDeviceStatusZmqConfig,
                    d, &QDialog::show);
        }
    }
    if (readoutAppMainWindow && hptdcStatusDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionHptdcStatus,
                hptdcStatusDialog, &QDialog::show);
    }
    if (readoutAppMainWindow && mStreamStatisticDialog) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionMstreamStatistics,
                mStreamStatisticDialog, &QDialog::show);
    }

    if (readoutAppCore) {
        connect(readoutAppCore, &ReadoutAppCore::runStateChanged,
                mStreamStatisticDialog, &MStreamStatisticDialog::set_run_state);
    }

    if (readoutAppMainWindow) {
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::set_run_state,
                mStreamStatisticDialog, &MStreamStatisticDialog::set_run_state); // TODO: core

        if (readoutAppMainWindow && feLinkStatusDialog) {
            connect(readoutAppMainWindow, &ReadoutAppMainWindow::actionFelinkStatus,
                    feLinkStatusDialog, &QDialog::show);
        }

        connect(readoutAppMainWindow, &ReadoutAppMainWindow::setStatus,
                deviceStatusTableView, &DeviceStatusTableView::setStatus);
        connect(readoutAppMainWindow, &ReadoutAppMainWindow::setMultiStatus,
                deviceStatusTableView, &DeviceStatusTableView::setMultiStatus);
    }
}

void ReadoutAppSignalRouter::connect_ConfigurationDispatcher() const
{
    if (!configDispatcher)
        return;

    if (readoutAppCore) {
        configDispatcher->subscribeCS(readoutAppCore);
    }

    if (readoutAppMainWindow) {
        configDispatcher->subscribeConfig(readoutAppMainWindow);
        configDispatcher->subscribeCS(readoutAppMainWindow);
        configDispatcher->subscribeAllowRemoteControl(readoutAppMainWindow);
        configDispatcher->subscribeErrOnTrigOnXOff(readoutAppMainWindow);
    }

    if (feLinkStatusDialog)
        configDispatcher->subscribeConfig(feLinkStatusDialog->getModel());

    if (adcMpdTrigDialog)
        configDispatcher->subscribeConfig(adcMpdTrigDialog);

    if(adcSerDesCalDialog)
        configDispatcher->subscribeSerDesDelay(adcSerDesCalDialog);

    if(remoteControlServer) {
        configDispatcher->subscribeAllowRemoteControl(remoteControlServer);
    }

    if (configurationManager)
        configDispatcher->subscribeCS(configurationManager);

    if (clockControlDialog)
        configDispatcher->subscribeClockControlConfig(clockControlDialog);

    if (crosspointSwitchDilog)
        configDispatcher->subscribeCrosspointSwitchConfig(crosspointSwitchDilog);

    if (networkPortConfigDialog)
        configDispatcher->subscribeNetworkPortConfig(networkPortConfigDialog);

    if (sparseDialog)
        configDispatcher->subscribeSparseParams(sparseDialog);

    if(auto zmq = readoutAppMainWindow->zmqDialog) {
        configDispatcher->subscribeZmqConfig(zmq);
    }
}

void ReadoutAppSignalRouter::connect_Core() const
{
    if (!readoutAppCore)
        return;
    auto runLogicController = readoutAppCore->runLogicController;
    if (runLogicController) {
        if (configDispatcher) {
            configDispatcher->subscribeRunLogicConfig(runLogicController);
        }
        if (deviceCommander) {
            runLogicController->setDeviceCommander(deviceCommander);
        }
        if (deviceDispatcher) {
            connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                    runLogicController, &RunLogicController::setDeviceList);
        }
    }
    if (auto stPub = readoutAppCore->deviceStatusZmqPublisher) {
        connect(readoutAppCore, &ReadoutAppCore::setStatus,
                stPub, &DeviceStatusZmqPublisher::setStatus);

        if(configDispatcher)
            configDispatcher->subscribeZmqConfig(stPub);
    }
}

void ReadoutAppSignalRouter::connect_FSM() const
{
    BaseSignalRouter::connect_FSM();

    if (!deviceMgrFsm)
        return;
    if (remoteControlServer) {
        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                remoteControlServer, &RemoteControlServer::onFsmStateChange);
    }
    if(deviceDispatcher)
        deviceDispatcher->subscribeDevices(deviceMgrFsm);

    if(configDispatcher)
        connect(configDispatcher, &BaseConfigDispatcher::configKeyChanged,
                deviceMgrFsm, &DeviceMgrFsm::setMainSetupKey);

    if(adcMpdTrigDialog)
        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                adcMpdTrigDialog, &AdcMpdTrigDialog::onFsmStateChange);

    if(adcSerDesCalDialog)
        connect(deviceMgrFsm, &BaseFsm::stateChanged,
                adcSerDesCalDialog, &AdcSerDesCalDialog::onFsmStateChange);
}

void ReadoutAppSignalRouter::connect_deviceCommander() const
{
    if (!deviceCommander)
        return;
    if (deviceDispatcher) {
        connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                deviceCommander, &AsyncDeviceCommander::setDeviceList);
    }
}

void ReadoutAppSignalRouter::connect_deviceDispatcher() const
{
    if (!deviceDispatcher)
        return;
    auto &lldpDiscovery = LldpDiscovery::instance();
    lldpDiscovery.setDeviceIdFilter(deviceIdFilter);
    lldpDiscovery.subscribe(deviceDispatcher);
    if (discoverDialog)
        lldpDiscovery.subscribe(discoverDialog);
    if (discoverDialog)
        deviceDispatcher->subscribeDevices(discoverDialog);
    if (deviceStatusTableView)
        deviceDispatcher->subscribeDevices(deviceStatusTableView);
    if (adcMpdTrigDialog)
        deviceDispatcher->subscribeDevices(adcMpdTrigDialog);
    if (adcSerDesCalDialog)
        deviceDispatcher->subscribeDevices(adcSerDesCalDialog);
    if (feLinkStatusDialog)
        deviceDispatcher->subscribeDevices(feLinkStatusDialog);
    if (hptdcStatusDialog)
        deviceDispatcher->subscribeDevices(hptdcStatusDialog);
    if (mStreamStatisticDialog)
        deviceDispatcher->subscribeDevices(mStreamStatisticDialog);
    if (sparseDialog)
        deviceDispatcher->subscribeDevices(sparseDialog);
}

void ReadoutAppSignalRouter::connect_EvNumChecker() const
{
    if(!evNumChecker)
        return;

    if(remoteControlServer) {
        connect(evNumChecker, &EvNumChecker::completedEvNumCheck,
                remoteControlServer, &RemoteControlServer::evNumCheckResponse);
        connect(remoteControlServer, &RemoteControlServer::evNumCheckRequest,
                evNumChecker, &EvNumChecker::onRemoteEvNumCheckRequest);
    }

    if(readoutAppCore) {
        connect(evNumChecker, &EvNumChecker::checkFail,
                readoutAppCore, &ReadoutAppCore::onEvNumCheckFail);
    }

    if (deviceDispatcher) {
        connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                evNumChecker, &EvNumChecker::setDeviceList);
    }
}

void ReadoutAppSignalRouter::connect_MainWindow() const
{
    BaseSignalRouter::connect_MainWindow();
    if(readoutAppCore && readoutAppMainWindow) {
        auto zmqPublisher = readoutAppCore->deviceStatusZmqPublisher;
        auto zmqDialog = readoutAppMainWindow->zmqDialog;

        if(zmqDialog && zmqPublisher) {
            connect(zmqPublisher, &ZmqPublisher::statusUpdated,
                    zmqDialog, &ZmqDialog::statusUpdated);
        }

        if (adcSerDesCalDialog && readoutAppCore)
            connect(readoutAppMainWindow, &ReadoutAppMainWindow::requestAdcClkDivSync,
                    readoutAppCore, &ReadoutAppCore::requestAdcClkDivSync);
    }
}

void ReadoutAppSignalRouter::connect_RemoteControlServer() const
{
    BaseSignalRouter::connect_RemoteControlServer();

    if (!remoteControlServer)
        return;

    if(readoutAppMainWindow)
        connect(remoteControlServer, &RemoteControlServer::connectionStatusUpdated,
                readoutAppMainWindow, &ReadoutAppMainWindow::onRemoteStatusUpdated);
}

void ReadoutAppSignalRouter::connect_device(const MstreamMregDevGlobal *devGlob, const ModularMregDevice *dev) const
{
    if (!devGlob || !dev)
        return;

    if (deviceStatusTableView) {
        connect(devGlob, &MstreamMregDevGlobal::stat_updated,
                deviceStatusTableView, &DeviceStatusTableView::stat_updated);
        connect(devGlob, &MstreamMregDevGlobal::stat_from_receiver_updated,
                deviceStatusTableView, &DeviceStatusTableView::stat_from_receiver_updated);
        connect(devGlob, &MstreamMregDevGlobal::mstreamConnectionStateChanged,
                deviceStatusTableView, &DeviceStatusTableView::mstreamConnectionStateChanged);
        connect(deviceStatusTableView, &DeviceStatusTableViewFancy::rebootDevice,
                dev, &ModularMregDevice::rebootDevice);
    }
    if (mStreamStatisticDialog) {
        connect(devGlob, &MstreamMregDevGlobal::stat_updated,
                mStreamStatisticDialog, &MStreamStatisticDialog::stat_updated);
        connect(devGlob, &MstreamMregDevGlobal::stat_from_receiver_updated,
                mStreamStatisticDialog, &MStreamStatisticDialog::stat_from_receiver_updated);
    }

    if(readoutAppMainWindow) {
        connect(dev, &ModularMregDevice::deviceSdbListUpdated,
                readoutAppMainWindow, &ReadoutAppMainWindow::setDeviceSdbList);
        connect(dev, &ModularMregDevice::adcDeserConfigChecked,
                readoutAppMainWindow, &ReadoutAppMainWindow::adcDeserConfigChecked);
        connect(dev, &ModularMregDevice::clockDividerStatusUpdated,
                readoutAppMainWindow, &ReadoutAppMainWindow::clockDividerStatusUpdated);
    }

    if (feLinkStatusDialog) {
        connect(feLinkStatusDialog->getModel(), &FeLinkTreeModel::phyReset,
                dev, &ModularMregDevice::feLinkPhyReset);
        connect(dev, &ModularMregDevice::feLinkBpUpdated,
                feLinkStatusDialog->getModel(), &FeLinkTreeModel::setFeLinkBp);
        connect(dev, &ModularMregDevice::feLinkStatusUpdated,
                feLinkStatusDialog, &FeLinkStatusDialog::setFeLinkStatus);
        connect(dev, &ModularMregDevice::deviceConnected,
                feLinkStatusDialog, &FeLinkStatusDialog::onDeviceConnected);
    }

    if (clockControlDialog) {
        connect(dev, &ModularMregDevice::clockControlStatusUpdated,
                clockControlDialog, &ClockControlDialog::setClockControlStatus);
        connect(dev, &ModularMregDevice::clockControlBpUpdated,
                clockControlDialog, &ClockControlDialog::setClockControlBp);
    }

    if (hptdcStatusDialog) {
        connect(dev, &ModularMregDevice::hptdcStatusUpdated,
                hptdcStatusDialog, &HptdcStatusDialog::deviceStatusUpdated);
    }

    if (adcMpdTrigDialog) {
        connect(dev, &ModularMregDevice::adcMpdTrigBpUpdated,
                adcMpdTrigDialog->getModel(), &AdcMpdTrigModel::setAdcMpdTrigBp);
        // connect(dev, &ModularMregDevice::adcMpdTrigStatusUpdated,
        //         adcMpdTrigDialog, &FeLinkStatusDialog::setAdcMpdTrigStatus);
    }

    if (adcSerDesCalDialog) {
        connect(dev, &ModularMregDevice::deserializersControlBpUpdated,
                adcSerDesCalDialog, &AdcSerDesCalDialog::setDeserializersControlBp);
        connect(dev, &ModularMregDevice::devSerDesDelayChanged,
                adcSerDesCalDialog, &AdcSerDesCalDialog::setDevSerDesDelay);
        connect(dev, &ModularMregDevice::adcSerDesProgressUpdated,
                adcSerDesCalDialog, &AdcSerDesCalDialog::progressUpdated);
    }
    if(sparseDialog) {
        connect(dev, &ModularMregDevice::adcBuildParamsUpdated,
                sparseDialog, &SparseDialog::deviceInfoUpdated);
    }

    if(crosspointSwitchDilog) {
        connect(dev, &ModularMregDevice::crosspointSwitchBpUpdated,
                crosspointSwitchDilog, &CrosspointSwitchDilog::crosspointSwitchBpUpdated);
    }

    if (readoutAppCore) {
        connect(devGlob, &MstreamMregDevGlobal::updateProgramDescription,
                readoutAppCore, &ReadoutAppCore::updateProgramDescription);

        connect(devGlob, &MstreamMregDevGlobal::closeProgram,
                readoutAppCore, &ReadoutAppCore::closeProgram);

        //    connect(dev, &ModularMregDevice::deviceConnected,
        //            readoutAppCore, &ReadoutAppCore::deviceConnected);
        connect(dev, &ModularMregDevice::deviceDisconnected,
                readoutAppCore, &ReadoutAppCore::deviceDisconnected);

        if (auto runLogicController = readoutAppCore->runLogicController) {
                connect(dev, &ModularMregDevice::runLogicStatusUpdated,
                        runLogicController, &RunLogicController::setRunLogicStatus);
        }
        connect(dev, &ModularMregDevice::statusUpdated,
                readoutAppCore, &ReadoutAppCore::setStatus);
        connect(readoutAppCore, &ReadoutAppCore::adcClkDivSync,
                                    dev, &ModularMregDevice::adcClkDivSync);
    }

    if(configDispatcher) {
        configDispatcher->forwardDeviceConfig(dev);
    }

    if (deviceMgrFsm) {
        connect(dev, &ModularMregDevice::deviceConfigWritten,
                deviceMgrFsm, &DeviceMgrFsm::setDevSetupKey);
    }
    if (deviceCommander) {
        connect(deviceCommander, &AsyncDeviceCommander::submitDeviceCommand,
                dev, &ModularMregDevice::submitDeviceCommand);
        connect(dev, &ModularMregDevice::deviceCommandComplete,
                deviceCommander, &AsyncDeviceCommander::deviceCommandComplete);
    }

    if(evNumChecker) {
        connect(evNumChecker, &EvNumChecker::evNumRequest, dev, &ModularMregDevice::evNumRequest);
        connect(dev, &ModularMregDevice::evNumResponse, evNumChecker, &EvNumChecker::gotDevEvNum);
    }
}
