/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MongoQt.h"
#include "MongoQt_p.h"
#include "BsonQtConverter.h"
#include "mongo/MongoConsts.h"
#include "BsonQtConverter.h"

#include <QDataStream>
#include <QDateTime>
#include <QDebug>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>
#include <QLoggingCategory>

#ifndef Q_MOC_RUN
#include <mongocxx/client.hpp>
#include <mongocxx/exception/exception.hpp>
#include <bsoncxx/json.hpp>
#endif

using bsoncxx::builder::basic::make_document;
using bsoncxx::builder::basic::kvp;
using namespace bsoncxx;
using namespace mongocxx;

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_mongoqt)
Q_LOGGING_CATEGORY(lc_mongoqt, "afi.mongoqt")
}

namespace {
QJsonObject dotNotation(const QJsonObject &obj) {
    QJsonObject ret;
    for(auto it=obj.begin(); it!=obj.end(); ++it) {
        if(it->type() == QJsonValue::Object) {
            QJsonObject child = dotNotation(it.value().toObject());
            for(auto chIt=child.begin(); chIt!=child.end(); ++chIt)
                ret[it.key()+"."+chIt.key()] = chIt.value();
        } else {
            ret[it.key()] = it.value();
        }
    }
    return ret;
}
template<typename T> bsoncxx::document::view_or_value mk_doc(std::string key, T val)
{
    return make_document(kvp(key, val));
}
}

MongoQt::MongoQt()
    : d_ptr(new MongoQtPrivate(this))
    , db(QString::fromStdString(DB))
    , collection_name(COLLECTION_SAVED)
{}

MongoQt::~MongoQt()
{
    delete d_ptr;
}

void MongoQt::set_db_collection(const QString &db_, const QString &collection_name_)
{
    db = db_;
    collection_name = collection_name_;
}

mongocxx::pool::entry MongoQt::get_connection()
{
    Q_D(MongoQt);
    return d->get_connection();
}

mongocxx::collection MongoQt::get_collection()
{
    Q_D(MongoQt);
    auto client = d->get_connection();
    return (*client)[db.toStdString()][collection_name.toStdString()];
}

bool MongoQt::set_uri(const QString &uri)
{
    Q_D(MongoQt);
    return d->configure(uri.toStdString());
}

bool MongoQt::try_connection()
{
    try {
        Q_D(MongoQt);
        auto client = d->get_connection();
        database mongo_db = (*client)[QString("admin").toStdString()];
        bsoncxx::document::value res = mongo_db.run_command(mk_doc("ping", 1));
        return res.view()["ok"].get_double();
    } catch (const mongocxx::exception &e){
        qCritical(lc_mongoqt).noquote() << __func__ << e.what();
        return false;
    }
}

int MongoQt::count_documents()
{
    try {
        auto collection = get_collection();
        return collection.count_documents({});
    } catch (const std::exception& xcp) {
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return 0;
}

QList<QJsonObject> MongoQt::find(const QJsonObject &query,
                                 const QJsonObject &projection)
{
    QList<QJsonObject> result;
    try {
        auto collection = get_collection();
        mongocxx::options::find opt{};
        bsoncxx::document::value doc_projection = BsonQtConverter::from_json(projection);
        auto queryInDotNotation = dotNotation(query);
        bsoncxx::document::value doc_query = BsonQtConverter::from_json(queryInDotNotation);
        opt.projection(doc_projection.view());
        auto cursor = collection.find(doc_query.view(), opt);
        for (const auto &i : cursor) {
            result.push_back(BsonQtConverter::to_json(i));
        }
    } catch (const std::exception& xcp) {
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return result;
}

bool MongoQt::replace_one(const QJsonObject &filter,
                          const QJsonObject &replacement,
                          bool upsert)
{
    try {
        if(filter.isEmpty() || replacement.isEmpty())
            return 0;

        auto collection = get_collection();
        options::replace options;
        options.upsert(upsert);

        bsoncxx::document::value doc_filter = BsonQtConverter::from_json(filter);
        bsoncxx::document::view_or_value doc_replacement = BsonQtConverter::from_json(replacement);

        const auto result = collection.replace_one(
            doc_filter.view(),
            doc_replacement.view(),
            options
        );

        int modified = result->modified_count();
        int upserted = result->upserted_id().has_value();
        return modified + upserted;
    } catch (const std::exception& xcp) {
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return false;
}

bool MongoQt::insert_one(const QJsonObject &obj)
{
    try {
        if(obj.isEmpty())
            return 0;
        auto collection = get_collection();
        bsoncxx::document::value doc = BsonQtConverter::from_json(obj);
        const auto insert_res = collection.insert_one(doc.view());
        const auto &result = insert_res->result();
        bool ok = (result.inserted_count() == 1);
        if (!ok)
            qWarning(lc_mongoqt).noquote() << __func__ << QString("inserted_count (%1) != 1")
                                    .arg(result.inserted_count());
        return ok;
    } catch (const std::exception& xcp){
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return false;
}

bool MongoQt::delete_one(const QJsonObject &obj)
{
    try {
        if(obj.isEmpty())
            return false;
        auto collection = get_collection();
        auto delete_result = collection.delete_one(
            BsonQtConverter::from_json(obj).view()
        );
        return delete_result->deleted_count() == 1;
    } catch (const std::exception& xcp) {
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return false;
}

int MongoQt::delete_many(const QJsonObject &obj)
{
    try {
        if(obj.isEmpty())
            return false;
        auto collection = get_collection();
        auto delete_result = collection.delete_many(
            BsonQtConverter::from_json(obj).view()
        );
        return delete_result->deleted_count();
    } catch (const std::exception& xcp) {
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return 0;
}

int MongoQt::delete_all()
{
    try {
        auto collection = get_collection();
        bsoncxx::document::view_or_value doc;
        auto delete_result = collection.delete_many(doc);
        return delete_result->deleted_count();
    } catch (const std::exception& xcp) {
        qCritical(lc_mongoqt).noquote() << __func__ << xcp.what();
    }
    return false;
}
