/*
**    Copyright 2011-2022 Ivan Filippov
**    Copyright 2011-2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MSTREAMDUMP_H
#define MSTREAMDUMP_H

#include <QElapsedTimer>
#include <QObject>
#include <QVector>

#include "AdcEventTrigInfo.h"
#include "MStreamDumpParameters.h"
#include "lib-common/MpdRawTypes.h"
#include "mstream-lib/MStreamStat.h"
#include "pnp-server/PNPMessage.h"
#include "tcpserver/TcpServerStatus.h"

class QTimer;
class QSemaphore;

class ConfigIndexName;
class DeviceIndex;
class LldpInfoCache;
class MStreamFileWriter;
class MldpListener;
class MlinkStreamReceiver;
class PacketBuffer;
class PacketRingBuffer;
class QSemaphore;
class RawCheck;
class ReceiverStat;
class TcpServer;
struct BaseFragment;
struct MStreamFragmentHeader;
struct MStreamStat;

class MStreamDump : public QObject , protected MStreamDumpParameters
{
    Q_OBJECT
public:
    explicit MStreamDump(const DeviceIndex &index, bool _doPrintStat = true);
    explicit MStreamDump(MStreamDumpParameters msParams, bool _doPrintStat = true);
    virtual ~MStreamDump() override;
    void setDeleteFlag();
    bool hasDeleteFlag() const;

public slots:
    void msdump_reset(const DeviceIndex &index, int seqNum, QString fn, bool doStatClear);
    void sendJsonBlock(const DeviceIndex &, const QByteArray &data);
    void sendCustomBlock(const DeviceIndex &, const QByteArray &data);
    void lldpCacheUpdated(const LldpInfoCache &dd);
    void connectToDevice(const DeviceIndex &index); //todo: make it private (w/o external control)
    void disconnectFromDevice(const DeviceIndex &index); //todo: make it private (w/o external control)
    void configSelectorChanged(const ConfigIndexName &cin);
    void requestLastRecord() const;
    void onPreRunCheck();

private slots:
    void start();
    void printStat();
    void gotData(const QVector<PacketBuffer> &tasks);
    void tcpServerStatusUpdated(const TcpServerStatus &status);
    void sendMetrics();
    void waitForData();
    void rxPortUpdated(quint16 port);
    void receiverDestroyed();
    void peerConnected();
    void peerDisconnected();

signals:
    void preRunCheck() const;
    void updateProgramDescription(const ProgramDescription&) const;
    void closeProgram(const ProgramDescription&) const;
    void connectToHardware(bool force = false);
    void disconnectFromHardware();
    void hostDiscovered(const QHostAddress &address);
    void releaseData(int);
    void stat_updated(const MStreamStat &stat);
    void stat_from_receiver_updated(const ReceiverStat &stat);
    void reset(int seqNum);
    void reseted(const DeviceIndex &index, int seqNum);
    void connectionStateChanged(const DeviceIndex &index, bool);
    void lastRecordUpdated(const QByteArray &) const;

private:
    bool verboseDataCheck(const char *buf, int recvSize);
    void processData(const PacketBuffer &pkbuf);
    void closeFragment(quint16 fragmentId, bool force=false);
    void closeAdc();
    void closeHrb(BaseFragment &fragment, const quint8 exactTrigPos, bool keepEvNum=true);
    void closeTimeSlice(BaseFragment &fragment, const quint8 usrDefBits);
    void sendHeaders();
    void flushData(const QByteArray &data);
    void addHistory(int fragId);
    void printHistory();
    void printFragParts(quint16 fragId);
    void initialize();
    void checkDeviceSerial(uint32_t serial);
    void debugPrintBufInHex(const char *buf, int size);
    void updateProgDescription();
    void createReciever();
    void destroyReciever();
    void print(QString s, bool noLog=false);
    QString formatStatText(int timeMs) const;
    bool isMyAddress(const QHostAddress &addr) const;
    void updateLastRecord();
    void sendJsonBlockInternal(const DeviceIndex &, const QByteArray &data);

    std::atomic<bool> deleteFlag{false};
    ProgramDescription progDescr;
    QSharedPointer<PacketRingBuffer> sharedBuffer;
    MlinkStreamReceiver * mr;
    bool mr_destroy = false;
    MStreamFileWriter * fw;
    TcpServer * const tcpServer;
    bool debugPrint = false;
    ClientIndex parentClientIndex;


    // < <deviceId, subtype>, <eventNum:24, <fragmentId, fragmentData> > >
//    QMap<DeviceKey,  BaseDevice> devices;
    QVector<BaseFragment> fragments; // <fragmentId (quint32:24), BaseEvent>
    AdcEventTrigInfo adcTrigInfo;

    MpdHdr mpdHdr;
    MpdEventHdr mpdEventHdr;
    MpdDeviceHdr mpdDeviceHdr;
    MpdMStreamHdr mpdMStreamHdr;

    const MStreamFragmentHeader *fragmentHdr;
//    quint16 deviceId;
    quint16 lastClosedFragmentId;
    bool initStage;
    QVector<quint16> histFrags;
    int histPos;
    MStreamStat prevStat, stat;
    QTime * const statTime;
    QTimer * const statTimer;
    QTimer * const waitDataTimer;
    QTimer * const metricTimer;
    QSet<int> initPackageSet;
    QVector<QPair<quint16,quint16> > lastEventFragments;
    quint32 lastSpillEvNum; // for HRB only
    quint32 prevEvNum; // for HRB only
    QString devIdStr;
    bool isDevBusy;
    int dataIndex;
    TcpServerStatus tcpStatus;
    uint16_t myReceiverPort;
    bool waitForDataEn=false;
    bool dumpAsApp;
    MpdRawDataDevRecord lastRec;
    MpdRawDataDevRecord curRec;
    bool lastRecSent = true;
    QElapsedTimer reconnectTimer;
};

#endif // MSTREAMDUMP_H
