/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MStreamV2Client.h"

#include <cstring>

#include <QDebug>
#include <QLoggingCategory>

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_ms_client)
Q_LOGGING_CATEGORY(lc_ms_client, "afi.mstream.client")
}
const int UDP_PORT_MSTREAM = 33301;

MStreamV2Client::~MStreamV2Client()
{
    close_mstream();
}

HwipSocketAddress MStreamV2Client::getLocalAddress() const
{
    HwipSocketAddress socketAddr;
    socketAddr.ip = ntohl(si_me.sin_addr.s_addr);
    socketAddr.port = ntohs(si_me.sin_port);
    return socketAddr;
}

HwipSocketAddress MStreamV2Client::getPeerAddress() const
{
    HwipSocketAddress socketAddr;
    socketAddr.ip = ntohl(si_other.sin_addr.s_addr);
    socketAddr.port = ntohs(si_other.sin_port);
    return socketAddr;
}

void MStreamV2Client::logErrno(const QString &message) const
{
    qCritical(lc_ms_client).noquote() << QString("[%1] recvmmsg: %2").arg(name, strerror(errno));
}

uint16_t MStreamV2Client::getLocalPort() const
{
    return ntohs(si_me.sin_port);
}

void MStreamV2Client::receivedFlags(quint32 seq, int flags)
{
    if ((protoState == SynSent) && (flags & Syn) && (flags & Ack)) {
        protoState = Open;
    }
}

int MStreamV2Client::recvmmsg (struct mmsghdr *vmessages, unsigned int vlen)
{
    if (-1 == s_mstream)
        return 0;
    int ret = ::recvmmsg(s_mstream, vmessages, vlen, MSG_WAITFORONE, nullptr);
    if (-1 == ret)
        if (errno != EAGAIN)
            logErrno("recvmmsg");
    return ret;
}

bool MStreamV2Client::send_mlink_packet(const QByteArray &payload)
{
    QByteArray qbuf;
    qbuf.resize(sizeof(MLinkFrameHdr) + payload.size() + 4);
    auto buf = qbuf.data();
    const auto bufsize = qbuf.size();

    MLinkFrameHdr *hdr = reinterpret_cast<MLinkFrameHdr *>(buf);
    uint32_t *crc = reinterpret_cast<uint32_t *>(&buf[sizeof(MLinkFrameHdr)+payload.size()]);
    hdr->sync = ML_FRAME_SYNC;
    hdr->type = FRAME_TYPE_MSTREAM;
    hdr->seq = mlinkSeq++;
    hdr->len = static_cast<uint16_t>(4 + payload.size() / 4);
    hdr->src = 1;
    hdr->dst = 0xFEFE;
    uint32_t *data = reinterpret_cast<uint32_t *>(&buf[sizeof(MLinkFrameHdr)]);
    std::memcpy(data, payload.constData(), payload.size());
    *crc = 0x12206249;

    int ret = sendto(s_mstream, buf, bufsize, 0,
                     reinterpret_cast<const sockaddr *>(&si_other), sizeof(si_other));
    if (-1 == ret)
        logErrno("sendto");
    return (ret > 0);
}

static uint32_t encode_mstream_v2_ack(int fid, int foff)
{
    return ((fid&0xFFFF) << 16) | (foff &0xFFFF);
}

bool MStreamV2Client::send_ack(const AckList &ackList)
{
    QVector<AckItem> buf;
    for (auto i : ackList) buf.push_back(i);
    if (false) {
        std::sort(buf.begin(), buf.end());
        auto last = std::unique(buf.begin(), buf.end());
        buf.erase(last, buf.end());
    }
    int mtu = 576;
    int max_packet_size = mtu - (24 + 8 + 16 + 4); // IP+opt, UDP, MLink, mstream
    int part_size = max_packet_size / 4;
    if (ackSizeLimit > 0 && part_size > ackSizeLimit)
        part_size = ackSizeLimit;
    int size = buf.size();
    if (size <= part_size)
        return send_ack_fragment(buf);
    bool ok = true;
    for (int start=0; start < size; start += part_size) {
        QVector<AckItem> part;
        for (int i=start; i<size; i++)
            part.push_back(buf[i]);
        send_ack_fragment(part);
    }
    return ok;
}

bool MStreamV2Client::send_ack_fragment(const QVector<AckItem> &ackList)
{
    if (ackList.empty())
        return true;
    int mtu = 576;
    int max_packet_size = mtu - (24 + 8 + 16 + 4); // IP+opt, UDP, MLink, mstream
    Q_ASSERT(ackList.size() <= max_packet_size);
    MStreamFragmentHeader h;
    h.deviceId = 1;
    h.flgs = Ack;
    h.fragmentLength  = 4 * (ackList.size() - 1);
    h.subType = 0;
    QByteArray ba;
    QVector<uint32_t> ackVect;
    for (const auto &i : ackList)
        ackVect.push_back(encode_mstream_v2_ack(i.fid, i.foff));
    auto rawData = reinterpret_cast<const char *>(&ackVect[0]);
    auto rawSize = ackVect.size() * sizeof(ackVect[0]);
    ba.resize(4 + rawSize);
    char *buf = ba.data();
    std::memcpy(buf, &h, 4);
    std::memcpy(buf+4, rawData, rawSize);
    return send_mlink_packet(ba);
}

void MStreamV2Client::runProtoFsm()
{
    switch (protoState) {
    case Init: break;
    case Closed: break;
    case SynSent: protoState = Open; break;
    case SynRcvd: protoState = Open; break;
    case Open: break;
    case CloseWait: protoState = Closed; break;
    default: break;
    }
}

void MStreamV2Client::setPeerAddress(const QHostAddress &address)
{
    peerAddress = address;
}

bool MStreamV2Client::connectToHardware()
{
    if (-1 != s_mstream)
        return true;
    qInfo(lc_ms_client).noquote() << QString("[%1] connectToHardware(%2), socket=%3")
                                         .arg(name, peerAddress.toString()).arg(s_mstream);
    close_mstream();
    if (peerAddress.isNull())
        return true;
    init_si_other(peerAddress);
    bool ok = bind_mstream();
    protoState = Closed;

    ok &= sendProtoFlags(Syn);
    protoState = SynSent;

    // TODO: wait for Syn+Ack received
    ok &= sendProtoFlags(Ack);
    protoState = Open;
    if (ok)
        qInfo(lc_ms_client).noquote() << QString("[%1] connected to %2")
                                             .arg(name, peerAddress.toString());
    else
        qCritical(lc_ms_client).noquote() << QString("[%1] unable to connect to %2")
                                             .arg(name, peerAddress.toString());
    return ok;
}

bool MStreamV2Client::disconnectFromHardware()
{
    bool ok = true;

    // TCP style handshake
    if (protoState == Open || protoState == SynRcvd || protoState == CloseWait)
        ok &= sendProtoFlags(Fin);
    ok &= close_mstream();
    return ok;
}

bool MStreamV2Client::sendProtoFlags(int flags)
{
    MStreamFragmentHeader h;
    h.deviceId = 1;
    h.flgs = flags;
    h.fragmentLength  = 0;
    h.fragmentOffsetId = -1;
    h.subType = 0;
    auto payload = QByteArray::fromRawData(reinterpret_cast<const char *>(&h), sizeof(h));
    return send_mlink_packet(payload);
}

bool MStreamV2Client::close_mstream()
{
    if (-1 == s_mstream) {
        protoState = Init;
        return true;
    }
    sendProtoFlags(Rst);
    protoState = CloseWait;
    // ...
    protoState = Closed;
    auto port = getLocalPort();
    int ret = close(s_mstream);
    if (-1 == ret)
        logErrno("close");
    else
        qInfo(lc_ms_client).noquote() << QString("[%1] closed port %2").arg(name).arg(port);
    s_mstream = -1;
    protoState = Init;
    return (0 == ret);
}

bool MStreamV2Client::isBound() const
{
    return (-1 != s_mstream) && (si_me.sin_port != 0);
}

bool MStreamV2Client::isConnected() const
{
    return protoState == Open;
}

bool MStreamV2Client::bind_mstream()
{
    s_mstream = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (-1 == s_mstream) {
        logErrno("socket");
        return false;
    }

    memset(static_cast<void *>(&si_me), 0, sizeof(si_me));
    si_me.sin_family = AF_INET;
    si_me.sin_addr.s_addr = htonl(INADDR_ANY);
    si_me.sin_port = 0;
    if (-1 == bind(s_mstream, reinterpret_cast<const sockaddr *>(&si_me), sizeof(si_me))) {
        logErrno("bind");
        return false;
    } else {
        timeval tv;
        tv.tv_sec = 0;
        tv.tv_usec = timeout;
        int ret = setsockopt(s_mstream, SOL_SOCKET, SO_RCVTIMEO, &tv,sizeof(tv));
        if (-1 == ret)
            logErrno("setsockopt");
        socklen_t len = sizeof(si_me);
        getsockname(s_mstream, reinterpret_cast<sockaddr *>(&si_me), &len);
    }
    qInfo(lc_ms_client).noquote() << QString("[%1] listening on port %2").arg(name).arg(getLocalPort());
    return true;
}

void MStreamV2Client::init_si_other(const QHostAddress &ipaddr)
{
    const in_addr_t peer_addr = inet_addr(ipaddr.toString().toLatin1());

    memset(reinterpret_cast<char *>(&si_other), 0, sizeof(si_other));
    si_other.sin_family = AF_INET;
    si_other.sin_addr.s_addr = peer_addr;
    si_other.sin_port = htons(UDP_PORT_MSTREAM);
//    cerr << QString("[MStream] [%1]: MlinkStreamReceiver receiving stream from %2:%3")
//                .arg(stat.index.getSerialStr())
//                .arg(inet_ntoa(si_other.sin_addr)).arg(ntohs(si_other.sin_port)).toStdString().c_str()
//         << endl;
}
