#include "MultiCastListener.h"

#include <QHostAddress>
#include <QTimer>
#include <QPointer>

//#include "impl/MulticastInterfaceReceiver.h"
#include "impl/SingleMultiCastListener.h"
#include "base/ThreadController.h"

static const int interfaceWatchIntervalMs = 1000;

static bool isMulticastCapable(const QNetworkInterface &interface)
{
    if (interface.flags() & QNetworkInterface::IsLoopBack) return false;
    if (! (interface.flags() & QNetworkInterface::IsUp)) return false;
    if (! (interface.flags() & QNetworkInterface::IsRunning)) return false;
    if (! (interface.flags() & QNetworkInterface::CanMulticast)) return false;
    const QList<QNetworkAddressEntry> &addressEntries = interface.addressEntries();
    if (addressEntries.isEmpty()) return false;
    return true;
}

MultiCastListener::MultiCastListener(const QHostAddress &_host, quint16 _port, QObject *parent) :
    QObject(parent),
    host(_host),
    port(_port)
{
    QTimer::singleShot(0, this, &MultiCastListener::watchInterfaces);
}

MultiCastListener::~MultiCastListener()
{
    receivers.clear();
}

void MultiCastListener::watchInterfaces()
{
    QList<int> currentInterfaces;
    for (const auto &interface : QNetworkInterface::allInterfaces()) {
        if (! isMulticastCapable(interface))
            continue;
        currentInterfaces.push_back(interface.index());
    }
    const auto &currentIndices = receivers.keys();
    bool isInterfacesUpdated = false;
    for (const auto &i : currentIndices) {
        if (!currentInterfaces.contains(i)) {
            stopInterface(i);
            isInterfacesUpdated = true;
        }
    }
    for (auto i : currentInterfaces) {
        if (! currentIndices.contains(i)) {
            startInterface(i);
            isInterfacesUpdated = true;
        }
    }
    if(isInterfacesUpdated)
        emit multicastInterfacesUpdated();
    QTimer::singleShot(interfaceWatchIntervalMs, this, &MultiCastListener::watchInterfaces);
}

void MultiCastListener::startInterface(int index)
{
    const auto &interface= QNetworkInterface::interfaceFromIndex(index);
    auto listener = new SingleMultiCastListener(this, interface, host, port);
    listener->setObjectName(QString("%1 multicast").arg(interface.humanReadableName()));
    connect(listener, &SingleMultiCastListener::received, this, &MultiCastListener::received);
    auto controller = new ThreadController(listener, this);
    receivers.insert(index, controller);
}

void MultiCastListener::stopInterface(int index)
{
    const auto &interface = QNetworkInterface::interfaceFromIndex(index);
    qDebug() << QString("interface %1 is down").arg(interface.humanReadableName());
    delete receivers.take(index);
}
